/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	b64 "encoding/base64"
	"encoding/json"
	"errors"
	"fmt"
	"os"
	"path"
	"regexp"
	"strconv"
	"strings"
	"text/tabwriter"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/bckapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	"github.com/jinzhu/copier"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
)

// -------------------------------------------------
// Types definition --------------------------------
// -------------------------------------------------

// -------------------------------------------------
// -------------------------------------------------

// -------------------------------------------------
// Global variable (inside this package) -----------
// -------------------------------------------------

// Applications cached
var cachedRPApplication map[uuid.UUID]rpapiv1.RPApplication

// -------------------------------------------------
// -------------------------------------------------

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

// Initialization function for the applications commands
func initAppCmd(rootCmd *cobra.Command) {
	// appsCmd represents the applications command
	var appsCmd = &cobra.Command{
		Use:     appsCommand,
		Aliases: []string{appsCommandAlias},
		Short:   appsCommandShortUsage,
	}

	// listAppsCmd represents the "applications list" command
	var listAppsCmd = &cobra.Command{
		Use:     appsListCmd,
		Aliases: []string{appsListCmdAlias},
		Short:   appsListCmdShortUsage,
		RunE:    listApplications,
		Annotations: map[string]string{
			"__flag_need_bck_connect__": "TRUE",
			"__read_object_command__":   "TRUE"},
	}

	// getAppsCmd represents the "applications get" command
	var getAppsCmd = &cobra.Command{
		Use:               appsGetCmd,
		Short:             appsGetCmdShortUsage,
		RunE:              getApplication,
		ValidArgsFunction: getFullListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__read_object_command__":  "TRUE"},
	}

	// addAppsCmd represents the "applications add" command
	var addAppsCmd = &cobra.Command{
		Use:     appsAddCmd,
		Aliases: []string{appsAddAlias},
		Short:   appsAddCmdShortUsage,
		PreRunE: initEventApp,
		RunE:    createApplication,
		Annotations: map[string]string{
			"__flag_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":        "TRUE"},
	}

	// addAppsFromSampleCmd represents the "applications add-from-sample" command
	var addAppsFromSampleCmd = &cobra.Command{
		Use:               appsAddFromSampleCmd,
		Short:             appsAddFromSampleCmdShortUsage,
		Long:              appsAddFromSampleCmdLongUsage,
		PreRunE:           initEventApp,
		RunE:              createApplicationFromSample,
		ValidArgsFunction: getListAppsSamplesAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// updateAppsCmd represents the "applications update" command
	var updateAppsCmd = &cobra.Command{
		Use:               appsUpdateCmd,
		Aliases:           []string{appsUpdateAlias},
		Short:             appsUpdateCmdShortUsage,
		PreRunE:           initEventApp,
		RunE:              updateApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// delAppsCmd represents the "applications delete" command
	var delAppsCmd = &cobra.Command{
		Use:               appsDelCmd,
		Aliases:           []string{appsDelAlias},
		Short:             appsDelCmdShortUsage,
		PreRunE:           initEventApp,
		RunE:              deleteApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// bookmarkAppCmd represents the "applications bookmark" command
	var bookmarkAppCmd = &cobra.Command{
		Use:               appsBookmarkCmd,
		Aliases:           []string{appsBookmarkAlias},
		Short:             appsBookmarkCmdShortUsage,
		PreRunE:           initEventApp,
		RunE:              bookmarkApplication,
		ValidArgsFunction: getFullListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__read_object_command__":  "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// contentAppsCmd represents the "applications content" command
	var contentAppsCmd = &cobra.Command{
		Use:               appsContentCmd,
		Aliases:           []string{appsContentAlias},
		Short:             appsContentCmdShortUsage,
		RunE:              contentSpecApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE"},
	}

	// removeAppsCmd represents the "applications remove" command
	var removeAppsCmd = &cobra.Command{
		Use:               appsRemoveCmd,
		Aliases:           []string{appsRemoveAlias},
		Short:             appsRemoveCmdShortUsage,
		RunE:              removeFilesApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE"},
	}

	// downloadAppsCmd represents the "applications download" command
	var downloadAppsCmd = &cobra.Command{
		Use:               appsDownloadCmd,
		Aliases:           []string{appsDownloadAlias},
		Short:             appsDownloadCmdShortUsage,
		RunE:              downloadFilesApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE"},
	}

	// uploadAppsCmd represents the "applications upload" command
	var uploadAppsCmd = &cobra.Command{
		Use:               appsUploadCmd,
		Aliases:           []string{appsUploadAlias},
		Short:             appsUploadCmdShortUsage,
		PreRunE:           initEventApp,
		RunE:              uploadFilesApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// buildAppsCmd represents the "applications build" command
	var buildAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildCmd, "app-ID"),
		Aliases:           []string{genericBuildAlias},
		Short:             fmt.Sprintf(genericBuildCmdShortUsage, "application"),
		Long:              fmt.Sprintf(genericBuildCmdLongUsage, "application", "application", "applications"),
		PreRunE:           initEventBuildAndApp,
		RunE:              buildApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// buildsCmd represents the "applications builds" command
	var buildsCmd = &cobra.Command{
		Use:   genericBuildsCmd,
		Short: fmt.Sprintf(genericBuildsCmdShortUsage, "application"),
	}

	// buildsListAppsCmd represents the "applications builds list" command
	var buildsListAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericSubBuildsListCmd, "app-ID"),
		Aliases:           []string{genericSubBuildsListAlias},
		Short:             fmt.Sprintf(genericSubBuildsListCmdShortUsage, "application"),
		RunE:              buildsListApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsGetAppsCmd represents the "applications builds get" command
	var buildsGetAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildsGetCmd, "app-ID"),
		Short:             genericBuildsGetCmdShortUsage,
		Long:              fmt.Sprintf(genericBuildsGetCmdLongUsage, "application", "application", "app"),
		RunE:              appBuildsGetAndDisplay,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsCancelAppsCmd represents the "applications builds cancel" command
	var buildsCancelAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildsCancelCmd, "app-ID"),
		Short:             genericBuildsCancelCmdShortUsage,
		Long:              fmt.Sprintf(genericBuildsCancelCmdLongUsage, "application", "application", "app"),
		RunE:              buildCancelApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsLogsAppsCmd represents the "applications builds logs" command
	var buildsLogsAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildLogsCmd, "app-ID"),
		Short:             fmt.Sprintf(genericBuildLogsCmdShortUsage, "application"),
		Long:              fmt.Sprintf(genericBuildLogsCmdLongUsage, "application", "application", "application", "app"),
		RunE:              buildGetLogsApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsStatsCmd represents the "applications builds stats" command
	var buildsStatsCmd = &cobra.Command{
		Use:               appsBuildsStatsCmd,
		Aliases:           []string{appsBuildsStatsAlias},
		Short:             appsBuildsStatsCmdShortUsage,
		SilenceUsage:      true,
		RunE:              getAppBuildsStats,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// listAppsCmd represents the "applications list" command
	var seriousCmd = &cobra.Command{
		Use:          "serious",
		SilenceUsage: true,
		Hidden:       true,
		RunE:         seriousCmdHandler,
	}

	// testAppsCmd represents the "applications test" command
	var testAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestCmd, "app-ID"),
		Aliases:           []string{genericTestAlias},
		Short:             fmt.Sprintf(genericTestCmdShortUsage, "application"),
		Long:              fmt.Sprintf(genericTestCmdLongUsage, "application", "application", "application", "--force-app"),
		PreRunE:           initEventTests,
		RunE:              testGeneric,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// testsCmd represents the "applications tests" command
	var testsCmd = &cobra.Command{
		Use:   genericTestsCmd,
		Short: fmt.Sprintf(genericTestsCmdShortUsage, "application"),
	}

	// testsListAppsCmd represents the "applications tests list" command
	var testsListAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestsListCmd, "app-ID"),
		Aliases:           []string{genericTestsListAlias},
		Short:             fmt.Sprintf(genericTestsListCmdShortUsage, "application"),
		RunE:              testsListGeneric,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testsGetAppsCmd represents the "applications tests get" command
	var testsGetAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestsGetCmd, "app-ID"),
		Short:             genericTestsGetCmdShortUsage,
		Long:              fmt.Sprintf(genericTestsGetCmdLongUsage, "application", "application", "--force-app"),
		RunE:              testsGetGeneric,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testsLogsAppsCmd represents the "applications tests logs" command
	var testsLogsAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestLogsCmd, "app-ID"),
		Short:             fmt.Sprintf(genericTestLogsCmdShortUsage, "application"),
		Long:              fmt.Sprintf(genericTestLogsCmdLongUsage, "application", "application", "application", "--force-app"),
		RunE:              testGetLogsGeneric,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testCancelAppsCmd represents the "applications tests cancel" command
	var testsCancelAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestCancelCmd, "app-ID"),
		Short:             genericTestCancelCmdShortUsage,
		Long:              fmt.Sprintf(genericTestCancelCmdLongUsage, "application", "application", "--force-app"),
		RunE:              testCancelGeneric,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testsStatsCmd represents the "applications tests stats" command
	var testsStatsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestsStatsCmd, "app-ID"),
		Aliases:           []string{genericTestsStatsAlias},
		Short:             fmt.Sprintf(genericTestsStatsCmdShortUsage, "application"),
		SilenceUsage:      true,
		RunE:              getAppTestsStats,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// servicesCmd represents the "applications services" command
	var servicesCmd = &cobra.Command{
		Use:   appsServicesCmd,
		Short: appsServicesCmdShortDesc,
		RunE:  rpServices,
	}

	// auditAppsCmd represents the "applications audit" command
	var auditAppsCmd = &cobra.Command{
		Use:               appsAuditCmd,
		Aliases:           []string{appsAuditAlias},
		Short:             appsAuditCmdShortUsage,
		PreRunE:           initEventAuditApp,
		RunE:              auditApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// auditsCmd represents the "applications audits" command
	var auditsCmd = &cobra.Command{
		Use:   appsAuditsCmd,
		Short: appsAuditsCmdShortUsage,
	}

	// auditsListToolsCmd represents the "applications audits tools" command
	var auditsListToolsCmd = &cobra.Command{
		Use:     appsAuditsListToolsCmd,
		Aliases: []string{appsAuditsListToolsAlias},
		Short:   appsAuditsListToolsCmdShortUsage,
		RunE:    auditsListTools,
	}

	// auditsListCmd represents the "applications audits list" command
	var auditsListCmd = &cobra.Command{
		Use:               appsAuditsListCmd,
		Aliases:           []string{appsAuditsListAlias},
		Short:             appsAuditsListCmdShortUsage,
		RunE:              auditsList,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// auditsGetCmd represents the "applications audits get" command
	var auditsGetCmd = &cobra.Command{
		Use:               appsAuditsGetCmd,
		Short:             appsAuditsGetCmdShortUsage,
		Long:              appsAuditsGetCmdLongUsage,
		RunE:              auditsGetAndDisplay,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// auditsLogsCmd represents the "applications audits logs" command
	var auditsLogsCmd = &cobra.Command{
		Use:         appsAuditsLogsCmd,
		Short:       appsAuditsLogsCmdShortUsage,
		Long:        appsAuditsLogsCmdLongUsage,
		RunE:        auditGetLogs,
		Annotations: map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// branchCmd represents the "applications branch" command
	var branchCmd = &cobra.Command{
		Use:               appsBranchCmd,
		Short:             appsBranchCmdShortDesc,
		Long:              appsBranchCmdLongDesc,
		RunE:              appCreateOrSwitchBranch,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// branchesCmd represents the "applications branches" command
	var branchesCmd = &cobra.Command{
		Use:   appsBranchesCmd,
		Short: appsBranchesCmdShortUsage,
	}

	// branchesListCmd represents the "applications branches list" command
	var branchesListCmd = &cobra.Command{
		Use:               appsBranchesListCmd,
		Aliases:           []string{appsBranchesListAlias},
		Short:             appsBranchesListCmdShortUsage,
		RunE:              appBranchesList,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// branchesDeleteCmd represents the "applications branches delete" command
	var branchesDeleteCmd = &cobra.Command{
		Use:               appsBranchesDelCmd,
		Aliases:           []string{appsBranchesDelCmdAlias},
		Short:             appsBranchesDelCmdShortUsage,
		RunE:              appBranchesDelete,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// copyCmd represents the "applications copy" command
	var copyAppCmd = &cobra.Command{
		Use:               appsCopyCmd,
		Aliases:           []string{appsCopyAlias},
		Short:             appsCopyCmdShortDesc,
		Long:              appsCopyCmdLongDesc,
		RunE:              appCopyToProject,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// refreshAppCmd represents the "applications refresh-files-pkg" command
	var refreshAppCmd = &cobra.Command{
		Use:               appsRefreshFilesPkgCmd,
		Aliases:           []string{appsRefreshFilesPkgAlias},
		Short:             appsRefreshFilesPkgShortDesc,
		PreRunE:           initEventApp,
		RunE:              appRefreshGitpkgFiles,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	// Define the flags specific to "list" command
	listAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	listAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	listAppsCmd.Flags().BoolP(appsListCmdBookmarkedFlag, appsListCmdBookmarkedFlagAlias, appsListCmdBookmarkedFlagDefault, appsListCmdBookmarkedFlagUsage)
	// Define the flags specific to "get" command
	getAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	getAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "build" command
	buildAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	buildAppsCmd.Flags().StringSliceP(appsBuildCmdDistroFlag, appsBuildCmdDistroFlagAlias, []string{}, appsBuildCmdDistroFlagUsage)
	buildAppsCmd.Flags().StringP(appsBuildCmdCommentFlag, appsBuildCmdCommentFlagAlias, "", appsBuildCmdCommentFlagUsage)
	buildAppsCmd.Flags().BoolP(appsBuildCmdSrpmFlag, appsBuildCmdSrpmFlagAlias, false, appsBuildCmdSrpmFlagUsage)
	buildAppsCmd.Flags().BoolP(appsBuildCmdCheckDepsFlag, appsBuildCmdCheckDepsFlagAlias, false, appsBuildCmdCheckDepsFlagUsage)
	// Define the flags specific to "test" command
	testAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	testAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	testAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	testAppsCmd.Flags().StringP(genericTestCmdCommentFlag, genericTestCmdCommentFlagAlias, "", genericTestCmdCommentFlagUsage)
	// Define the flags specific to "builds list" command
	buildsListAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	buildsListAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	buildsListAppsCmd.Flags().BoolP(appsSubBuildsListCmdSRpmFlag, appsSubBuildsListCmdSRpmFlagAlias, false, appsSubBuildsListCmdSRpmFlagUsage)
	// Define the flags specific to "builds get" command
	buildsGetAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	buildsGetAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	buildsGetAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "builds logs" command
	buildsLogsAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	buildsLogsAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	buildsLogsAppsCmd.Flags().StringP(genericBuildLogsDirFlag, genericBuildLogsDirFlagAlias, genericBuildLogsDirFlagDefault, genericBuildLogsDirFlagUsage)
	buildsLogsAppsCmd.Flags().BoolP(genericBuildZipFlag, genericBuildZipFlagAlias, false, genericBuildZipFlagUsage)
	// Define the flags specific to "builds cancel" command
	buildsCancelAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	buildsCancelAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "builds stats" command
	buildsStatsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	buildsStatsCmd.Flags().Int(appsBuildsStatsCmdLastFlag, 0, appsBuildsStatsCmdLastFlagUsage)
	buildsStatsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Define the flags specific to "tests list" command
	testsListAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	testsListAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "tests get" command
	testsGetAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	testsGetAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	testsGetAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "tests logs" command
	testsLogsAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	testsLogsAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	testsLogsAppsCmd.Flags().StringP(genericBuildLogsDirFlag, genericBuildLogsDirFlagAlias, genericBuildLogsDirFlagDefault, genericBuildLogsDirFlagUsage)
	testsLogsAppsCmd.Flags().BoolP(genericBuildZipFlag, genericBuildZipFlagAlias, false, genericBuildZipFlagUsage)
	// Define the flags specific to "tests cancel" command
	testsCancelAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	testsCancelAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	// Define the flags specific to "tests stats" command
	testsStatsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	testsStatsCmd.Flags().Int(genericTestsStatsCmdLastFlag, 0, genericTestsStatsCmdLastFlagUsage)
	testsStatsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Define the flags specific to "delete" command
	delAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "bookmark" command
	bookmarkAppCmd.Flags().BoolP(appsBookmarkCmdResetFlag, appsBookmarkCmdResetFlagAlias, false, appsBookmarkCmdResetFlagUsage)
	bookmarkAppCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Slice flags needed for application list
	listAppsCmd.Flags().StringSlice(appsListCmdTypeFlag, []string{"standard"}, appsListCmdTypeFlagUsage)

	// Slice flags needed for application creation/update
	// Creation
	addAppsCmd.Flags().StringSlice(appsAddCmdMandDistFlag, []string{}, appsAddCmdMandDistFlagUsage)
	addAppsCmd.Flags().StringSlice(appsAddCmdOptDistFlag, []string{}, appsAddCmdOptDistFlagUsage)
	// Update
	updateAppsCmd.Flags().StringSlice(appsAddCmdMandDistFlag, []string{flagUpdateKeepCurrentValue}, appsAddCmdMandDistFlagUsage)
	updateAppsCmd.Flags().StringSlice(appsAddCmdOptDistFlag, []string{flagUpdateKeepCurrentValue}, appsAddCmdOptDistFlagUsage)

	// Other flags needed for application creation/update
	// Creation
	addAppsCmd.Flags().StringP(appsAddCmdNameFlag, appsAddCmdNameFlagAlias, "", appsAddCmdNameFlagUsage)
	addAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsAddCmdPrjFlagUsage)
	addAppsCmd.Flags().StringP(appsAddCmdAliasFlag, appsAddCmdAliasFlagAlias, "", appsAddCmdAliasFlagUsage)
	addAppsCmd.Flags().StringP(appsAddCmdDescFlag, appsAddCmdDescFlagAlias, "", appsAddCmdDescFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdPkgFlag, "", appsAddCmdPkgFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdSrcURLFlag, "", appsAddCmdSrcURLFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdSrcRevFlag, "", appsAddCmdSrcRevFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdExtSpecFlag, "", appsAddCmdExtSpecFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdSrcSpecFlag, "", appsAddCmdSrcSpecFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdLocalSpecFlag, "", appsAddCmdLocalSpecFlagUsage)
	addAppsCmd.Flags().Bool(appsAddCmdAutoTestFlag, false, appsAddCmdAutoTestFlagUsage)
	addAppsCmd.Flags().Bool(appsAddCmdAutoSRpmBuildFlag, appsAddCmdAutoSRpmBuildFlagDefault, appsAddCmdAutoSRpmBuildFlagUsage)
	addAppsCmd.Flags().StringP(appsAddCmdBuildTypeFlag, appsAddCmdBuildTypeFlagAlias, appsAddCmdBuildTypeFlagDefault, appsAddCmdBuildTypeFlagUsage)
	addAppsCmd.Flags().Bool(appsAddCmdAutoAuditFlag, false, appsAddCmdAutoAuditFlagUsage)
	addAppsCmd.Flags().StringSlice(appsAddCmdDefAuditFlag, []string{}, appsAddCmdDefAuditFlagUsage)
	addAppsCmd.Flags().StringP(appsAddCmdActiveBranchFlag, appsAddCmdActiveBranchFlagAlias, "", appsAddCmdActiveBranchFlagUsage)

	// App services
	addAppsCmd.Flags().StringArrayP(appsAddCmdServiceFlag, appsAddCmdServicesFlagAlias, []string{}, appsAddCmdServicesFlagUsage)

	seriousCmd.Flags().Bool("why", false, "")
	// Creation from sample
	addAppsFromSampleCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsAddCmdPrjFlagUsage)
	// Update
	updateAppsCmd.Flags().StringP(appsAddCmdNameFlag, appsAddCmdNameFlagAlias, flagUpdateKeepCurrentValue, appsAddCmdNameFlagUsage)
	updateAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	updateAppsCmd.Flags().StringP(appsAddCmdDescFlag, appsAddCmdDescFlagAlias, flagUpdateKeepCurrentValue, appsAddCmdDescFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdSrcRevFlag, flagUpdateKeepCurrentValue, appsAddCmdSrcRevFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdSrcURLFlag, flagUpdateKeepCurrentValue, appsUpdateCmdSrcURLFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdExtSpecFlag, flagUpdateKeepCurrentValue, appsAddCmdExtSpecFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdSrcSpecFlag, flagUpdateKeepCurrentValue, appsAddCmdSrcSpecFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdAutoTestFlag, flagUpdateKeepCurrentValue, appsAddCmdAutoTestFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdAutoSRpmBuildFlag, flagUpdateKeepCurrentValue, appsAddCmdAutoSRpmBuildFlagUsage)
	updateAppsCmd.Flags().StringP(appsAddCmdBuildTypeFlag, appsAddCmdBuildTypeFlagAlias, flagUpdateKeepCurrentValue, appsAddCmdBuildTypeFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdAutoAuditFlag, flagUpdateKeepCurrentValue, appsAddCmdAutoAuditFlagUsage)
	updateAppsCmd.Flags().StringSlice(appsAddCmdDefAuditFlag, []string{flagUpdateKeepCurrentValue}, appsAddCmdDefAuditFlagUsage)

	// Update app services
	updateAppsCmd.Flags().StringArrayP(appsAddCmdServiceFlag, appsAddCmdServicesFlagAlias, []string{flagUpdateKeepCurrentValue}, appsAddCmdServicesFlagUsage)
	updateAppsCmd.Flags().StringSlice(appsUpdateCmdDisableServiceFlag, []string{}, appsUpdateCmdDisableServicesFlagUsage)

	// Hidden updates flags
	updateAppsCmd.Flags().Bool(appsForceStateFlag, false, appsForceStateFlagUsage)
	updateAppsCmd.Flags().MarkHidden(appsForceStateFlag)

	// Flag for content list
	contentAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	contentAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flag for remove file
	removeAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flag for download file
	downloadAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flag needed for spec file upload
	uploadAppsCmd.Flags().String(appsUploadCmdFilePathFlag, "", appsUploadCmdFilePathFlagUsage)
	uploadAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flag for the test command
	testAppsCmd.Flags().StringP(appsTestCmdImageFlag, appsTestCmdImageFlagAlias, "", appsTestCmdImageFlagUsage)
	testAppsCmd.Flags().StringP(genericTestCmdBoardModelFlag, genericTestCmdBoardModelFlagAlias, "", genericTestCmdBoardModelFlagUsage)
	testAppsCmd.Flags().Int(genericTestCmdTimeOutFlag, 5, genericTestCmdTimeOutFlagUsage)
	testAppsCmd.Flags().Bool(genericTestCmdProgressFlag, genericTestCmdProgressFlagDefault, genericTestCmdProgressFlagUsage)

	// Flag for the available services command
	servicesCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flag for the audit command
	auditAppsCmd.Flags().StringSlice(appsAuditToolFlag, []string{}, appsAuditToolFlagUsage)
	auditAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flag for the audits tools list command
	auditsListToolsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flag for the audits list command
	auditsListCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	auditsListCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flags for the audits get command
	auditsGetCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	auditsGetCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	auditsGetCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flags for the audits logs command
	auditsLogsCmd.Flags().StringP(genericBuildLogsDirFlag, genericBuildLogsDirFlagAlias, genericBuildLogsDirFlagDefault, genericBuildLogsDirFlagUsage)
	auditsLogsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	auditsLogsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flags for the branch command
	branchCmd.Flags().StringP(appsBranchSourceFlag, appsBranchSourceFlagAlias, "", appsBranchSourceFlagUsage)
	branchCmd.Flags().StringP(appsBranchDestinationFlag, appsBranchDestinationFlagAlias, "", appsBranchDestinationFlagUsage)
	branchCmd.Flags().Bool(appsBranchDefaultFlag, false, appsBranchDefaultFlagUsage)
	branchCmd.Flags().Bool(appsBranchNotActiveFlag, false, appsBranchNotActiveFlagUsage)
	branchCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	branchCmd.Flags().StringSlice(appsBranchDistrosFlag, []string{}, appsBranchDistrosFlagUsage)
	branchCmd.Flags().BoolP(appsBranchResetDistrosFlag, appsBranchResetDistrosFlagAlias, false, appsBranchResetDistrosFlagUsage)

	// Flags for the branch list command
	branchesListCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	branchesListCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flags for the branch delete command
	branchesDeleteCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	branchesDeleteCmd.Flags().StringP(appsBranchFlag, appsBranchFlagAlias, "", appsBranchFlagUsage)

	// Flags for the copy app command
	copyAppCmd.Flags().StringP(appCopyCmdDestinationPrjFlag, appCopyCmdDestinationPrjAlias, "", appCopyCmdDestinationPrjFlagUsage)
	copyAppCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flags for the refresh gitpkgs files command
	refreshAppCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required
	addAppsCmd.MarkFlagRequired(appsAddCmdPkgFlag)
	addAppsCmd.MarkFlagRequired(appsAddCmdNameFlag)

	// Specify dynamic flag completion
	addAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	addAppsFromSampleCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	listAppsCmd.RegisterFlagCompletionFunc(appsListCmdTypeFlag, getListTypeAsTab)
	listAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getFullListPrjsAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	uploadAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	getAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getFullListPrjsAsTab)
	buildAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	buildAppsCmd.RegisterFlagCompletionFunc(appsBuildCmdDistroFlag, getListDistAsTab)
	buildsStatsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testAppsCmd.RegisterFlagCompletionFunc(appsTestCmdImageFlag, getListImagesAsTab)
	testAppsCmd.RegisterFlagCompletionFunc(genericTestCmdBoardModelFlag, getListRTMBoardModelsAsTab)
	buildsListAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	buildsLogsAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	buildsCancelAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testsListAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testsLogsAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testsCancelAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testsStatsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	delAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	// Creation
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdMandDistFlag, getListDistAsTab)
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdOptDistFlag, getListDistAsTab)
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdDefAuditFlag, getAuditToolsAsTab)
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdBuildTypeFlag, getBuildTypesAsTab)
	// Add app services
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdServiceFlag, getServicesNamesAsTab)
	// Update
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdMandDistFlag, getListDistAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdOptDistFlag, getListDistAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdDefAuditFlag, getAuditToolsAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdBuildTypeFlag, getBuildTypesAsTab)
	// Update app services
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdServiceFlag, getServicesNamesAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsUpdateCmdDisableServiceFlag, getServicesNamesAsTab)
	// Audit
	auditAppsCmd.RegisterFlagCompletionFunc(appsAuditToolFlag, getAuditToolsAsTab)
	// Branches
	branchCmd.RegisterFlagCompletionFunc(appsBranchSourceFlag, getAppBranchesAsTab)
	branchCmd.RegisterFlagCompletionFunc(appsBranchDestinationFlag, getAppBranchesAsTab)
	branchCmd.RegisterFlagCompletionFunc(appsBranchDistrosFlag, getListDistAsTab)
	branchesDeleteCmd.RegisterFlagCompletionFunc(appsBranchFlag, getAppBranchesAsTab)

	// App Copy
	copyAppCmd.RegisterFlagCompletionFunc(appCopyCmdDestinationPrjFlag, getListPrjsAsTab)
	copyAppCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)

	// App refresh
	refreshAppCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)

	// Add the group commands "applications" as subcommand of rootCmd
	rootCmd.AddCommand(appsCmd)
	rootCmd.AddCommand(seriousCmd)

	// Add the command "list" as a subcommand of appsCmd
	appsCmd.AddCommand(listAppsCmd)
	// Add the command "get" as a subcommand of appsCmd
	appsCmd.AddCommand(getAppsCmd)
	// Add the command "add" as a subcommand of appsCmd
	appsCmd.AddCommand(addAppsCmd)
	// Add the command "add-from-sample" as a subcommand of appsCmd
	appsCmd.AddCommand(addAppsFromSampleCmd)
	// Add the command "update" as a subcommand of appsCmd
	appsCmd.AddCommand(updateAppsCmd)
	// Add the command "del" as a subcommand of appsCmd
	appsCmd.AddCommand(delAppsCmd)
	// Add the command "bookmark" as a subcommand of appsCmd
	appsCmd.AddCommand(bookmarkAppCmd)
	// Add the command "content" as a subcommand of appsCmd
	appsCmd.AddCommand(contentAppsCmd)
	// Add the command "remove" as a subcommand of appsCmd
	appsCmd.AddCommand(removeAppsCmd)
	// Add the command "download" as a subcommand of appsCmd
	appsCmd.AddCommand(downloadAppsCmd)
	// Add the command "upload" as a subcommand of appsCmd
	appsCmd.AddCommand(uploadAppsCmd)

	// Add the command "build" as a subcommand of appsCmd
	appsCmd.AddCommand(buildAppsCmd)
	// Add the command "builds" as a subcommand of appsCmd
	appsCmd.AddCommand(buildsCmd)
	// Add the command "list" as a subcommand of buildsCmd
	buildsCmd.AddCommand(buildsListAppsCmd)
	// Add the command "get" as a subcommand of buildsCmd
	buildsCmd.AddCommand(buildsGetAppsCmd)
	// Add the command "logs" as a subcommand of appsCmd
	buildsCmd.AddCommand(buildsLogsAppsCmd)
	// Add the command "cancel" as a subcommand of appsCmd
	buildsCmd.AddCommand(buildsCancelAppsCmd)
	// Add the command "stats" as a subcommand of appsCmd
	buildsCmd.AddCommand(buildsStatsCmd)

	// Add the command "test" as a subcommand of appsCmd
	appsCmd.AddCommand(testAppsCmd)
	// Add the command "tests" as a subcommand of appsCmd
	appsCmd.AddCommand(testsCmd)
	// Add the command "list" as a subcommand of testsCmd
	testsCmd.AddCommand(testsListAppsCmd)
	// Add the command "get" as a subcommand of testsCmd
	testsCmd.AddCommand(testsGetAppsCmd)
	// Add the command "logs" as a subcommand of testsCmd
	testsCmd.AddCommand(testsLogsAppsCmd)
	// Add the command "cancel" as a subcommand of testsCmd
	testsCmd.AddCommand(testsCancelAppsCmd)
	// Add the command "stats" as a subcommand of testsCmd
	testsCmd.AddCommand(testsStatsCmd)

	// Add the command "services" as a subcommand of appsCmd
	appsCmd.AddCommand(servicesCmd)

	// Add the command "audit" as a subcommand of appsCmd
	appsCmd.AddCommand(auditAppsCmd)
	// Add the command "audits" as a subcommand of appsCmd
	appsCmd.AddCommand(auditsCmd)
	// Add the command "tools" as a subcommand of auditsCmd
	auditsCmd.AddCommand(auditsListToolsCmd)
	// Add the command "list" as a subcommand of auditsCmd
	auditsCmd.AddCommand(auditsListCmd)
	// Add the command "get" as a subcommand of auditsCmd
	auditsCmd.AddCommand(auditsGetCmd)
	// Add the command "logs" as a subcommand of auditsCmd
	auditsCmd.AddCommand(auditsLogsCmd)

	// Add the command "branch" as a subcommand of appsCmd
	appsCmd.AddCommand(branchCmd)
	// Add the command "branches" as a subcommand of appsCmd
	appsCmd.AddCommand(branchesCmd)
	// Add the command "list" as a subcommand of branchesCmd
	branchesCmd.AddCommand(branchesListCmd)
	// Add the command "delete" as a subcommand of branchesCmd
	branchesCmd.AddCommand(branchesDeleteCmd)

	// Add the command "copy" as a subcommand of appsCmd
	appsCmd.AddCommand(copyAppCmd)

	// Add the command "refresh" as a subcommand of appsCmd
	appsCmd.AddCommand(refreshAppCmd)
}

// initEventApp inits everything that is needed for the application add
// command to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventApp(cmd *cobra.Command, args []string) error {

	// Init the events list
	eventsList := []string{
		rpapiv1.EvtAppAdd,
		rpapiv1.EvtAppDelete,
		rpapiv1.EvtAppStateChange,
		rpapiv1.EVTError,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// initEventBuildAndApp inits everything that is needed for the application and
// the build command
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventBuildAndApp(cmd *cobra.Command, args []string) error {

	// First to the initialization for the App
	err := initEventApp(cmd, args)
	if err != nil {
		return err
	}

	// Init the events list
	eventsList := []string{
		rpapiv1.EvtAppBuildInfo,
		rpapiv1.EvtAppBuildSubTaskInfo,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// initEventAuditApp inits everything that is needed for the application audit
// command to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventAuditApp(cmd *cobra.Command, args []string) error {

	// Init the events list
	eventsList := []string{
		rpapiv1.EvtAuditAdd,
		rpapiv1.EvtAuditDelete,
		rpapiv1.EvtAuditStateChange,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getListAppsAsTab gets all the applications available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListAppsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	appsList := []string{}
	typesList := []string{"standard", "released"}

	// Depending if the user is admin, get all types of projects or just "standard" and "released"
	userAdmin, _ := isUserAdminOnCurrentStack()
	if userAdmin {
		typesList = append(typesList, "external")
	}

	// Get applications list
	apps := rpapiv1.RPApplications{}
	if err := bckAPI.ApplicationsListGetOfTypes(&apps, "", typesList); err != nil {
		// No error logs in autocompletion functions!
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the projects list, fill the prjsList tab
	for _, application := range apps.Apps {
		// Add the slug of the application in the suggestions
		appsList = append(appsList, application.Slug)
	}

	return appsList, cobra.ShellCompDirectiveNoFileComp
}

// getFullListAppsAsTab gets all (even external) the applications available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getFullListAppsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	appsList := []string{}
	typesList := []string{"standard", "released", "external"}

	// Get applications list
	apps := rpapiv1.RPApplications{}
	if err := bckAPI.ApplicationsListGetOfTypes(&apps, "", typesList); err != nil {
		// No error logs in autocompletion functions!
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the projects list, fill the prjsList tab
	for _, application := range apps.Apps {
		// Add the slug of the application in the suggestions
		appsList = append(appsList, application.Slug)
	}

	return appsList, cobra.ShellCompDirectiveNoFileComp
}

// getListAppsSamplesAsTab gets all the sample applications available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListAppsSamplesAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	appsList := []string{}

	// Get sample applications list
	apps := []rpapiv1.RPSampleApplication{}
	if err := bckAPI.ApplicationsSamplesListGet(&apps); err != nil {
		// No error logs in autocompletion functions!
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the apps samples list, fill the prjsList tab
	for _, appSample := range apps {
		// Add the package-name of the application in the suggestions
		appsList = append(appsList, appSample.PackageName)
	}

	return appsList, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// listApplications gets all the applications available in the Redpesk
// instance and prints it. If a project is specified, only the applications
// from this project is retrieved.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func listApplications(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the project flag
	projectID, err := cmd.Flags().GetString(appsCmdPrjFlag)
	if err != nil {
		logFlagError(appsCmdPrjFlag)
		return err
	}

	// Get the bookmarked flag
	bookmarkOnly, err := cmd.Flags().GetBool(appsListCmdBookmarkedFlag)
	if err != nil {
		logFlagError(appsListCmdBookmarkedFlag)
		return err
	}

	// Get the type flag
	typesSlice, err := cmd.Flags().GetStringSlice(appsListCmdTypeFlag)
	if err != nil {
		logFlagError(appsListCmdTypeFlag)
		return err
	}

	// Convert "distro" to "external", and remove duplicated values
	utils.ReplaceInStringSlice(&typesSlice, "distro", "external")
	typesSlice = utils.RemoveDuplicateInStrSlice(typesSlice)

	var prjUUIDStr string
	if projectID != "" {
		// Retrieve the true project ID from UUID or its slug
		projectUUID, err := getProjectUUID(projectID)
		if err != nil {
			return err
		}
		prjUUIDStr = projectUUID.String()
	} else {
		prjUUIDStr = ""
	}

	// Prepare the function needed by the printer to get Applications
	getAppsCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		// Get applications list
		apps := rpapiv1.RPApplications{}
		if err := bckAPI.ApplicationsListByTypePage(&apps, prjUUIDStr, typesSlice, pageNb, perPage, bookmarkOnly); err != nil {
			return apps.Apps, printer.RPPrinterPaging{}, err
		}
		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&apps)

		return apps.Apps, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: appDisplayVerboseLightLineNb,
		GetObjCB:      getAppsCB,
		PrintFancyCB:  displayApps,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct)
}

// getApplication gets the application information available in the Redpesk
// instance and prints it
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func getApplication(cmd *cobra.Command, args []string) error {

	// Retrieve the true application ID from UUID or its slug
	appUUID, err := genericGetAppUUID(cmd, args)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the application that we are looking for
	appStructure, err := getAppStructure(cmd, appUUID, true)
	if err != nil {
		return err
	}

	// Display the application obtained
	err = RPPrinter.CustomPrintObject(appStructure, verbose, displayOneAppFull)

	return err
}

// buildsListApplication gets all the builds available for the given application
// and prints it.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildsListApplication(cmd *cobra.Command, args []string) error {
	return buildsList(cmd, args, "application")
}

// buildGetLogsApplication allows to retrieve logs from an application build.
// If the build ID is not specified, the logs of the latest build are retrieved
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildGetLogsApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(genericBuildLogsCmdAnsLine0)

	// Retrieve the build ID from the args given by the user
	buildID, err := getBuildIDFromArgs(cmd, args, appsForceAppFlag, "application", genericGetAppUUID)
	if err != nil {
		return err
	}

	return buildGetLogs(cmd, buildID, "application")
}

// buildCancelApplication allows to cancel a build.
// If the build ID is not specified, the latest build is cancelled
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildCancelApplication(cmd *cobra.Command, args []string) error {

	// Retrieve the build ID from the args given by the user
	buildID, err := getBuildIDFromArgs(cmd, args, appsForceAppFlag, "application", genericGetAppUUID)
	if err != nil {
		return err
	}

	return buildCancel(buildID)
}

// getAppBuildsStats display the builds statistics for the given
// application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getAppBuildsStats(cmd *cobra.Command, args []string) error {
	// Retrieve the true application ID from UUID or its slug
	appUUID, err := genericGetAppUUID(cmd, args)
	if err != nil {
		return err
	}

	// Get the application that we are looking for
	appStructure, err := getAppStructure(cmd, appUUID, false)
	if err != nil {
		return err
	}

	// Now request and display builds stats for this project
	return getAndDisplayBuildsStats(cmd, appStructure.Parent.ProjectID, &appStructure)
}

// getAppTestsStats display the tests statistics for the given
// application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getAppTestsStats(cmd *cobra.Command, args []string) error {
	// Retrieve the true application ID from UUID or its slug
	appUUID, err := genericGetAppUUID(cmd, args)
	if err != nil {
		return err
	}

	// Get the application that we are looking for
	appStructure, err := getAppStructure(cmd, appUUID, false)
	if err != nil {
		return err
	}

	// Now request and display builds stats for this project
	return getAndDisplayTestsStats(cmd, appStructure.Parent.ProjectID, appUUID, uuid.Nil)
}

// uploadFilesApplication uploads files for the given application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func uploadFilesApplication(cmd *cobra.Command, args []string) error {

	// Communication to user
	RPPrinter.CustomPrintLn(appsUploadCmdAnsLine0)

	// Retrieve the true application ID from UUID or its slug
	appUUID, err := genericGetAppUUID(cmd, args)
	if err != nil {
		return err
	}

	// Get the path to the specfile
	files := []string{}
	if len(args) > 1 {
		files = args[1:]
	}

	// Compatibility with old behaviour
	filePath, err := cmd.Flags().GetString(appsUploadCmdFilePathFlag)
	if err == nil && len(filePath) > 0 {
		files = append(files, filePath)
	}

	// Uploading files
	err = nil
	for _, name := range files {
		// Upload the file to the Redpesk instance
		err = uploadFileUtility(name, appUUID)
		if err != nil {
			break
		}
	}
	return err
}

// contentSpecApplication content of the given application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func contentSpecApplication(cmd *cobra.Command, args []string) error {

	// Retrieve the true application ID from UUID or its slug
	appUUID, errargs := genericGetAppUUID(cmd, args)
	if errargs != nil {
		return errargs
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// ask for the content (the list of files)
	content := rpapiv1.RPFileList{}
	errreq := bckAPI.ApplicationGetListFiles(appUUID.String(), &content)
	if errreq != nil {
		return errreq
	}

	// display the content
	errprt := RPPrinter.CustomPrintObject(content, verbose, displayContent)
	return errprt
}

// displayContent display the list of files
//
// :param content: List of services to display
// :type content: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayContent(content interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	list := content.(rpapiv1.RPFileList)

	// Display each service
	for _, item := range list.ListFiles {
		if verbose {
			fmt.Fprintf(writer, "%s\t(%d bytes)\n", item.Name, item.Size)
		} else {
			fmt.Fprintln(writer, item.Name)
		}
	}
	return nil
}

// removeFilesApplication content of the given application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func removeFilesApplication(cmd *cobra.Command, args []string) error {

	// Retrieve the true application ID from UUID or its slug
	appUUID, errargs := genericGetAppUUID(cmd, args)
	if errargs != nil {
		return errargs
	}

	RPPrinter.CustomPrintLn(appsRemoveCmdAnsLine0)
	for _, name := range args[1:] {
		// Communication to user
		RPPrinter.CustomPrint(appsRemoveCmdAnsLine1, name, " ")
		err := bckAPI.ApplicationRemoveFile(appUUID.String(), name)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			Log.Error("error when removing ", name, ": ", err)
		} else {
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		}
	}
	return nil
}

// downloadApplication content of the given application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func downloadFilesApplication(cmd *cobra.Command, args []string) error {

	// Retrieve the true application ID from UUID or its slug
	appUUID, errargs := genericGetAppUUID(cmd, args)
	if errargs != nil {
		return errargs
	}

	files := []string{}
	if len(args) > 1 {
		files = args[1:]
	} else {
		// ask for the content (the list of files)
		content := rpapiv1.RPFileList{}
		errreq := bckAPI.ApplicationGetListFiles(appUUID.String(), &content)
		if errreq != nil {
			return errreq
		}
		for _, item := range content.ListFiles {
			files = append(files, item.Name)
		}
	}

	return downloadGitPkgsAppFiles(appUUID, "", files, true)
}

// prepareAppStruct prepare a application structure to send to the backend
// to create or update an application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param app: Application structure to fill
// :type app: *rpapiv1.RPApplicationFull
// :param parentPrj: ID of the parent project of the application
// :type parentPrj: uuid.UUID
// :param isCreation: Indicate if we are creating the application
// :type isCreation: bool
// :returns: pointer to the RPAPIInterface object
func prepareAppStruct(cmd *cobra.Command, args []string, app *rpapiv1.RPApplicationFull, parentPrj uuid.UUID, isCreation bool) error {

	// Get all the attributes available only during creation
	if isCreation {
		// Get the alias of the application wanted by the user (optional)
		appAlias, err := cmd.Flags().GetString(appsAddCmdAliasFlag)
		if err != nil {
			logFlagError(appsAddCmdAliasFlag)
			return err
		}

		// Get the package name wanted by the user
		appPkgName, err := cmd.Flags().GetString(appsAddCmdPkgFlag)
		if err != nil {
			logFlagError(appsAddCmdPkgFlag)
			return err
		}

		// Get the optional active branch to set
		activeBranch, err := cmd.Flags().GetString(appsAddCmdActiveBranchFlag)
		if err != nil {
			logFlagError(appsAddCmdActiveBranchFlag)
			return err
		}

		// Set the attributes
		app.Slug = appAlias
		app.PackageName = appPkgName
		app.PackageActiveBranch = activeBranch
	}

	// Get the name of the application to create
	appName, err := cmd.Flags().GetString(appsAddCmdNameFlag)
	if err != nil {
		logFlagError(appsAddCmdNameFlag)
		return err
	}

	// Get the description of the application wanted by the user (optional)
	appDesc, err := cmd.Flags().GetString(appsAddCmdDescFlag)
	if err != nil {
		logFlagError(appsAddCmdDescFlag)
		return err
	}

	// Get the source url wanted by the user
	appSrcURL, err := cmd.Flags().GetString(appsAddCmdSrcURLFlag)
	if err != nil {
		logFlagError(appsAddCmdSrcURLFlag)
		return err
	}

	// Get the source revision wanted by the user
	appSrcRev, err := cmd.Flags().GetString(appsAddCmdSrcRevFlag)
	if err != nil {
		logFlagError(appsAddCmdSrcRevFlag)
		return err
	}

	// Get the build type wanted by the user
	buildType, err := cmd.Flags().GetString(appsAddCmdBuildTypeFlag)
	if err != nil {
		logFlagError(appsAddCmdBuildTypeFlag)
		return err
	}

	// Set these attributes now
	if isCreation {
		app.Name = appName
		app.Description = appDesc
		app.Metadata.Branch.Version = appsMetadataVersion
		app.Metadata.Branch.Data.SourcesURI = appSrcURL
		app.Metadata.Branch.Data.Revision = appSrcRev
		app.Metadata.Branch.Data.BuildType = rpapiv1.RPAppBuildMethod(buildType)

	} else {
		if appName != flagUpdateKeepCurrentValue {
			app.Name = appName
		}
		if appDesc != flagUpdateKeepCurrentValue {
			app.Description = appDesc
		}
		if appSrcRev != flagUpdateKeepCurrentValue {
			app.Metadata.Branch.Data.Revision = appSrcRev
		}
		if appSrcURL != flagUpdateKeepCurrentValue {
			app.Metadata.Branch.Data.SourcesURI = appSrcURL
		}
		if buildType != flagUpdateKeepCurrentValue {
			app.Metadata.Branch.Data.BuildType = rpapiv1.RPAppBuildMethod(buildType)
		}
	}

	// Get the auto-test value wanted by the user
	app.AutoTests.Set, err = genericGetBoolFlagValue(cmd, isCreation, app.AutoTests.Set, appsAddCmdAutoTestFlag,
		fmt.Sprintf("%s \"%s\" flag", pbFlagMainPrefix, appsAddCmdAutoTestFlag))
	if err != nil {
		return err
	}

	// Get the auto-audit value wanted by the user
	app.AutoAudits.Set, err = genericGetBoolFlagValue(cmd, isCreation, app.AutoAudits.Set, appsAddCmdAutoAuditFlag,
		fmt.Sprintf("%s \"%s\" flag", pbFlagMainPrefix, appsAddCmdAutoAuditFlag))
	if err != nil {
		return err
	}

	app.AutoRSrpmBuild, err = genericGetBoolFlagValue(cmd, isCreation, app.AutoRSrpmBuild, appsAddCmdAutoSRpmBuildFlag,
		fmt.Sprintf("%s \"%s\" flag", pbFlagMainPrefix, appsAddCmdAutoSRpmBuildFlag))
	if err != nil {
		return err
	}

	// Get the external specfile url wanted by the user
	appExtSpecURL, err := cmd.Flags().GetString(appsAddCmdExtSpecFlag)
	if err != nil {
		logFlagError(appsAddCmdExtSpecFlag)
		return err
	}

	// Get the specfile path in the source wanted by the user
	appSrcSpecPath, err := cmd.Flags().GetString(appsAddCmdSrcSpecFlag)
	if err != nil {
		logFlagError(appsAddCmdSrcSpecFlag)
		return err
	}

	// Manage spec file cases
	// Set as default internal specfile
	if isCreation {
		app.Metadata.Branch.Data.SpecfileType = rpapiv1.RPAppSpecTypeInternal
	}

	// Check if external or relative have been selected
	switch {
	case isCreation && (appExtSpecURL != ""):
		app.Metadata.Branch.Data.SpecfileLocation = appExtSpecURL
		app.Metadata.Branch.Data.SpecfileType = rpapiv1.RPAppSpecTypeExternal
	case isCreation && (appSrcSpecPath != ""):
		app.Metadata.Branch.Data.SpecfileLocation = appSrcSpecPath
		app.Metadata.Branch.Data.SpecfileType = rpapiv1.RPAppSpecTypeRelative
	case !isCreation:
		if appExtSpecURL != flagUpdateKeepCurrentValue {
			app.Metadata.Branch.Data.SpecfileLocation = appExtSpecURL
			app.Metadata.Branch.Data.SpecfileType = rpapiv1.RPAppSpecTypeExternal
		}
		if appSrcSpecPath != flagUpdateKeepCurrentValue {
			app.Metadata.Branch.Data.SpecfileLocation = appSrcSpecPath
			app.Metadata.Branch.Data.SpecfileType = rpapiv1.RPAppSpecTypeRelative
		}
	}

	// Get the services given by the user
	inputServices, err := cmd.Flags().GetStringArray(appsAddCmdServiceFlag)
	if err != nil {
		logFlagError(appsAddCmdServiceFlag)
		return err
	}

	bckServices := bckapi.RPBCKServices{}
	needToModifyAppService := true
	servicesToDisable := []string{}
	if isCreation {
		// Get the list of services available on redpesk factory
		err = bckAPI.RPServicesGet(&bckServices)
		if err != nil {
			return err
		}
	} else {
		// Try to get possible services to disable
		servicesToDisable, err = cmd.Flags().GetStringSlice(appsUpdateCmdDisableServiceFlag)
		if err != nil {
			logFlagError(appsUpdateCmdDisableServiceFlag)
			return err
		}

		// Delete flagUpdateKeepCurrentValue from inputServices
		utils.DeleteStringFromSlice(&inputServices, flagUpdateKeepCurrentValue)

		// Check if user gave service flag
		if len(inputServices) > 0 || len(servicesToDisable) > 0 {
			// Check if the Services structure is nil
			if app.State == rpapiv1.RPAppLockState {
				return errors.New("application currently locked: update impossible")
			}

			// Convert from map[string]interfaces to RPBCKServices
			// decode first map of interface{}, so it can be encode into RPBCKServices
			jsonbody, err := json.Marshal(app.Services)
			if err != nil {
				return err
			}

			// then convert from map[string]interfaces to RPBCKServices
			if err := json.Unmarshal(jsonbody, &bckServices); err != nil {
				return err
			}
		} else {
			needToModifyAppService = false
		}
	}

	if needToModifyAppService {
		// Modify the services according to what the user gave
		newServices, err := decodeServiceFromCmdOptions(inputServices, bckServices)
		if err != nil {
			return err
		}

		// If we are in "update" mode, the user can ask to disable services
		if !isCreation {
			// Loop on services and disable the ones requested by the user
			for _, service := range servicesToDisable {
				if serviceStruct, ok := newServices[service]; ok {
					// Disable the service
					serviceStruct.Used = false
				} else {
					// Print a warning saying that the service does not exist
					Log.Warnf(appUpdateDisableServiceNotExisting, service)
				}
			}
		}

		// Copy the services as a map[string]interface{} in the application
		err = copier.Copy(&app.Services, &newServices)
		if err != nil {
			return err
		}
	}

	// Manage the default audits requested by the user
	defAuditsWanted, err := cmd.Flags().GetStringSlice(appsAddCmdDefAuditFlag)
	if err != nil {
		logFlagError(appsAddCmdDefAuditFlag)
		return err
	}
	// Check if we need to update the value in the application structure
	if isCreation || !utils.ContainsString(defAuditsWanted, flagUpdateKeepCurrentValue) {

		// In the case where the user gave only "" or "reset", gives an empty slice
		if (len(defAuditsWanted) == 1) && (utils.ContainsString(defAuditsWanted, "") || utils.ContainsString(defAuditsWanted, "reset")) {
			defAuditsWanted = []string{}
		}

		// Prepare the structure
		auditsStruct, err := checkAndPrepareToolsRequest(defAuditsWanted)
		if err != nil {
			return err
		}
		// Convert the structure to a string
		bytesAudits, err := json.Marshal(auditsStruct)
		if err != nil {
			return err
		}
		// Save in the app structure
		app.AuditsConf = string(bytesAudits)
	}

	// Get the list of all the mandatory distributions
	appMandDistros, err := cmd.Flags().GetStringSlice(appsAddCmdMandDistFlag)
	if err != nil {
		logFlagError(appsAddCmdMandDistFlag)
		return err
	}
	if !isCreation {
		if utils.ContainsString(appMandDistros, flagUpdateKeepCurrentValue) {
			// Retrieve the distros already existing in application
			appMandDistros = []string{}
			for _, distro := range app.Distributions {
				if distro.Set && distro.Mandatory {
					appMandDistros = append(appMandDistros, distro.Slug)
				}
			}
		} else {
			// Reset the mandatory distro for the apps and let the end of the function do its work
			app.Distributions = []rpapiv1.RPDistro{}
		}
	}

	// Get the list of all the optional distributions
	appOptDistros, err := cmd.Flags().GetStringSlice(appsAddCmdOptDistFlag)
	if err != nil {
		logFlagError(appsAddCmdOptDistFlag)
		return err
	}
	if !isCreation {
		if utils.ContainsString(appOptDistros, flagUpdateKeepCurrentValue) {
			// Retrieve the distros already existing in application
			appOptDistros = []string{}
			for _, distro := range app.Distributions {
				if distro.Set && !distro.Mandatory {
					appOptDistros = append(appOptDistros, distro.Slug)
				}
			}
		} else {
			// Reset the mandatory distro for the apps and let the end of the function do its work
			app.Distributions = []rpapiv1.RPDistro{}
		}
	}

	// Fill the last attributes
	app.AutoTests.ID = "Automatic tests"
	app.AutoAudits.ID = "Automatic audits"
	app.Parent.ProjectID = parentPrj

	// Get the parent project
	FatherPrj := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectGet(parentPrj.String(), &FatherPrj); err != nil {
		return err
	}

	// Fill the distributions infos for the app structure
	RPPrinter.CustomPrint(appsAddCmdAnsLine2)
	err = fillDistInfo(appMandDistros, appOptDistros, &FatherPrj, app)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	return nil
}

// decodeServiceFromCmdOptions decode --service option values
//
// :param inputServices:
// :type inputServices:
// :returns: error -- an error if something went wrong, nil in the other case
func decodeServiceFromCmdOptions(inputServices []string, defaultBckServices bckapi.RPBCKServices) (bckapi.RPBCKServices, error) {

	// Create the resulting map
	res := make(bckapi.RPBCKServices)

	// Copy the default services inside the result
	copier.Copy(&res, &defaultBckServices)

	for _, values := range inputServices {

		// Split the flag value given by the user using "="
		serviceSplitted := strings.Split(values, "=")
		if len(serviceSplitted) != 2 {
			return res, fmt.Errorf("invalid service definition %s", values)
		}

		serviceName := serviceSplitted[0]

		// Replace simple quote by escaped double quotes to be able to unmarshal the json string
		serviceValue := strings.ReplaceAll(serviceSplitted[1], "'", "\"")

		// Convert the json string to a map
		settedParams := make(map[string]interface{})
		err := json.Unmarshal([]byte(serviceValue), &settedParams)
		if err != nil {
			return res, fmt.Errorf("invalid definition services values %s : error = %v", serviceValue, err)
		}

		// Check that the service name passed is existing in the available services
		if bckSrvStruct, ok := defaultBckServices[serviceName]; ok {
			// For each parameter given by the user, take the value and set it in the result map
			for paramName, paramValue := range settedParams {
				// Check that the parameter exists for this service in the available service
				if _, ok := bckSrvStruct.Parameters[paramName]; ok {
					// Modify the value as given by the user
					tmpParam := bckapi.RPBCKServiceParam{
						Type:        res[serviceName].Parameters[paramName].Type,
						Description: res[serviceName].Parameters[paramName].Description,
						Value:       paramValue,
					}
					// Assign new parameter value
					res[serviceName].Parameters[paramName] = tmpParam
					res[serviceName].Used = true
				} else {
					return res, fmt.Errorf("param '%s' not found for service %s", paramName, serviceName)
				}
			}
		} else {
			fmt.Println(bckSrvStruct, "\n", serviceName)
			return res, fmt.Errorf("unsupported service name %s", serviceName)
		}
	}

	return res, nil
}

// createApplication creates an application as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func createApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsAddCmdAnsLine0)

	// Get the name of the project where the application needs to be created
	appPrjName, err := cmd.Flags().GetString(appsCmdPrjFlag)
	if err != nil {
		logFlagError(appsCmdPrjFlag)
		return err
	}
	if appPrjName == "" {
		Log.Error(appPrjFlagNotSetCreateMsg)
		return errors.New(appPrjFlagRequiredCreateMsg)
	}

	// Retrieve the true project ID from UUID or its slug
	projectUUID, err := getProjectUUID(appPrjName)
	if err != nil {
		return err
	}

	// Create the structure to fill
	app := rpapiv1.RPApplicationFull{}

	// Prepare the application structure to send to the backend
	if err = prepareAppStruct(cmd, args, &app, projectUUID, true); err != nil {
		return err
	}

	// Request the creation of the application to the backend
	if err = requestAppCreation(&app, projectUUID); err != nil {
		return err
	}

	// Once that the application has been created, upload the specfile if specified as locally
	appLocalSpecPath, err := cmd.Flags().GetString(appsAddCmdLocalSpecFlag)
	if err != nil {
		logFlagError(appsAddCmdLocalSpecFlag)
		return err
	}
	if appLocalSpecPath != "" {
		// Upload the specfile to the redpesk factory
		RPPrinter.CustomPrintLn()
		RPPrinter.ForceFlush()
		RPPrinter.CustomPrintLn(appsUploadCmdAnsLine0)

		err = uploadFileUtility(appLocalSpecPath, app.ID)
		if err != nil {
			return err
		}
	}

	return nil
}

// createApplicationFromSample creates an application from a sample
// as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func createApplicationFromSample(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsAddCmdAnsLine0)

	// Get the application ID and return an error if there is nothing
	appID, err := getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "application"), fmt.Sprintf(objectIDRequiredMsg, "application"))
	if err != nil {
		return err
	}

	// Get the available samples from the backend
	apps := []rpapiv1.RPSampleApplication{}
	if err := bckAPI.ApplicationsSamplesListGet(&apps); err != nil {
		return err
	}

	// Check that specified app exists in the samples furnished by the backend
	appIndex := -1
	for index, appSample := range apps {
		if appSample.PackageName == appID {
			appIndex = index
			break
		}
	}

	if appIndex == -1 {
		errMsg := fmt.Sprintf(appSampleNotFoundMsg, appID)
		return errors.New(errMsg)
	}

	// Get the name of the project where the application needs to be created
	appPrjName, err := cmd.Flags().GetString(appsCmdPrjFlag)
	if err != nil {
		logFlagError(appsCmdPrjFlag)
		return err
	}
	if appPrjName == "" {
		Log.Error(appPrjFlagNotSetCreateMsg)
		return errors.New(appPrjFlagRequiredCreateMsg)
	}

	// Retrieve the true project ID from UUID or its slug
	projectUUID, err := getProjectUUID(appPrjName)
	if err != nil {
		return err
	}

	// Cast rpapiv1.RPSampleApplication to rpapiv1.RPApplication using json as "interface"
	castedApp := rpapiv1.RPApplicationFull{}
	byteArrayApp, err := json.Marshal(apps[appIndex])
	if err != nil {
		return err
	}
	err = json.Unmarshal(byteArrayApp, &castedApp)
	if err != nil {
		return err
	}

	// Fill the parent project UUID in the structure
	castedApp.Parent.ProjectID = projectUUID

	// Get the parent project
	FatherPrj := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectGet(projectUUID.String(), &FatherPrj); err != nil {
		return err
	}

	// Fill the distributions infos for the app structure
	RPPrinter.CustomPrint(appsAddCmdAnsLine2)
	err = fillDistInfo([]string{}, []string{}, &FatherPrj, &castedApp)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Request the creation of the application from the sample
	if err = requestAppCreation(&castedApp, projectUUID); err != nil {
		return err
	}

	return nil
}

// updateApplication update an application as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func updateApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsUpdateCmdAnsLine0)

	// Retrieve the true application ID from UUID or its slug
	appUUID, err := genericGetAppUUID(cmd, args)
	if err != nil {
		return err
	}

	// Get the application that we are looking for
	app, err := getAppStructure(cmd, appUUID, true)
	if err != nil {
		return err
	}

	// Prepare the application structure to send to the backend
	if err = prepareAppStruct(cmd, args, &app, app.Parent.ProjectID, false); err != nil {
		return err
	}

	// Get the "force-stable-state" flag
	forceStable, err := cmd.Flags().GetBool(appsForceStateFlag)
	if err != nil {
		return err
	}
	// If the user wants it, force the app state at "stable"
	if forceStable {
		app.State = rpapiv1.RPAppStableState
	}

	Log.Infof(appUpdateInfoMsg, app.Name)
	Log.Debugf(appUpdateDebugMsg, app)

	// Tell the user that the application is currently being updated
	RPPrinter.CustomPrintF(appsUpdateCmdAnsLine1, app.Name)

	// Request the update of this application to the backend
	appAns := rpapiv1.RPApplicationFull{}
	if err := bckAPI.ApplicationUpdate(appUUID.String(), &app, &appAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(appAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project addition
		if rEvent.Type == rpapiv1.EvtAppStateChange {
			// Decode the application
			rApp, err := rEvent.DecodeApplication()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			if utils.UUIDEqual(rApp.ID, appAns.ID) && (rApp.State == rpapiv1.RPAppStableState) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rApp, false, displayNothing)
				break
			}
		}

		// Check if we get an error event during the operation
		if rEvent.Type == rpapiv1.EVTError {
			// Function handling the events
			err = handleErrorEvent(appAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// deleteApplication deletes an existing application as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func deleteApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsDelCmdAnsLine0)

	// Retrieve the app UUID and its Slug
	appUUID, appSlug, err := getAppUUIDAndSlug(cmd, args)
	if err != nil {
		return err
	}

	// Tell the user that the application is currently being deleted
	RPPrinter.CustomPrintF(appsDelCmdAnsLine1, appSlug)

	// Ask the backend for deletion
	appAns := rpapiv1.RPApplicationFull{}
	if err := bckAPI.ApplicationDelete(appUUID.String(), &appAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(appAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a application deletion addition
		if rEvent.Type == rpapiv1.EvtAppDelete {
			// Function handling the events
			rApp, err := rEvent.DecodeApplication()
			if err != nil {
				return err
			}

			if utils.UUIDEqual(rApp.ID, appAns.ID) {
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rApp, false, displayNothing)
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				break
			}
		}
	}

	return nil
}

// bookmarkApplication bookmarks an existing application as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func bookmarkApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsBookmarkCmdAnsLine0)

	// Retrieve the app UUID and its Slug
	appUUID, appSlug, err := getAppUUIDAndSlug(cmd, args)
	if err != nil {
		return err
	}

	// Get the reset flag
	reset, err := cmd.Flags().GetBool(appsBookmarkCmdResetFlag)
	if err != nil {
		logFlagError(appsBookmarkCmdResetFlag)
		return err
	}

	// Get the application concerned by the bookmarking
	appStructure, err := getAppStructure(cmd, appUUID, false)
	if err != nil {
		return err
	}

	// Do the appropriate action according to the variables
	resetDone := false
	setDone := false
	bookmarkObj := rpapiv1.RPApplicationBookmark{}
	switch {
	case (reset && !appStructure.Bookmarked):
		// Reset is requested but the app is not bookmarked, nothing to do
		RPPrinter.CustomPrintF(appsBookmarkCmdAnsLine3, appSlug)
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	case (reset && appStructure.Bookmarked):
		// Reset is requested and the app is bookmarked
		RPPrinter.CustomPrintF(appsBookmarkCmdAnsLine2, appSlug)
		if err := bckAPI.ApplicationBookmark(appUUID.String(), &bookmarkObj); err != nil {
			return err
		}
		resetDone = true

	case (!reset && appStructure.Bookmarked):
		// Bookmark is requested but the app is already bookmarked, nothing to do
		RPPrinter.CustomPrintF(appsBookmarkCmdAnsLine4, appSlug)
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	case (!reset && !appStructure.Bookmarked):
		// Bookmark is requested and the app is not bookmarked
		RPPrinter.CustomPrintF(appsBookmarkCmdAnsLine1, appSlug)
		if err := bckAPI.ApplicationBookmark(appUUID.String(), &bookmarkObj); err != nil {
			return err
		}
		setDone = true
	}

	if setDone || resetDone {

		// Display something to the user if in raw mode
		RPPrinter.CustomPrintObject(bookmarkObj, false, displayNothing)

		// Check from viper the non-blocking flag to see if we wait
		nonBlocking := viper.GetBool(nonBlockingFlag)
		if nonBlocking {
			// The user do not want blocking functions
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
			return nil
		}

		// Block until we receive a corresponding websocket event
		for {
			rEvent := <-cmdEventChannel

			// Check that the received event is for a project addition
			if rEvent.Type == rpapiv1.EvtAppStateChange {
				// Function handling the events
				rApp, err := rEvent.DecodeApplication()
				if err != nil {
					return err
				}

				if utils.UUIDEqual(rApp.ID, bookmarkObj.ApplicationID) {
					if rApp.Bookmarked != reset {
						// Display OK if not in raw mode
						RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
						// Display something to the user if in raw mode
						RPPrinter.CustomPrintObject(rApp, false, displayNothing)
						break
					}
				}
			}
		}
	}
	return nil
}

// buildApplication launch a build on the existing application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(fmt.Sprintf(genericBuildCmdAnsLine0, "Application"))

	// Retrieve the app UUID and its Slug
	appUUID, appSlug, err := getAppUUIDAndSlug(cmd, args)
	if err != nil {
		return err
	}

	// Get the additional information (comment) to save with the build
	comment, err := cmd.Flags().GetString(appsBuildCmdCommentFlag)
	if err != nil {
		logFlagError(appsBuildCmdCommentFlag)
		return err
	}

	// Get the value of the srpm flag
	bldType := rpapiv1.RPBuildApp
	isSrpmBld, err := cmd.Flags().GetBool(appsBuildCmdSrpmFlag)
	if err != nil {
		logFlagError(appsBuildCmdSrpmFlag)
		return err
	}

	// Get the value of the check-deps flag
	isCheckDepsBld, err := cmd.Flags().GetBool(appsBuildCmdCheckDepsFlag)
	if err != nil {
		logFlagError(appsBuildCmdCheckDepsFlag)
		return err
	}

	if isSrpmBld {
		bldType = rpapiv1.RPBuildSrpm
	} else if isCheckDepsBld {
		bldType = rpapiv1.RPBuildDeps
	}

	// Manage the distribution to build
	startBuildObj, err := createDistroToBuild(cmd, appUUID)
	if err != nil {
		return err
	}

	// Update the comment field of the start build object
	startBuildObj.Comment = comment

	// Tell the user that we start the build of the application
	RPPrinter.CustomPrintF(genericBuildCmdAnsLine1, appSlug, "application")

	// Ask the backend to start the build
	buildsAns := rpapiv1.RPBuilds{}
	if err := bckAPI.ApplicationBuild(appUUID.String(), startBuildObj, bldType, &buildsAns); err != nil {
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(buildsAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Create a map for all the builds we are monitoring
	buildsMap := make(map[string]string)
	longestDistroSlugValue := 0
	for _, bld := range buildsAns.Builds {
		buildsMap[bld.ID] = "building"
		// Get the longest distribution slug
		if len(bld.DistributionSlug) > longestDistroSlugValue {
			longestDistroSlugValue = len(bld.DistributionSlug)
		}
	}

	// Create the format string with the right length for distribution
	lengthDistroStr := strconv.Itoa(longestDistroSlugValue)
	// Change the format line used in builds.go in getLiveBuildLine()
	liveBuildFormatLine = genericBuildLiveCmdLinePart0 + "%-" + lengthDistroStr + "v" + genericBuildLiveCmdLinePart1

	// Create a tab that will contains the results we want to display at the end
	resultsTab := []string{}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		Log.Tracef("buildApplication: receive event=%+v", rEvent)

		// Check that the received event is for application build
		if rEvent.Type == rpapiv1.EvtAppBuildInfo || rEvent.Type == rpapiv1.EvtAppBuildSubTaskInfo {
			// Function handling the events data
			rBuild, err := rEvent.DecodeBuild()
			if err != nil {
				return err
			}

			// Check if we are monitoring this build ID
			_, ok := buildsMap[rBuild.ID]
			if ok {
				// Print the build in the output
				RPPrinter.CustomPrintObject(rBuild, true, printOneBuildLive)

				// Check if the build is done
				if rBuild.Status == rpapiv1.RPBuildTaskStatusDone {

					// This build is finished and does not need to be monitored anymore
					buildsMap[rBuild.ID] = "done"

					// Add a result line for this build
					resultsTab = append(resultsTab, getLiveBuildLine(rBuild, genericBuildLiveLabelSuccess))

				} else if rBuild.Status == rpapiv1.RPBuildTaskStatusCanceled {
					buildsMap[rBuild.ID] = "canceled"

					// Add a result line for this build
					resultsTab = append(resultsTab, getLiveBuildLine(rBuild, genericBuildLiveLabelCanceled))

					// Display the comment if there is one
					if rBuild.Comment != "" {
						resultsTab = append(resultsTab, fmt.Sprintf("%s Cancellation comment: '%s'", genericBuildLiveLabelCanceled, rBuild.Comment))

						// If it fit the regexp, tell the user to do a 'build get' to have more details about the error/cancellation
						regularExp := regexp.MustCompile(`build id ([0-9]+) status is failed`)
						matchGroups := regularExp.FindStringSubmatch(rBuild.Comment)
						if len(matchGroups) >= 2 {
							resultsTab = append(resultsTab, fmt.Sprintf("%s Please run 'rp-cli applications builds get %s' for more info about cancellation", genericBuildLiveLabelCanceled, matchGroups[1]))
						}
					}
				}

				// Check if the build is failed
				if rBuild.Status == "failed" {
					// This build is finished and does not need to be monitored anymore
					buildsMap[rBuild.ID] = "failed"

					// Add a result line for this build
					resultsTab = append(resultsTab, getLiveBuildLine(rBuild, genericBuildLiveLabelFailed))

					// Save output of error fault code and string if there are some
					if rBuild.Task.Error.FaultString != "" {
						resultsTab = append(resultsTab, fmt.Sprintf("%s %s%d", genericBuildLiveLabelFailed, genericBuildCmdVerboseErrorFaultCode, rBuild.Task.Error.FaultCode))
						resultsTab = append(resultsTab, fmt.Sprintf("%s %s'%s'", genericBuildLiveLabelFailed, genericBuildCmdVerboseErrorFaultMsg, rBuild.Task.Error.FaultString))
					}
				}

				// Check if there are still builds to monitor
				if !utils.MapContainsString(buildsMap, "building") {
					return printLiveBuildResults(resultsTab, appBuildErrorMsg)
				}
			}
		}

		// Check if we get an error event during the operation
		if rEvent.Type == rpapiv1.EVTError {
			// Function handling the events
			err = handleErrorEvent(appUUID, rEvent, false)
			if err != nil {
				return err
			}
		}
	}
}

// appCopyToProject copy the application to another project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func appCopyToProject(cmd *cobra.Command, args []string) error {

	// Retrieve the app UUID and its Slug
	appUUID, appSlug, err := getAppUUIDAndSlug(cmd, args)
	if err != nil {
		return err
	}

	//  Get the name of the project where the application will be copied
	destProject, err := cmd.Flags().GetString(appCopyCmdDestinationPrjFlag)
	if err != nil {
		logFlagError(appCopyCmdDestinationPrjFlag)
		return err
	}

	// Check that the user gave a correct value for the destination project
	if destProject == "" {
		return errors.New(copyAppDestPrjEmptyMsg)
	}
	// Retrieve the true project ID from UUID or its slug
	projectUUID, err := getProjectUUID(destProject)
	if err != nil {
		return err
	}

	// Construct the right structure and send the request to the backend
	rpApplicationCopy := rpapiv1.RPApplicationCopy{Destination: projectUUID}
	if err := bckAPI.ApplicationCopy(appUUID.String(), &rpApplicationCopy); err != nil {
		return err
	}

	// Tell the user that everything went well
	RPPrinter.CustomPrintF(appsCopyCmdAnsLine0, appSlug, destProject)
	return nil
}

// appRefreshGitpkgFiles requests the backend to refresh the gitpkg files for the
// given application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func appRefreshGitpkgFiles(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsRefreshFilesPkgAnsLine0)

	// Retrieve the app UUID and its slug
	appUUID, appSlug, err := getAppUUIDAndSlug(cmd, args)
	if err != nil {
		return err
	}

	RPPrinter.CustomPrintF(appsRefreshFilesPkgAnsLine1, appSlug)

	// Construct the right structure and send the request to the backend
	if err := bckAPI.ApplicationRefreshFiles(appUUID.String()); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project addition
		if rEvent.Type == rpapiv1.EvtAppStateChange {
			// Decode the application
			rApp, err := rEvent.DecodeApplication()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			if utils.UUIDEqual(rApp.ID, appUUID) && (rApp.State == rpapiv1.RPAppStableState) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				break
			}
		}

		// Check if we get an error event during the operation
		if rEvent.Type == rpapiv1.EVTError {
			// Function handling the events
			err = handleErrorEvent(appUUID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getCachedRPApps is a function allowing to optimize the number of
// requests done to the backend. Indeed, it allows to just do the
// "get lists of applications" once to the backend and afterwards,
// to store them locally. The program just has to call this function
// anytime it needs to have the list of applications
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param projectUUID: the UUID of the parent project (optional)
// :type projectUUID: uuid.UUID
// :returns: map[uuid.UUID]rpapiv1.RPApplication, error -- the map of
// the apps and an error if something went wrong, nil in the other case
func getCachedRPApps(cmd *cobra.Command, projectUUID uuid.UUID) (map[uuid.UUID]rpapiv1.RPApplication, error) {

	// Firstly, check if the global variable has been initialized
	if (cachedRPApplication == nil) || (runMode == "testing") {
		// The cache has not been initialized, fill it right now

		// Get the project flag if it exists
		prjUUIDStr := ""
		if !projectUUID.IsNil() {
			prjUUIDStr = projectUUID.String()
		} else {
			appPrjName, err := cmd.Flags().GetString(appsCmdPrjFlag)
			if (err == nil) && (appPrjName != "") {
				// Retrieve the true project ID from UUID or its slug
				projectUUID, err := getProjectUUID(appPrjName)
				if err == nil {
					prjUUIDStr = projectUUID.String()
				}
			}
		}

		// Initialize the type list to request
		typesList := []string{"standard"}

		// Check if the user is an admin or not
		userAdmin, _ := isUserAdminOnCurrentStack()
		if userAdmin {
			typesList = append(typesList, "released", "external")
		} else {
			// Check if it is a command where we can return all the applications
			isOnlyReadCommand := false
			if annotationValue, ok := cmd.Annotations["__read_object_command__"]; ok {
				isOnlyReadCommand, _ = strconv.ParseBool(annotationValue)
			}

			if isOnlyReadCommand {
				typesList = append(typesList, "released", "external")
			}
		}

		// Get the list of the application existing on the Redpesk instance
		apps := rpapiv1.RPApplications{}
		if err := bckAPI.ApplicationsListGetOfTypes(&apps, prjUUIDStr, typesList); err != nil {
			return nil, err
		}

		// Fill the cache with it
		cachedRPApplication = make(map[uuid.UUID]rpapiv1.RPApplication)
		for index, application := range apps.Apps {
			// Copy each application in the cache
			tmpApplication := rpapiv1.RPApplication{}
			copier.Copy(&tmpApplication, &apps.Apps[index])
			cachedRPApplication[application.ID] = tmpApplication
		}
	}

	// Return the cache
	return cachedRPApplication, nil
}

// genericGetAppUUID allows a function to retrieve the UUID of the
// application targetted by the user.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- the UUID of the app and an error if something went wrong, nil in the other case
func genericGetAppUUID(cmd *cobra.Command, args []string) (uuid.UUID, error) {

	// Get the application ID and return an error if there is nothing
	appID, err := getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "application"), fmt.Sprintf(objectIDRequiredMsg, "application"))
	if err != nil {
		return uuid.Nil, err
	}

	// Test if the appID could be a valid UUID
	isValidUUID := utils.IsValidUUID(appID)

	// In the case where the the id could be directly a valid UUID, return it
	if isValidUUID {
		return uuid.FromString(appID)
	}

	// Get the name of the project if the user set it
	appPrjName, err := cmd.Flags().GetString(appsCmdPrjFlag)
	if err != nil {
		logFlagError(appsCmdPrjFlag)
		return uuid.Nil, err
	}

	projectUUID := uuid.Nil
	if appPrjName != "" {
		// Retrieve the true project ID from UUID or its slug
		projectUUID, err = getProjectUUID(appPrjName)
		if err != nil {
			return uuid.Nil, err
		}
	}

	// Retrieve the true application ID from UUID or its slug
	appUUID, err := getAppUUID(cmd, appID, projectUUID)
	if err != nil {
		return uuid.Nil, err
	}

	return appUUID, nil
}

// getAppUUID is a private function allowing to retrieve the
// UUID of an application from a given identifier. This identifier can
// be the UUID itself as a string representation or the Slug of the
// project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param identifier: the identifier of the application
// :type identifier: string
// :param projectUUID: the UUID of the parent project
// :type projectUUID: uuid.UUID
// :returns: (uuid.UUID, error) -- the UUID of the project and an error if something went bad
func getAppUUID(cmd *cobra.Command, identifier string, projectUUID uuid.UUID) (uuid.UUID, error) {
	uuidValue := uuid.Nil
	moreThanOne := false

	// Get the current applications in cache
	apps, err := getCachedRPApps(cmd, projectUUID)
	if err != nil {
		return uuidValue, err
	}

	// Check if the identifier is the UUID
	tmpUUID, err := uuid.FromString(identifier)
	if err == nil {
		if _, ok := apps[tmpUUID]; ok {
			// The identifier is our UUID!
			return tmpUUID, nil
		}
	}

	// The identifier is not the UUID, go through the cached applications
	for currUUID, currApp := range apps {
		if (currApp.Slug == identifier) && utils.UUIDEqual(projectUUID, currApp.Parent.ProjectID) {
			// The slug of the application and the project ID match, it is our app!
			uuidValue = currUUID
			break
		} else if currApp.Slug == identifier && (projectUUID == uuid.Nil) {
			// In this last case, only the slug match. We have to check until
			// the end of the list before to be sure.
			if utils.UUIDEqual(uuid.Nil, uuidValue) {
				uuidValue = currUUID
			} else {
				// uuidValue != uuid.Nil means that there are several applications
				// with the name we want. We have to return an error
				moreThanOne = true
				break
			}
		}
	}

	if utils.UUIDEqual(uuid.Nil, uuidValue) {
		return uuid.Nil, fmt.Errorf(appIdentifierNotFound, identifier)
	}

	if moreThanOne {
		return uuid.Nil, fmt.Errorf(appSeveralIdentifierFound, identifier)
	}

	return uuidValue, nil
}

// getAppStructure gets the application structure for the given ID
// The caller needs to specify if he wants a truly full structure or not
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param appID: ID of the application wanted
// :type appID: uuid.UUID
// :param fullNeeded: true if the caller needs all the fields from the full structure
// :type fullNeeded: bool
// :returns: rpapiv1.RPApplicationFull, error -- the researched application and an error if something went wrong, nil in the other case
func getAppStructure(cmd *cobra.Command, appID uuid.UUID, fullNeeded bool) (rpapiv1.RPApplicationFull, error) {
	app := rpapiv1.RPApplicationFull{}
	var err error

	if fullNeeded {
		// Request the full application to the backend
		err = bckAPI.ApplicationGet(appID.String(), &app)
	} else {
		// Retrieve the cached applications
		cachedApps, err := getCachedRPApps(cmd, uuid.Nil)
		if err != nil {
			return app, err
		}

		// Try to get our light App
		tmpLightApp, ok := cachedApps[appID]
		if !ok {
			return app, fmt.Errorf(appIdentifierNotFound, appID.String())
		}
		app.RPApplication = tmpLightApp
	}

	return app, err
}

// getAppSlugFromUUID allows to get the slug of an application from its UUID
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param appID: ID of the application wanted
// :type appID: uuid.UUID
// :returns: string, error -- the app's slug and an error if something went wrong, nil in the other case
func getAppSlugFromUUID(cmd *cobra.Command, appID uuid.UUID) (string, error) {

	appSlug := ""

	// Get the current applications in cache
	apps, err := getCachedRPApps(cmd, uuid.Nil)
	if err != nil {
		return appSlug, err
	}

	// Check if we find the UUID
	tmpApplication, ok := apps[appID]
	if ok {
		// We found the application!
		return tmpApplication.Slug, nil
	}

	return appSlug, errors.New(appNotFoundMsg)
}

// getAppUUIDAndSlug allows a function to get the UUID and the slug of an app from one
// of the identifiers
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: uuid, string, error -- the UUID of the app, its slug and an error if something went wrong
func getAppUUIDAndSlug(cmd *cobra.Command, args []string) (uuid.UUID, string, error) {

	// Retrieve the true application ID from UUID or its slug
	appUUID, err := genericGetAppUUID(cmd, args)
	if err != nil {
		return uuid.Nil, "", err
	}

	// Get the slug of our application
	appSlug, err := getAppSlugFromUUID(cmd, appUUID)

	return appUUID, appSlug, err
}

// getBoardStruct allows to retrieve the full board structure
// from its name
//
// :param name: Name of the board that interests us
// :type name: string
// :returns: rpapiv1.RPModelBoard, error -- the model board structure and an error if a problem appends
func getBoardStruct(name string) (rpapiv1.RPModelBoard, error) {
	boardResult := rpapiv1.RPModelBoard{
		ID: uuid.Nil,
	}

	// Get the list of every available boards
	boards := []rpapiv1.RPModelBoard{}
	if err := bckAPI.RPBoardModelsGet(&boards, false); err != nil {
		return boardResult, err
	}

	// We got correctly the boards list, find the one that interest us
	for _, board := range boards {
		if board.Name == name {
			boardResult = board
			break
		}
	}

	return boardResult, nil
}

// fillDistInfo can be used to fill the Distributions part of the
// application structure from what the user wants and from the project ID
//
// :param mandDistWanted: tab of string containing the mandatory distributions
// wanted by the user
// :type mandDistWanted: []string
// :param optDistWanted: tab of string containing the optional distributions
// wanted by the user
// :type optDistWanted: []string
// :param prj: pointer to the project structure
// :type prj: *rpapiv1.RPProjectFull
// :param pApp: Pointer to the structure to fill
// :type pApp: *rpapiv1.RPApplicationFull
// :returns: error -- an error if something went wrong, nil in the other case
func fillDistInfo(mandDistWanted []string, optDistWanted []string, prj *rpapiv1.RPProjectFull, pApp *rpapiv1.RPApplicationFull) error {

	atLeastOne := false

	// Convert the tabs of string containing the distributions wanted to tab of uuid.UUID
	mandDistWantedUUID, err := getTabOfDistrosUUID(mandDistWanted)
	if err != nil {
		return err
	}
	optDistWantedUUID, err := getTabOfDistrosUUID(optDistWanted)
	if err != nil {
		return err
	}

	// If the user gave nothing, use the project distributions
	if len(mandDistWanted)+len(optDistWanted) <= 0 {
		Log.Infof(appNoDistByUser, prj.Slug)
		for _, tmpDist := range prj.Distributions {
			if tmpDist.Set {
				pApp.Distributions = append(pApp.Distributions, tmpDist)
				atLeastOne = true
			}
		}
	} else {
		// If the user has specified some distributions, check that they match the project one
		// and if they does, add it to the structure
		for _, tmpDist := range prj.Distributions {

			if tmpDist.Set && tmpDist.Mandatory {
				// Set a log saying that we are adding it anyway
				Log.Infof(appMandDistMsg, tmpDist.ID, prj.Slug)
				pApp.Distributions = append(pApp.Distributions, tmpDist)
				atLeastOne = true

			} else if tmpDist.Set && !tmpDist.Mandatory {
				// Check if the user wants this architecture as optional
				if utils.ContainsString(optDistWantedUUID, tmpDist.ID.String()) {
					pApp.Distributions = append(pApp.Distributions, tmpDist)
					atLeastOne = true
				}
				// Check if the user wants this architecture as mandatory
				if utils.ContainsString(mandDistWantedUUID, tmpDist.ID.String()) {
					tmpDist.Mandatory = true
					pApp.Distributions = append(pApp.Distributions, tmpDist)
					atLeastOne = true
				}
			}
		}
	}

	// Check if we have at least one architecture set
	if !atLeastOne {
		return errors.New(appDistNotSetMsg)
	}

	return nil
}

// fileToBase64Str is a private function allowing to extract the text from a
// file and then to encode this content to base64.
//
// :param filePath: path to the file to encode
// :type filePath: string
// :returns: string, error -- the encoded string and an error if something went bad
func fileToBase64Str(filePath string) (string, error) {
	// First get the content of the file as text
	clearContent, err := os.ReadFile(filePath)
	if err != nil {
		return "", err
	}

	// Now encode the content of the file
	encodedContent := b64.StdEncoding.EncodeToString([]byte(clearContent))

	return encodedContent, nil
}

// uploadFileUtility is a private function allowing to upload a given file to the
// Redpesk factory (for the application given in parameters)
//
// :param filePath: path to the file to upload
// :type filePath: string
// :param appUUID: ID of the application concerned by the file
// :type appUUID: uuid.UUID
// :returns: error -- an error if something went bad
func uploadFileUtility(filePath string, appUUID uuid.UUID) error {
	// Initialize the file structure and the error we will use
	rpFile := rpapiv1.RPFile{}
	var err error

	// Communication to user
	RPPrinter.CustomPrint(appsUploadCmdAnsLine1, filePath, " ")

	// Set the content of the file encoded in base64
	rpFile.Content, err = fileToBase64Str(filePath)
	if err != nil {
		// Communication to user
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Set the filename in the structure
	rpFile.Name = path.Base(filePath)

	// Communication to user
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	RPPrinter.CustomPrint(appsUploadCmdAnsLine2, filePath, " ")

	// Send the request to the backend
	if err := bckAPI.ApplicationUploadFile(appUUID.String(), &rpFile); err != nil {
		// Communication to user
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project addition
		if rEvent.Type == rpapiv1.EvtAppStateChange {
			// Function handling the events
			rApp, err := rEvent.DecodeApplication()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			if utils.UUIDEqual(rApp.ID, appUUID) && (rApp.State == rpapiv1.RPAppStableState) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rApp, false, displayNothing)
				break
			}
		}

		// Check if we get an error event during the operation
		if rEvent.Type == rpapiv1.EVTError {
			// Function handling the events
			err = handleErrorEvent(appUUID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// manageAppCreationEvents is a function allowing to manage the events received during the
// application creation. The boolean returned allows the caller of the function to know when
// all the needed events have been received.
//
// :param event: event received through the websocket
// :type event: rpapiv1.EventMsg
// :param appAnsUUID: UUID of the application that is being created
// :type appAnsUUID: uuid.UUID
// :returns: bool, error -- true if the app is "stable" or "created" and an error if something went wrong
func manageAppCreationEvents(event rpapiv1.EventMsg, appAnsUUID uuid.UUID) (bool, error) {
	// Initialize the result
	result := false

	// Check that the received event is for a application addition
	if event.Type == rpapiv1.EvtAppAdd || event.Type == rpapiv1.EvtAppStateChange {
		// Decode the application
		rApp, err := event.DecodeApplication()
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return result, err
		}

		if utils.UUIDEqual(rApp.ID, appAnsUUID) {
			// The application slug corresponds to the one we have created
			if event.Type == rpapiv1.EvtAppAdd {
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rApp, false, displayNothing)
				// Write that we are waiting for confirmation
				RPPrinter.CustomPrint(appsAddCmdAnsLine4)
			} else if (event.Type == rpapiv1.EvtAppStateChange) && (rApp.State == rpapiv1.RPAppStableState) {
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rApp, false, displayNothing)
				// The application is completely created
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				result = true
			}
		}
	}
	return result, nil
}

// requestAppCreation is a function allowing to request the creation of an application
// to the backend. It manages also all the prints done during the process.
//
// :param application: pointer to the application to create
// :type application: *rpapiv1.RPApplicationFull
// :param projectUUID: UUID of the parent project
// :type projectUUID: uuid.UUID
// :returns: error -- an error if something went wrong
func requestAppCreation(application *rpapiv1.RPApplicationFull, projectUUID uuid.UUID) error {

	// Few logs
	Log.Infof(appCreationInfoMsg, application.Name)
	Log.Debugf(appCreationDebugMsg, application)

	// Tell the user that the project is currently being created
	RPPrinter.CustomPrintF(appsAddCmdAnsLine3, application.Name)

	// Request the creation of this project to the backend
	appAns := rpapiv1.RPApplicationFull{}
	if err := bckAPI.ApplicationCreate(projectUUID.String(), application, &appAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Set the right UUID in our application (for caller)
	application.ID = appAns.ID

	// The creation request has been taken into account by the backend
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(appAns, false, displayNothing)

	// Create the sentence to display when the object will be added
	slugSentence := fmt.Sprintf(objectSlugDisplayMsg, "application", appAns.Slug)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		// Display the slug of the created object
		RPPrinter.CustomPrintLn(slugSentence)
		return nil
	}

	// Block until we receive all the websocket events we need to confirm creation
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a application addition
		isEndOfWait, err := manageAppCreationEvents(rEvent, appAns.ID)
		if err != nil {
			return err
		}

		// If this is the end of waiting time, escape!
		if isEndOfWait {
			// Display the slug of the created object
			RPPrinter.CustomPrintLn(slugSentence)
			break
		}

		// Check if we get an error event during the operation
		if rEvent.Type == rpapiv1.EVTError {
			// Function handling the events
			err = handleErrorEvent(appAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// manageDownloadDir manages the creation of a download directory
// where the user specifies it. In order to do that, it retrieves the
// value of the "--directory" flag to do so.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param subDirName: name of the subdirectory to create
// :type subDirName: string
// :param objectID: ID of the concerned object (will be used as another subdirectory)
// :type objectID: string
// :returns: string, error -- path to the created directory and an error if something went wrong
func manageDownloadDir(cmd *cobra.Command, subDirName, objectID string) (string, error) {
	// Get the directory given by the user
	dir, err := cmd.Flags().GetString(genericBuildLogsDirFlag)
	if err != nil {
		logFlagError(genericBuildLogsDirFlag)
		return "", err
	}

	// Get the absolute path from what the user gave us
	cleanDir, err := utils.SanitizePath(dir)
	if err != nil {
		return "", err
	}

	// Create the full path of our directory
	fullPathDir := path.Join(cleanDir, subDirName, objectID)

	// Create the directory
	err = utils.CreateDirIfNotExists(fullPathDir, true)
	return fullPathDir, err
}

// createDistroToBuild is the function managing the distribution to pass to the backend
// when we request to build an application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param appUUID: Unique identifier of the application to build
// :type appUUID: uuid.UUID
// :returns: *rpapiv1.RPDistrosToBuild, error -- pointer to the distro to build and an error if something went wrong
func createDistroToBuild(cmd *cobra.Command, appUUID uuid.UUID) (*rpapiv1.RPBuildStart, error) {
	distrosResult := &rpapiv1.RPBuildStart{
		Distributions: []rpapiv1.RPDistro{},
		Comment:       "",
	}

	// Get the distributions given by the user
	userDistros, err := cmd.Flags().GetStringSlice(appsBuildCmdDistroFlag)
	if err != nil {
		logFlagError(appsBuildCmdDistroFlag)
		return distrosResult, err
	}
	emptyUserList := (len(userDistros) == 0)

	// Get the distributions available in the project

	// Get application structure (in order to have the project ID)
	appStruct, err := getAppStructure(cmd, appUUID, false)
	if err != nil {
		return distrosResult, err
	}

	// Fill a map containing the distribution slugs as key (easier for next steps)
	distroMap := make(map[string]rpapiv1.RPDistro)
	for _, distro := range appStruct.Distributions {
		if distro.Set {
			distroMap[distro.Slug] = distro
			// If the user did not give any distro set them all
			if emptyUserList {
				userDistros = append(userDistros, distro.Slug)
			}
		}
	}

	// Now compare the available distro with what the user gave us
	for _, usrDistro := range userDistros {
		// Check that the distro is available for this project
		distroVal, ok := distroMap[usrDistro]
		if !ok {
			errMsg := fmt.Sprintf(distroNotAvailableMsg, usrDistro, appStruct.Slug)
			return distrosResult, errors.New(errMsg)
		} else {
			// Add the distribution in the result
			distrosResult.Distributions = append(distrosResult.Distributions, rpapiv1.RPDistro{
				ID:        distroVal.ID,
				Slug:      distroVal.Slug,
				Name:      distroVal.Name,
				Version:   distroVal.Version,
				Set:       distroVal.Set,
				Mandatory: distroVal.Mandatory,
			})
		}
	}

	if len(distrosResult.Distributions) == 0 {
		distrosResult = nil
	}

	return distrosResult, nil
}

// downloadGitPkgsAppFiles is the function allowing to download a list of files from an application gitpkg.
// A download directory can be specified. If the caller wants to print user messages, the 'printMsg' argument
// needs to be set to true
//
// :param appUUID: Unique identifier of the application containing the files
// :type appUUID: uuid.UUID
// :param dlDirectory: Download directory, if it is the current one that is wanted, give empty string
// :type dlDirectory: string
// :param filesList: List of files to download
// :type filesList: []string
// :param printMsg: Set to true to display messaged in the command line during downloads
// :type printMsg: bool
// :returns: error -- an error if something went wrong
func downloadGitPkgsAppFiles(appUUID uuid.UUID, dlDirectory string, filesList []string, printMsg bool) error {

	// Custom function allowing to print OK or KO if needed
	funcPrintOKKOIfNeeded := func(color string, a ...interface{}) {
		if printMsg {
			RPPrinter.CustomColorPrint(color, a...)
		}
	}

	// Print a title message if wanted
	if printMsg {
		RPPrinter.CustomPrintLn(appsDownloadCmdAnsLine0)
	}

	// Now iterate over the list of files
	for _, name := range filesList {

		// Communication to user
		if printMsg {
			RPPrinter.CustomPrint(appsDownloadCmdAnsLine1, name, " ")
		}

		file := rpapiv1.RPFile{}
		errdwn := bckAPI.ApplicationDownloadFile(appUUID.String(), path.Base(name), &file)
		if errdwn != nil {
			funcPrintOKKOIfNeeded(printer.ColorRed, globalKOLabel)
			Log.Error("error when downloading ", name, ": ", errdwn)
		} else {
			content, errdec := b64.StdEncoding.DecodeString(file.Content)
			if errdec != nil {
				funcPrintOKKOIfNeeded(printer.ColorRed, globalKOLabel)
				Log.Error("error when decoding ", name, ": ", errdec)
			} else {
				// Now write the file in local
				fullFilePath := name

				// If a download directory has been set, add it to the filename
				if dlDirectory != "" {
					fullFilePath = path.Join(dlDirectory, fullFilePath)
				}

				// Extract all the parents directories and create them if needed
				parentDirs := path.Dir(fullFilePath)
				if parentDirs != "." {
					err := os.MkdirAll(parentDirs, 0750)
					if err != nil {
						funcPrintOKKOIfNeeded(printer.ColorRed, globalKOLabel)
						Log.Error("error when creating parents directory ", parentDirs, ": ", err.Error())
					}
				}

				errwr := os.WriteFile(fullFilePath, content, 0640)
				if errwr != nil {
					funcPrintOKKOIfNeeded(printer.ColorRed, globalKOLabel)
					Log.Error("error when writing ", name, ": ", errdec)
				} else if printMsg {
					funcPrintOKKOIfNeeded(printer.ColorGreen, globalOKLabel)
				}
			}
		}
	}
	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// displayApps is a private function allowing to display
// a slice of applications in the console
//
// :param applications: tab of applications to display
// :type applications: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayApps(applications interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	apps := applications.([]rpapiv1.RPApplication)

	// Display result
	first := true
	for _, application := range apps {
		if verbose {
			if !first {
				fmt.Fprintln(writer)
			}

			// Print this app
			err := _printOneApp(application, verbose, false, false, writer)
			if err != nil {
				return err
			}

		} else {
			if first {
				manageHeaderPrint(appsListCmdHeader, writer)
			}
			// Print this app
			err := _printOneApp(application, verbose, false, false, writer)
			if err != nil {
				return err
			}
		}
		first = false
	}

	return nil
}

// displayOneApp is a private function allowing to display
// one application in the console
//
// :param application: Application to display
// :type application: interface{}
// :param verbose: Verbose flag passed in the command line
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayOneAppFull(application interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Print the app
	err := _printOneApp(application, verbose, true, true, writer)
	if err != nil {
		return err
	}
	return nil
}

// _printOneApp is a private function allowing to print one application in the console
//
// :param app: application to print (can be rpapiv1.RPApplicationFull or rpapiv1.RPApplication)
// :type app: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isAlone: True if there is only one project to display in the context where this function is called
// :type isAlone: bool
// :param isFull: True if we need to display the Full object, false in the other case
// :type isFull: bool
// :param writer: Writer to used to print the project information
// :type writer: *tabwriter.Writer
func _printOneApp(appInterface interface{}, verbose, isAlone, isFull bool, writer *tabwriter.Writer) error {
	var typeStr string
	app := rpapiv1.RPApplicationFull{}

	// Cast to the full type in all case but use "full" parameters as indicated by the user
	// We have to pass through the Marshal/Unmarshal tricks to not have problem during conversion from
	// RPApplication -> RPApplicationFull at runtime
	dataBytes, _ := json.Marshal(appInterface)
	json.Unmarshal(dataBytes, &app)

	// Create the string for type
	typeToDisplay := app.Parent.Type
	if typeToDisplay == "external" {
		typeToDisplay = "distro (external)"
	}

	if verbose {
		// Create the string containing all the distributions
		distributions := ""
		for _, distro := range app.Distributions {
			if distro.Set {
				if distributions != "" {
					distributions += ", "
				}
				if distro.Mandatory {
					typeStr = " [Mand]"
				} else {
					typeStr = " [Opt]"
				}
				distributions += (distro.Slug + typeStr)
			}
		}

		// Create the string about autotests
		autotests := ""
		if app.AutoTests.Set {
			autotests += "ON"
			if app.AutoTests.Mandatory {
				autotests += " [Mand]"
			} else {
				autotests += " [Opt]"
			}
		} else {
			autotests += "OFF"
		}

		// Create the strings about autoaudits
		autoaudits := ""
		if app.AutoAudits.Set {
			autoaudits = "ON"
		} else {
			autoaudits = "OFF"
		}

		auditSettings := mainNotSetMsg
		if app.AuditsConf != "" {
			auditSettings = app.AuditsConf
		}

		// Auto SRPM rebuild
		autoSRpmRebuild := "OFF"
		if app.AutoRSrpmBuild {
			autoSRpmRebuild = "ON"
		}

		// Create the string that can potentially be empty
		description := app.Description
		if description == "" {
			description = mainNotSetMsg
		}

		// Create the string containing the results
		fmt.Fprintln(writer, appsListCmdVerboseLine0, app.ID.String())
		fmt.Fprintln(writer, appsListCmdVerboseLine1, app.Name)
		fmt.Fprintln(writer, appsListCmdVerboseLine8, app.PackageName)
		fmt.Fprintln(writer, appsListCmdVerboseLine2, app.Slug)
		fmt.Fprintln(writer, appsListCmdVerboseLine15, typeToDisplay)
		fmt.Fprintln(writer, appsListCmdVerboseLine3, app.Parent.ProjectSlug)
		fmt.Fprintln(writer, appsListCmdVerboseLine9, description)
		fmt.Fprintln(writer, appsListCmdVerboseLine6, distributions)
		fmt.Fprintln(writer, appsListCmdVerboseLine13, autotests)
		fmt.Fprintln(writer, appsListCmdVerboseLine19, autoaudits)
		fmt.Fprintln(writer, appsListCmdVerboseLine20, autoSRpmRebuild)
		fmt.Fprintln(writer, appsListCmdVerboseLine21, auditSettings)
		fmt.Fprintln(writer, appsListCmdVerboseLine4, app.CreatedAt.String())
		fmt.Fprintln(writer, appsListCmdVerboseLine5, app.UpdatedAt.String())
		fmt.Fprintln(writer, appsListCmdVerboseLine14, app.Bookmarked)

		if isFull {
			// Print the metadata relative to this app
			extSpec := mainNotSetMsg
			if app.Metadata.Branch.Data.SpecfileLocation != "" {
				extSpec = app.Metadata.Branch.Data.SpecfileLocation
			}
			fmt.Fprintln(writer, appsListCmdVerboseLine18, app.Metadata.Branch.Data.BuildType)
			fmt.Fprintln(writer, appsListCmdVerboseLine10, app.Metadata.Branch.Data.SourcesURI)
			fmt.Fprintln(writer, appsListCmdVerboseLine11, app.Metadata.Branch.Data.Revision)
			fmt.Fprintln(writer, appsListCmdVerboseLine12, extSpec)

			// Print eventual overrides
			if len(app.Overrides) > 0 {
				fmt.Fprintln(writer, appsListCmdVerboseLine16, app.Overrides)
			}

			fmt.Fprintln(writer, appsListCmdVerboseLine17)
			fmt.Fprintln(writer)

			// Convert from map[string]interfaces to RPBCKServices
			// decode first map of interface{}, so it can be encode into RPBCKServices
			jsonbody, err := json.Marshal(app.Services)
			if err != nil {
				return err
			}

			// then convert from map[string]interfaces to RPBCKServices
			services := bckapi.RPBCKServices{}
			if err := json.Unmarshal(jsonbody, &services); err != nil {
				return err
			}

			// Display the services
			if err = displayAppServices(services, verbose, writer); err != nil {
				return err
			}

		}

	} else {
		if isAlone {
			fmt.Fprintln(writer, appsListCmdHeader)
		}
		fmt.Fprintln(writer, app.ID.String(), "\t", app.Name, "\t", app.Slug, "\t", typeToDisplay, "\t", app.Parent.ProjectSlug)
	}

	return nil
}

// displayAppServices display the list of the current services available
// for the specified application
//
// :param services: List of services to display
// :type services: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayAppServices(services interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	servicesToPrint := services.(bckapi.RPBCKServices)

	// Display each service
	for serviceName, serviceStruct := range servicesToPrint {
		// Do not display a service that is not used
		if !serviceStruct.Used {
			continue
		}

		fmt.Fprintln(writer, "\t", "-", appsServicesAnsLine1, serviceName)
		fmt.Fprintln(writer, "\t", "-", appsServicesAnsLine2)
		for paramName, param := range serviceStruct.Parameters {
			paramFullString := "    " + paramName + ":\t "

			paramValueAsStr := fmt.Sprintf("%v", param.Value)
			if paramValueAsStr == "" {
				paramValueAsStr = "\"\""
			}

			paramFullString += paramValueAsStr
			fmt.Fprintln(writer, "\t"+paramFullString)
		}
		fmt.Fprintln(writer)
	}

	return nil
}

// rpServices gets the list of all the available services on
// the Redpesk backend
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpServices(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the list of every available services
	services := bckapi.RPBCKServices{}
	if err = bckAPI.RPServicesGet(&services); err != nil {
		return err
	}

	// Here, for all services set "Used" to true in order to display it with "printServices()" function
	for _, service := range services {
		service.Used = true
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(services, verbose, printServices)

	return err
}

// printServices display the list of the current services available
// in the current Redpesk instance
//
// :param services: List of services to display
// :type services: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printServices(services interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	servicesToPrint := services.(bckapi.RPBCKServices)

	if verbose {
		// Display each service
		fmt.Fprintln(writer, appsServicesAnsLine0)
		fmt.Fprintln(writer)
		for serviceName, serviceStruct := range servicesToPrint {
			// Do not display anything if the service is not used
			if !serviceStruct.Used {
				continue
			}

			fmt.Fprintln(writer, appsServicesAnsLine1, serviceName)
			fmt.Fprintln(writer, appsServicesAnsLine4, serviceStruct.Version)
			fmt.Fprintln(writer, appsServicesAnsLine3, serviceStruct.Description)
			fmt.Fprintln(writer, appsServicesAnsLine2)
			for paramName, param := range serviceStruct.Parameters {
				paramFullString := "  " + paramName + "\t [" + param.Type + "]" + "\t => " + param.Description
				fmt.Fprintln(writer, paramFullString)
			}
			fmt.Fprintln(writer)
		}
	} else {
		// Print the header for services
		fmt.Fprintln(writer, appsServicesAnsHeader)
		// Print each service
		for serviceName, serviceStruct := range servicesToPrint {

			// Do not display anything if the service is not used
			if !serviceStruct.Used {
				continue
			}

			// Aggregate all the names of parameters (max 3)
			parametersSlice := []string{}
			index := 0
			for paramName := range serviceStruct.Parameters {
				index++
				if index > 3 {
					break
				}
				parametersSlice = append(parametersSlice, paramName)
			}
			if index > 3 {
				parametersSlice = append(parametersSlice, "etc.")
			}
			parametersStr := strings.Join(parametersSlice, ", ")

			fmt.Fprintln(writer, serviceName, "\t", serviceStruct.Summary, "\t", serviceStruct.Version, "\t", parametersStr)
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Banano's Touch -----------------------------------------------
// --------------------------------------------------------------

// seriousCmdHandler is a serious command
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func seriousCmdHandler(cmd *cobra.Command, args []string) error {

	// Get the flag
	why, _ := cmd.Flags().GetBool("why")

	if !why {
		RPPrinter.CustomPrintLn("Yes I am. But the real question is WHY are you so serious?")
	} else {
		// Split the string by line
		toDisplay := strings.Split(seriousASCIIArt, "\n")
		for _, line := range toDisplay {
			RPPrinter.CustomPrintLn(line)
			if runMode == "testing" {
				time.Sleep(2 * time.Millisecond)
			} else {
				time.Sleep(250 * time.Millisecond)
			}
			RPPrinter.ForceFlush()
		}
		RPPrinter.CustomPrintLn()
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
