/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"encoding/json"
	"errors"
	"fmt"
	"os"
	"strings"
	"text/tabwriter"
	"time"

	"github.com/gofrs/uuid"
	"github.com/jinzhu/copier"
	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
)

// -------------------------------------------------
// Types definition --------------------------------
// -------------------------------------------------

// RPProjectArchDist definition
type RPProjectArchDist struct {
	MandDistros []string
	OptDistros  []string
	MandArches  []string
	OptArches   []string
}

// -------------------------------------------------
// -------------------------------------------------

// -------------------------------------------------
// Global variable (inside this package) -----------
// -------------------------------------------------

// Projects cached
var cachedRPProjectsByUUID map[uuid.UUID]rpapiv1.RPProject
var cachedRPProjectsBySlug map[string]uuid.UUID

// Release projects cached
var cachedRPRelProjectsByUUID map[uuid.UUID]rpapiv1.RPReleaseProject
var cachedRPRelProjectsBySlug map[string]uuid.UUID

// Roles available for a project in RP instance
var cachedRPPrjRoles []rpapiv1.RPRole

// -------------------------------------------------
// -------------------------------------------------

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

func initPrjCmd(rootCmd *cobra.Command) {

	// projectsCmd represents the projects command
	var projectsCmd = &cobra.Command{
		Use:     prjsCommand,
		Aliases: []string{prjsCommandAlias},
		Short:   prjsCommandShortUsage,
	}

	// -------------------------------------------------
	// Project sub-command -----------------------------
	// -------------------------------------------------

	// listPrjCmd represents the "projects list" command
	var listPrjCmd = &cobra.Command{
		Use:     prjsListCmd,
		Aliases: []string{prjsListCmdAlias},
		Short:   prjsListCmdShortUsage,
		RunE:    listProjects,
	}

	// getPrjCmd represents the "projects get" command
	var getPrjCmd = &cobra.Command{
		Use:               prjsGetCmd,
		Short:             prjsGetCmdShortUsage,
		RunE:              getProject,
		ValidArgsFunction: getFullListPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// addPrjCmd represents the "projects add" command
	var addPrjCmd = &cobra.Command{
		Use:     prjsAddCmd,
		Aliases: []string{prjsAddAlias},
		Short:   prjsAddCmdShortUsage,
		PreRunE: initEventProject,
		RunE:    createProject,
		Annotations: map[string]string{
			"__flag_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":        "TRUE"},
	}

	// updatePrjCmd represents the "projects update" command
	var updatePrjCmd = &cobra.Command{
		Use:               prjsUpdateCmd,
		Aliases:           []string{prjsUpdateAlias},
		Short:             prjsUpdateCmdShortUsage,
		Long:              prjsUpdateCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventProject,
		RunE:              updateProject,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// delPrjCmd represents the "projects delete" command
	var delPrjCmd = &cobra.Command{
		Use:               prjsDelCmd,
		Aliases:           []string{prjsDelAlias},
		Short:             prjsDelCmdShortUsage,
		PreRunE:           initEventProject,
		RunE:              deleteProject,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// bookmarkPrjCmd represents the "projects bookmark" command
	var bookmarkPrjCmd = &cobra.Command{
		Use:               prjsBookmarkCmd,
		Aliases:           []string{prjsBookmarkAlias},
		Short:             prjsBookmarkCmdShortUsage,
		PreRunE:           initEventProject,
		RunE:              bookmarkProject,
		ValidArgsFunction: getFullListPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// memberCmd represents the "projects members" commands group
	var membersCmd = &cobra.Command{
		Use:          prjMembersCmd,
		Short:        prjMembersCmdShortUsage,
		SilenceUsage: true,
	}

	// exportCmd represents the "projects export" commands group
	var exportCmd = &cobra.Command{
		Use:          prjsExportCmd,
		Short:        prjsExportCmdShortUsage,
		SilenceUsage: true,
	}

	// importCmd represents the "projects import" commands group
	var importCmd = &cobra.Command{
		Use:          prjsImportCmd,
		Short:        prjsImportCmdShortUsage,
		SilenceUsage: true,
	}

	// releaseCmd represents the "projects release" command
	var releaseCmd = &cobra.Command{
		Use:               prjsReleaseCmd,
		Aliases:           []string{prjsReleaseCmdAlias},
		Short:             prjsReleaseShortUsage,
		Long:              prjsReleaseLongUsage,
		PreRunE:           initEventProjectRelease,
		RunE:              releaseProject,
		ValidArgsFunction: getListStandardPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// releasesCmd represents the "projects releases" commands group
	var releasesCmd = &cobra.Command{
		Use:          prjsReleasesCmd,
		Short:        prjsReleasesCmdShortUsage,
		SilenceUsage: true,
	}

	// buildsCmd represents the "projects builds" commands group
	var buildsCmd = &cobra.Command{
		Use:          prjsBuildsCmd,
		Short:        prjsBuildsCmdShortUsage,
		SilenceUsage: true,
	}

	// testsCmd represents the "projects tests" commands group
	var testsCmd = &cobra.Command{
		Use:          prjsTestsCmd,
		Short:        prjsTestsCmdShortUsage,
		SilenceUsage: true,
	}

	// -------------------------------------------------
	// -------------------------------------------------

	// -------------------------------------------------
	// Project's members sub-commands ------------------
	// -------------------------------------------------

	// getMembersCmd represents the "projects members get" command
	var getMembersCmd = &cobra.Command{
		Use:               prjMembersGetCmd,
		Short:             prjMembersGetCmdShortUsage,
		SilenceUsage:      true,
		RunE:              getProjectMembers,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// addMembersCmd represents the "projects members add" command
	var addMembersCmd = &cobra.Command{
		Use:               prjMembersAddCmd,
		Short:             prjMembersAddCmdShortUsage,
		Long:              prjMembersAddCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventProjectMembers,
		RunE:              addProjectMembers,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// updateMembersCmd represents the "projects members update" command
	var updateMembersCmd = &cobra.Command{
		Use:               prjMembersUpdateCmd,
		Short:             prjMembersUpdateCmdShortUsage,
		Long:              prjMembersUpdateCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventProjectMembers,
		RunE:              updateProjectMembers,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// deleteMembersCmd represents the "projects members delete" command
	var deleteMembersCmd = &cobra.Command{
		Use:               prjMembersDeleteCmd,
		Short:             prjMembersDeleteCmdShortUsage,
		Long:              prjMembersDeleteCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventProjectMembers,
		RunE:              deleteProjectMembers,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// -------------------------------------------------
	// -------------------------------------------------

	// -------------------------------------------------
	// Project's export sub-commands -------------------
	// -------------------------------------------------

	// startExportCmd represents the "projects export start" command
	var startExportCmd = &cobra.Command{
		Use:               prjsExportStartCmd,
		Short:             prjsExportStartCmdShortUsage,
		Long:              prjsExportStartCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventProjectExport,
		RunE:              startExportProject,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// getExportCmd represents the "projects export get" command
	var getExportCmd = &cobra.Command{
		Use:               prjsExportImportGetCmd,
		Short:             prjsExportGetCmdShortUsage,
		SilenceUsage:      true,
		RunE:              getExportProject,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// downloadExportCmd represents the "projects export download" command
	var downloadExportCmd = &cobra.Command{
		Use:               prjsExportDownloadCmd,
		Short:             prjsExportDownloadCmdShortUsage,
		SilenceUsage:      true,
		RunE:              downloadExportProject,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// -------------------------------------------------
	// -------------------------------------------------

	// -------------------------------------------------
	// Project's import sub-commands -------------------
	// -------------------------------------------------

	// startImportCmd represents the "projects import start" command
	var startImportCmd = &cobra.Command{
		Use:          prjsImportStartCmd,
		Short:        prjsImportStartCmdShortUsage,
		Long:         prjsImportStartCmdLongUsage,
		SilenceUsage: true,
		PreRunE:      initEventProjectImport,
		RunE:         startImportProject,
		Annotations: map[string]string{
			"__bck_ws_connect__": "TRUE"},
	}

	// getImportCmd represents the "projects import get" command
	var getImportCmd = &cobra.Command{
		Use:               prjsExportImportGetCmd,
		Short:             prjsImportGetCmdShortUsage,
		SilenceUsage:      true,
		RunE:              getImportProject,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// -------------------------------------------------
	// -------------------------------------------------

	// -------------------------------------------------
	// Project's releases sub-commands -----------------
	// -------------------------------------------------

	// listReleaseCmd represents the "projects releases list" command
	var listReleaseCmd = &cobra.Command{
		Use:               prjsListCmd,
		Aliases:           []string{prjsListCmdAlias},
		Short:             prjsListReleasedCmdShortUsage,
		SilenceUsage:      true,
		RunE:              listReleasedProjects,
		ValidArgsFunction: getListPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// getReleaseCmd represents the "projects releases get" command
	var getReleaseCmd = &cobra.Command{
		Use:               prjsGetCmd,
		Short:             prjsGetReleasedCmdShortUsage,
		SilenceUsage:      true,
		RunE:              getReleasedProject,
		ValidArgsFunction: getListReleasedPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// updateReleaseCmd represents the "projects releases update" command
	var updateReleaseCmd = &cobra.Command{
		Use:               prjsUpdateCmd,
		Aliases:           []string{prjsUpdateAlias},
		Short:             prjsUpdateReleasedCmdShortUsage,
		SilenceUsage:      true,
		RunE:              updateReleasedProject,
		ValidArgsFunction: getListReleasedPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// deleteReleaseCmd represents the "projects releases delete" command
	var deleteReleaseCmd = &cobra.Command{
		Use:               prjsDelCmd,
		Aliases:           []string{prjsDelAlias},
		Short:             prjsDeleteReleasedCmdShortUsage,
		SilenceUsage:      true,
		RunE:              deleteReleasedProject,
		ValidArgsFunction: getListReleasedPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// deployReleaseCmd represents the "projects releases deploy" command
	var deployReleaseCmd = &cobra.Command{
		Use:               prjsDeployReleaseCmd,
		Aliases:           []string{prjsDeployReleaseCmdAlias},
		Short:             prjsDeployReleaseShortUsage,
		Long:              prjsDeployReleaseLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventProjectDeploy,
		RunE:              deployReleasedProject,
		ValidArgsFunction: getListReleasedPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}
	// -------------------------------------------------
	// -------------------------------------------------

	// -------------------------------------------------
	// Project's builds sub-commands -------------------
	// -------------------------------------------------

	// getBuildsStatsCmd represents the "projects builds stats" command
	var getBuildsStatsCmd = &cobra.Command{
		Use:               prjsSubBuildsStatsCmd,
		Aliases:           []string{prjsSubBuildsStatsAlias},
		Short:             prjsSubBuildsStatsCmdShortUsage,
		SilenceUsage:      true,
		RunE:              getProjectBuildsStats,
		ValidArgsFunction: getListStandardPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// getBuildsMatrixCmd represents the "projects builds matrix" command
	var getBuildsMatrixCmd = &cobra.Command{
		Use:               prjsSubBuildsMatrixCmd,
		Aliases:           []string{prjsSubBuildsMatrixAlias},
		Short:             prjsSubBuildsMatrixCmdShortUsage,
		SilenceUsage:      true,
		RunE:              getBuildsMatrixCmd,
		ValidArgsFunction: getListStandardPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// -------------------------------------------------
	// -------------------------------------------------

	// -------------------------------------------------
	// Project's tests sub-commands --------------------
	// -------------------------------------------------

	// getBuildsStatsCmd represents the "projects builds stats" command
	var getTestsStatsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestsStatsCmd, "prj-ID"),
		Aliases:           []string{genericTestsStatsAlias},
		Short:             fmt.Sprintf(genericTestsStatsCmdShortUsage, "project"),
		SilenceUsage:      true,
		RunE:              getProjectTestsStats,
		ValidArgsFunction: getListStandardPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// -------------------------------------------------
	// -------------------------------------------------

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	listPrjCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	listPrjCmd.Flags().BoolP(prjsListCmdBookmarkedFlag, prjsListCmdBookmarkedFlagAlias, prjsListCmdBookmarkedFlagDefault, prjsListCmdBookmarkedFlagUsage)
	getPrjCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	getPrjCmd.Flags().BoolP(prjsGetCmdRepoFlag, prjsGetCmdRepoFlagAlias, false, prjsGetCmdRepoFlagUsage)
	getPrjCmd.Flags().Bool(prjsGetCmdRoleFlag, false, prjsGetCmdRoleFlagUsage)
	bookmarkPrjCmd.Flags().BoolP(prjsBookmarkCmdResetFlag, prjsBookmarkCmdResetFlagAlias, false, prjsBookmarkCmdResetFlagUsage)
	// Define the flag specific to "members get" command
	getMembersCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Slice flags needed for project list
	listPrjCmd.Flags().StringSlice(prjsListCmdTypeFlag, []string{"standard"}, prjsListCmdTypeFlagUsage)

	// Slice flags needed for project creation and update
	// Creation
	addPrjCmd.Flags().StringSlice(prjsAddCmdMandArchFlag, []string{}, prjsAddCmdMandArchFlagUsage)
	addPrjCmd.Flags().StringSlice(prjsAddCmdOptArchFlag, []string{}, prjsAddCmdOptArchFlagUsage)
	addPrjCmd.Flags().StringSlice(prjsAddCmdMandDistFlag, []string{}, prjsAddCmdMandDistFlagUsage)
	addPrjCmd.Flags().StringSlice(prjsAddCmdOptDistFlag, []string{}, prjsAddCmdOptDistFlagUsage)
	addPrjCmd.Flags().StringSliceP(prjsAddCmdInheritFlag, prjsAddCmdInheritFlagAlias, []string{}, prjsAddCmdInheritFlagUsage)
	// Update
	updatePrjCmd.Flags().StringSlice(prjsAddCmdMandArchFlag, []string{flagUpdateKeepCurrentValue}, prjsAddCmdMandArchFlagUsage)
	updatePrjCmd.Flags().StringSlice(prjsAddCmdOptArchFlag, []string{flagUpdateKeepCurrentValue}, prjsAddCmdOptArchFlagUsage)
	updatePrjCmd.Flags().StringSlice(prjsAddCmdMandDistFlag, []string{flagUpdateKeepCurrentValue}, prjsAddCmdMandDistFlagUsage)
	updatePrjCmd.Flags().StringSlice(prjsAddCmdOptDistFlag, []string{flagUpdateKeepCurrentValue}, prjsAddCmdOptDistFlagUsage)
	updatePrjCmd.Flags().StringSliceP(prjsAddCmdInheritFlag, prjsAddCmdInheritFlagAlias, []string{flagUpdateKeepCurrentValue}, prjsAddCmdInheritFlagUsage)

	// Other flags needed for project creation and update
	// Creation
	addPrjCmd.Flags().StringP(prjsAddCmdNameFlag, prjsAddCmdNameFlagAlias, "", prjsAddCmdNameFlagUsage)
	addPrjCmd.Flags().StringP(prjsAddCmdAliasFlag, prjsAddCmdAliasFlagAlias, "", prjsAddCmdAliasFlagUsage)
	addPrjCmd.Flags().Bool(prjsAddCmdShortReleaseFlag, prjsAddCmdShortReleaseFlagDefault, prjsAddCmdShortReleaseFlagUsage)
	addPrjCmd.Flags().StringP(prjsAddCmdDescFlag, prjsAddCmdDescFlagAlias, "", prjsAddCmdDescFlagUsage)
	addPrjCmd.Flags().Bool(prjsAddCmdAppSupportFlag, prjsAddCmdAppSupportFlagDefault, prjsAddCmdAppSupportFlagUsage)
	addPrjCmd.Flags().Bool(prjsAddCmdImgSupportFlag, prjsAddCmdImgSupportFlagDefault, prjsAddCmdImgSupportFlagUsage)
	addPrjCmd.Flags().Bool(prjsAddCmdCtnSupportFlag, prjsAddCmdCtnSupportFlagDefault, prjsAddCmdCtnSupportFlagUsage)

	// Update
	updatePrjCmd.Flags().StringP(prjsAddCmdNameFlag, prjsAddCmdNameFlagAlias, flagUpdateKeepCurrentValue, prjsAddCmdNameFlagUsage)
	updatePrjCmd.Flags().StringP(prjsAddCmdDescFlag, prjsAddCmdDescFlagAlias, flagUpdateKeepCurrentValue, prjsAddCmdDescFlagUsage)
	updatePrjCmd.Flags().String(prjsAddCmdShortReleaseFlag, flagUpdateKeepCurrentValue, prjsAddCmdShortReleaseFlagUsage)
	updatePrjCmd.Flags().StringP(prjsUpdateCmdBldTypeFlag, prjsUpdateCmdBldTypeFlagAlias, flagUpdateKeepCurrentValue, prjsUpdateCmdBldTypeFlagUsage)
	updatePrjCmd.Flags().StringSlice(prjsUpdateCmdBldTypeAppsFlag, []string{}, prjsUpdateCmdBldTypeAppsFlagUsage)
	updatePrjCmd.Flags().String(prjsAddCmdAppSupportFlag, flagUpdateKeepCurrentValue, prjsAddCmdAppSupportFlagUsage)
	updatePrjCmd.Flags().String(prjsAddCmdImgSupportFlag, flagUpdateKeepCurrentValue, prjsAddCmdImgSupportFlagUsage)
	updatePrjCmd.Flags().String(prjsAddCmdCtnSupportFlag, flagUpdateKeepCurrentValue, prjsAddCmdCtnSupportFlagUsage)
	updatePrjCmd.Flags().Bool(prjsForceStateFlag, false, prjsForceStateFlagUsage)

	// Flags needed by the "projects members add" command
	addMembersCmd.Flags().StringP(prjsMembersAddCmdGuestFlag, prjsMembersAddCmdGuestAlias, "", prjsMembersAddCmdGuestUsage)
	addMembersCmd.Flags().StringSliceP(prjsMembersAddCmdRoleFlag, prjsMembersAddCmdRoleAlias, []string{}, prjsMembersAddCmdRoleUsage)
	// Flags needed by the "projects members update" command
	updateMembersCmd.Flags().StringP(prjsMembersUpdateCmdMemberFlag, prjsMembersUpdateCmdMemberAlias, "", prjsMembersUpdateCmdMemberUsage)
	updateMembersCmd.Flags().StringSliceP(prjsMembersAddCmdRoleFlag, prjsMembersAddCmdRoleAlias, []string{}, prjsMembersUpdateCmdRoleUsage)
	// Flags needed by the "projects members delete" command
	deleteMembersCmd.Flags().StringP(prjsMembersUpdateCmdMemberFlag, prjsMembersUpdateCmdMemberAlias, "", prjsMembersDeleteCmdMemberUsage)

	// Flags needed by the "projects export start" command
	startExportCmd.Flags().StringP(prjsExportCmdDirFlag, prjsExportCmdDirFlagAlias, prjsExportCmdDirFlagDefault, prjsExportCmdDirFlagUsage)
	// Flags needed by the "projects export get" command
	getExportCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Flags needed by the "projects export download" command
	downloadExportCmd.Flags().StringP(prjsExportCmdDirFlag, prjsExportCmdDirFlagAlias, prjsExportCmdDirFlagDefault, prjsExportCmdDirFlagUsage)

	// Flags needed by the "projects import start" command
	startImportCmd.Flags().StringP(prjsImportCmdFileFlag, prjsImportCmdFileFlagAlias, "", prjsImportCmdFileFlagUsage)
	startImportCmd.Flags().BoolP(prjsImportCmdInteractiveFlag, prjsImportCmdInteractiveFlagAlias, false, prjsImportCmdInteractiveFlagUsage)
	// Flags needed by the "projects import get" command
	getImportCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flags needed by the "projects release" command
	releaseCmd.Flags().StringP(prjsReleaseCmdNameFlag, prjsReleaseCmdNameFlagAlias, "", prjsReleaseCmdNameFlagUsage)
	releaseCmd.Flags().StringP(prjsAddCmdDescFlag, prjsAddCmdDescFlagAlias, "", prjsReleaseCmdDescFlagUsage)
	releaseCmd.Flags().String(prjsReleaseCmdDistroFlag, "", prjsReleaseCmdDistroFlagUsage)
	releaseCmd.Flags().StringP(prjsReleaseCmdVersionFlag, prjsReleaseCmdVersionFlagAlias, "", prjsReleaseCmdVersionFlagUsage)
	releaseCmd.Flags().StringSlice(prjsReleaseCmdTeamFlag, []string{}, prjsReleaseCmdTeamFlagUsage)
	releaseCmd.Flags().StringSliceP(prjsReleaseCmdUserFlag, prjsReleaseCmdUserFlagAlias, []string{}, prjsReleaseCmdUserFlagUsage)
	releaseCmd.Flags().BoolP(prjsReleaseCmdAllUsersFlag, prjsReleaseCmdAllUsersFlagAlias, false, prjsReleaseCmdAllUsersFlagUsage)

	// Flags needed by the "projects releases list" command
	listReleaseCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flags needed by the "projects releases get" command
	getReleaseCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	getReleaseCmd.Flags().BoolP(prjsGetReleaseCmdAudienceFlag, prjsGetReleaseCmdAudienceFlagAlias, false, prjsGetReleaseCmdAudienceFlagUsage)

	// Flags needed by the "projects releases update" command
	updateReleaseCmd.Flags().StringP(prjsReleaseCmdNameFlag, prjsReleaseCmdNameFlagAlias, flagUpdateKeepCurrentValue, prjsReleaseCmdNameFlagUsage)
	updateReleaseCmd.Flags().StringP(prjsAddCmdDescFlag, prjsAddCmdDescFlagAlias, flagUpdateKeepCurrentValue, prjsReleaseCmdDescFlagUsage)
	updateReleaseCmd.Flags().StringSlice(prjsReleaseCmdTeamFlag, []string{flagUpdateKeepCurrentValue}, prjsReleaseCmdTeamFlagUsage)
	updateReleaseCmd.Flags().StringSliceP(prjsReleaseCmdUserFlag, prjsReleaseCmdUserFlagAlias, []string{flagUpdateKeepCurrentValue}, prjsReleaseCmdUserFlagUsage)
	updateReleaseCmd.Flags().StringP(prjsReleaseCmdAllUsersFlag, prjsReleaseCmdAllUsersFlagAlias, flagUpdateKeepCurrentValue, prjsReleaseCmdAllUsersFlagUsage)
	updateReleaseCmd.Flags().Bool(prjsUpdateReleaseCmdResetUsersFlag, false, prjsUpdateReleaseCmdResetUsersFlagUsage)
	updateReleaseCmd.Flags().Bool(prjsUpdateReleaseCmdResetTeamsFlag, false, prjsUpdateReleaseCmdResetTeamsFlagUsage)
	updateReleaseCmd.Flags().Bool(prjsForceStateFlag, false, prjsForceStateFlagUsage)

	// Flags needed by the "projects releases deploy" command
	deployReleaseCmd.Flags().StringSliceP(prjsDeployReleaseCmdBoardsFlag, prjsDeployReleaseCmdBoardsFlagAlias, []string{}, prjsDeployReleaseCmdBoardsFlagUsage)

	// Flags needed by the "projects builds stats" command
	getBuildsStatsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flags needed by the "projects tests stats" command
	getTestsStatsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required
	addPrjCmd.MarkFlagRequired(prjsAddCmdNameFlag)
	addMembersCmd.MarkFlagRequired(prjsMembersAddCmdGuestFlag)
	addMembersCmd.MarkFlagRequired(prjsMembersAddCmdRoleFlag)
	updateMembersCmd.MarkFlagRequired(prjsMembersUpdateCmdMemberFlag)
	updateMembersCmd.MarkFlagRequired(prjsMembersAddCmdRoleFlag)
	deleteMembersCmd.MarkFlagRequired(prjsMembersUpdateCmdMemberFlag)
	startImportCmd.MarkFlagRequired(prjsImportCmdFileFlag)
	releaseCmd.MarkFlagRequired(prjsReleaseCmdDistroFlag)
	releaseCmd.MarkFlagRequired(prjsReleaseCmdVersionFlag)
	deployReleaseCmd.MarkFlagRequired(prjsDeployReleaseCmdBoardsFlag)

	// Add some flags as hidden
	addPrjCmd.Flags().MarkHidden(prjsAddCmdShortReleaseFlag)
	updatePrjCmd.Flags().MarkHidden(prjsAddCmdShortReleaseFlag)
	updatePrjCmd.Flags().MarkHidden(prjsForceStateFlag)
	updateReleaseCmd.Flags().MarkHidden(prjsForceStateFlag)

	// Specify dynamic flag completion

	// List
	listPrjCmd.RegisterFlagCompletionFunc(prjsListCmdTypeFlag, getListTypeAsTab)
	// Creation
	addPrjCmd.RegisterFlagCompletionFunc(prjsAddCmdMandArchFlag, getListArchAsTab)
	addPrjCmd.RegisterFlagCompletionFunc(prjsAddCmdOptArchFlag, getListArchAsTab)
	addPrjCmd.RegisterFlagCompletionFunc(prjsAddCmdMandDistFlag, getListDistAsTab)
	addPrjCmd.RegisterFlagCompletionFunc(prjsAddCmdOptDistFlag, getListDistAsTab)
	addPrjCmd.RegisterFlagCompletionFunc(prjsAddCmdInheritFlag, getPrjsAndRelPrjsAsTab)
	// Update
	updatePrjCmd.RegisterFlagCompletionFunc(prjsAddCmdMandArchFlag, getListArchAsTab)
	updatePrjCmd.RegisterFlagCompletionFunc(prjsAddCmdOptArchFlag, getListArchAsTab)
	updatePrjCmd.RegisterFlagCompletionFunc(prjsAddCmdMandDistFlag, getListDistAsTab)
	updatePrjCmd.RegisterFlagCompletionFunc(prjsAddCmdOptDistFlag, getListDistAsTab)
	updatePrjCmd.RegisterFlagCompletionFunc(prjsAddCmdInheritFlag, getPrjsAndRelPrjsAsTab)
	updatePrjCmd.RegisterFlagCompletionFunc(prjsUpdateCmdBldTypeFlag, getBuildTypesAsTab)
	updatePrjCmd.RegisterFlagCompletionFunc(prjsUpdateCmdBldTypeAppsFlag, getAppsListFromPrjAsTab)
	// Add members
	addMembersCmd.RegisterFlagCompletionFunc(prjsMembersAddCmdGuestFlag, getUsersToInviteAsTab)
	addMembersCmd.RegisterFlagCompletionFunc(prjsMembersAddCmdRoleFlag, getListRoleAsTab)
	// Update members
	updateMembersCmd.RegisterFlagCompletionFunc(prjsMembersUpdateCmdMemberFlag, getListMembersAsTab)
	updateMembersCmd.RegisterFlagCompletionFunc(prjsMembersAddCmdRoleFlag, getListRoleAsTab)
	// Delete member
	deleteMembersCmd.RegisterFlagCompletionFunc(prjsMembersUpdateCmdMemberFlag, getListMembersAsTab)
	// Release
	releaseCmd.RegisterFlagCompletionFunc(prjsReleaseCmdDistroFlag, getListDistAsTab)
	releaseCmd.RegisterFlagCompletionFunc(prjsReleaseCmdTeamFlag, getListTeamsAsTab)
	releaseCmd.RegisterFlagCompletionFunc(prjsReleaseCmdUserFlag, getListUserAsTab)
	// Release update
	updateReleaseCmd.RegisterFlagCompletionFunc(prjsReleaseCmdTeamFlag, getListTeamsAsTab)
	updateReleaseCmd.RegisterFlagCompletionFunc(prjsReleaseCmdUserFlag, getListUserAsTab)
	// Release deploy
	deployReleaseCmd.RegisterFlagCompletionFunc(prjsDeployReleaseCmdBoardsFlag, getOTAListBoardsAsTab)

	// Projects subcommands ------------------------------------
	// Add the group commands "projects" as subcommand of rootCmd
	rootCmd.AddCommand(projectsCmd)
	// Add the command "list" as a subcommand of projectsCmd
	projectsCmd.AddCommand(listPrjCmd)
	// Add the command "get" as a subcommand of projectsCmd
	projectsCmd.AddCommand(getPrjCmd)
	// Add the command "add" as a subcommand of projectsCmd
	projectsCmd.AddCommand(addPrjCmd)
	// Add the command "del" as a subcommand of projectsCmd
	projectsCmd.AddCommand(delPrjCmd)
	// Add the command "update" as a subcommand of projectsCmd
	projectsCmd.AddCommand(updatePrjCmd)
	// Add the command "bookmark" as a subcommand of projectsCmd
	projectsCmd.AddCommand(bookmarkPrjCmd)
	// Add the subcommands group "members" inside projectsCmd
	projectsCmd.AddCommand(membersCmd)
	// Add the subcommands group "export" inside projectsCmd
	projectsCmd.AddCommand(exportCmd)
	// Add the subcommands group "import" inside projectsCmd
	projectsCmd.AddCommand(importCmd)
	// Add the command "release" as a subcommand of projectsCmd
	projectsCmd.AddCommand(releaseCmd)
	// Add the command "releases" as a subcommand of projectsCmd
	projectsCmd.AddCommand(releasesCmd)
	// Add the command "builds" as a subcommand of projectsCmd
	projectsCmd.AddCommand(buildsCmd)
	// Add the command "tests" as a subcommand of projectsCmd
	projectsCmd.AddCommand(testsCmd)
	// ---------------------------------------------------------

	// Projects members subcommands ----------------------------
	membersCmd.AddCommand(getMembersCmd)
	membersCmd.AddCommand(addMembersCmd)
	membersCmd.AddCommand(updateMembersCmd)
	membersCmd.AddCommand(deleteMembersCmd)
	// ---------------------------------------------------------

	// Projects export subcommands -----------------------------
	exportCmd.AddCommand(startExportCmd)
	exportCmd.AddCommand(getExportCmd)
	exportCmd.AddCommand(downloadExportCmd)
	// ---------------------------------------------------------

	// Projects import subcommands -----------------------------
	importCmd.AddCommand(startImportCmd)
	importCmd.AddCommand(getImportCmd)
	// ---------------------------------------------------------

	// Projects releases subcommands ---------------------------
	releasesCmd.AddCommand(listReleaseCmd)
	releasesCmd.AddCommand(getReleaseCmd)
	releasesCmd.AddCommand(updateReleaseCmd)
	releasesCmd.AddCommand(deleteReleaseCmd)
	releasesCmd.AddCommand(deployReleaseCmd)
	// ---------------------------------------------------------

	// Projects builds subcommands -----------------------------
	buildsCmd.AddCommand(getBuildsStatsCmd)
	buildsCmd.AddCommand(getBuildsMatrixCmd)
	// ---------------------------------------------------------

	// Projects tests subcommands ------------------------------
	testsCmd.AddCommand(getTestsStatsCmd)
	// ---------------------------------------------------------
}

// initEventProject inits everything that is needed for the project add
// command to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventProject(cmd *cobra.Command, args []string) error {

	// Init the events list
	eventsList := []string{
		rpapiv1.EVTProjectAdd,
		rpapiv1.EVTProjectDelete,
		rpapiv1.EVTProjectStateChange,
		rpapiv1.EVTError,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// initEventProjectMembers inits everything that is needed for the project members
// add, update and delete commands to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventProjectMembers(cmd *cobra.Command, args []string) error {
	// Init the events list
	eventsList := []string{
		rpapiv1.EVTMemberAdd,
		rpapiv1.EVTMemberDelete,
		rpapiv1.EVTMemberStateChange,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// initEventProjectExport inits everything that is needed for the 'projects export start'
// command to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventProjectExport(cmd *cobra.Command, args []string) error {
	return CLIRegisterToMultipleEvents([]string{rpapiv1.EVTProjectExport})
}

// initEventProjectImport inits everything that is needed for the 'projects import start'
// command to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventProjectImport(cmd *cobra.Command, args []string) error {
	return CLIRegisterToMultipleEvents([]string{rpapiv1.EVTProjectImport})
}

// initEventProjectRelease inits everything that is needed for the project release
// command to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventProjectRelease(cmd *cobra.Command, args []string) error {
	// First init the project events
	err := initEventProject(cmd, args)
	if err != nil {
		return err
	}

	// Then the project release events
	eventsList := []string{
		rpapiv1.EVTReleaseProjectAdd,
		rpapiv1.EVTReleaseProjectDelete,
		rpapiv1.EVTReleaseProjectStateChange,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// initEventProjectDeploy inits everything that is needed for the 'projects releases deploy'
// command to run correctly
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventProjectDeploy(cmd *cobra.Command, args []string) error {
	return CLIRegisterToMultipleEvents([]string{rpapiv1.EVTDeployReleasedProjectStateChange})
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// mainGetListPrjsAsTab gets all the projects available in the Redpesk
// instance according to the list of types furnished
// and return it as an array of strings
//
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func mainGetListPrjsAsTab(typesList []string) ([]string, cobra.ShellCompDirective) {
	prjsList := []string{}

	// Get projects list
	prjs := rpapiv1.RPProjects{}
	if err := bckAPI.ProjectsListGetOfTypes(&prjs, typesList); err != nil {
		return prjsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the projects list, fill the prjsList tab
	for _, project := range prjs.Projects {
		prjsList = append(prjsList, project.Slug)
	}

	return prjsList, cobra.ShellCompDirectiveNoFileComp
}

// getListPrjsAsTab gets all the projects available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListPrjsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	typesList := []string{"standard"}

	// Depending if the user is admin, get all types of projects or just "standard" and "released"
	userAdmin, _ := isUserAdminOnCurrentStack()
	if userAdmin {
		typesList = append(typesList, "external")
	}

	return mainGetListPrjsAsTab(typesList)
}

// getFullListPrjsAsTab gets all (even external) projects available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getFullListPrjsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return mainGetListPrjsAsTab([]string{"standard", "external"})
}

// getListExtPrjsAsTab gets only external projects available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListExtPrjsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return mainGetListPrjsAsTab([]string{"external"})
}

// getListStandardPrjsAsTab gets only standard projects available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListStandardPrjsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return mainGetListPrjsAsTab([]string{"standard"})
}

// getListReleasedPrjsAsTab gets only released projects available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListReleasedPrjsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	prjsList := []string{}

	// Get the release projects
	tmpList := &rpapiv1.RPReleaseProjects{}
	if err := bckAPI.ListReleasedProject(tmpList); err != nil {
		return prjsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the projects list, fill the prjsList tab
	for _, project := range tmpList.ReleaseProjects {
		prjsList = append(prjsList, project.Slug)
	}

	return prjsList, cobra.ShellCompDirectiveNoFileComp
}

// getPrjsAndRelPrjsAsTab gets all projects available in the Redpesk
// instance (standard, external and released) and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getPrjsAndRelPrjsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	// Get the projects the standard way
	prjsList, _ := mainGetListPrjsAsTab([]string{"standard", "external"})
	// Now get the released projects
	relPrjsList, _ := getListReleasedPrjsAsTab(cmd, args, toComplete)
	// Merge the two slices
	prjsList = append(prjsList, relPrjsList...)
	return prjsList, cobra.ShellCompDirectiveNoFileComp
}

// getListArchAsTab gets all the architectures available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListArchAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	archsList := []string{}

	// Get architectures list
	archis := []rpapiv1.RPArch{}
	if err := bckAPI.RPArchGet(&archis, false); err != nil {
		return archsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the architectures list, fill the archsList tab
	for _, arch := range archis {
		archsList = append(archsList, arch.ID)
	}

	return archsList, cobra.ShellCompDirectiveNoFileComp
}

// getListDistAsTab gets all the distributions available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListDistAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	distrosList := []string{}

	// Get distributions list
	distros := rpapiv1.RPDistros{}
	if err := bckAPI.RPDistrosGet(&distros, false); err != nil {
		return distrosList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the distributions list, fill the distrosList tab
	for _, dist := range distros.Distros {
		distrosList = append(distrosList, dist.Slug)
	}

	return distrosList, cobra.ShellCompDirectiveNoFileComp
}

// getUsersToInviteAsTab gets all users that are available to invite for
// a given project. This function works only if the user has passed the project ID
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getUsersToInviteAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	memberList := []string{}

	// Get the project UUID from what the user gave us
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return memberList, cobra.ShellCompDirectiveNoFileComp
	}

	// Now we have a project ID we can retrieve the possible members
	invitList := rpapiv1.RPMembers{}
	if err := bckAPI.ProjectGetMembersToInvite(prjUUID.String(), &invitList); err != nil {
		return memberList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the invitation list, fill the memberList tab
	for _, guest := range invitList {
		memberList = append(memberList, guest.Pseudo)
	}

	return memberList, cobra.ShellCompDirectiveNoFileComp
}

// getListRoleAsTab gets all roles that are available for a project.
// This function works only if the user has passed the project ID
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListRoleAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	rolesList := []string{}

	// Get the project UUID from what the user gave us
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return rolesList, cobra.ShellCompDirectiveNoFileComp
	}

	// Now we have a project ID we can retrieve the possible roles
	rolesAvailable := rpapiv1.RPRoleProjects{}
	if err := bckAPI.ProjectGetMembersRoles(prjUUID.String(), &rolesAvailable); err != nil {
		return rolesList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the roles list, fill the rolesList tab
	for _, role := range rolesAvailable {
		rolesList = append(rolesList, role.RoleLabel)
	}

	return rolesList, cobra.ShellCompDirectiveNoFileComp
}

// getListMembersAsTab gets all users that are members of the given project.
// This function works only if the user has passed the project ID
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListMembersAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	membersList := []string{}

	// Get the project UUID from what the user gave us
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return membersList, cobra.ShellCompDirectiveNoFileComp
	}

	// Now we have a project ID we can retrieve the members
	membersAvailable := rpapiv1.RPMembers{}
	if err := bckAPI.ProjectMembersGet(prjUUID.String(), &membersAvailable); err != nil {
		return membersList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the members list, fill the membersList tab
	for _, member := range membersAvailable {
		membersList = append(membersList, member.Pseudo)
	}

	return membersList, cobra.ShellCompDirectiveNoFileComp
}

// getListTypeAsTab returns the static type available on projects/apps/images:
// standard, external, released
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListTypeAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	sliceToReturn := []string{"external", "standard"}

	// Get command parent use
	parentUse := ""
	parentCmd := cmd.Parent()
	if parentCmd != nil {
		parentUse = parentCmd.Use
	}

	if strings.HasPrefix(parentUse, prjsCommand) || strings.HasPrefix(parentUse, appsCommand) {
		sliceToReturn = append(sliceToReturn, "distro")
	}

	if strings.HasPrefix(parentUse, imagesCmdName) {
		sliceToReturn = append(sliceToReturn, "iotbzh")
	}

	return sliceToReturn, cobra.ShellCompDirectiveNoFileComp
}

// getAppsListFromPrjAsTab gets all applications of the given project.
// This function works only if the user has passed the project ID
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getAppsListFromPrjAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	appsList := []string{}

	// Get the project UUID from what the user gave us
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	// Now get the list of applications
	apps := rpapiv1.RPApplications{}
	if err := bckAPI.ApplicationsListGetOfTypes(&apps, prjUUID.String(), []string{}); err != nil {
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the apps list, fill the tab
	for _, app := range apps.Apps {
		appsList = append(appsList, app.Slug)
	}

	return appsList, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// listProjects gets all the projects available in the Redpesk
// instance and prints it
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func listProjects(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the bookmarked flag
	bookmarkOnly, err := cmd.Flags().GetBool(prjsListCmdBookmarkedFlag)
	if err != nil {
		logFlagError(prjsListCmdBookmarkedFlag)
		return err
	}

	// Get the type flag
	typesSlice, err := cmd.Flags().GetStringSlice(prjsListCmdTypeFlag)
	if err != nil {
		logFlagError(prjsListCmdTypeFlag)
		return err
	}

	// Convert "distro" to "external", and remove duplicated values
	utils.ReplaceInStringSlice(&typesSlice, "distro", "external")
	typesSlice = utils.RemoveDuplicateInStrSlice(typesSlice)

	// Prepare the function needed by the printer to get Projects
	getProjectsCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		// Get projects list
		prjs := rpapiv1.RPProjects{}
		if err := bckAPI.ProjectsListByTypePage(&prjs, typesSlice, pageNb, perPage, bookmarkOnly); err != nil {
			return prjs.Projects, printer.RPPrinterPaging{}, err
		}
		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&prjs)

		return prjs.Projects, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: projectDisplayVerboseLightLineNb,
		GetObjCB:      getProjectsCB,
		PrintFancyCB:  displayProjects,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct)
}

// getProject gets the project information available in the Redpesk
// instance and prints it
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getProject(cmd *cobra.Command, args []string) error {

	// Get the project UUID from what the user gave us
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Check that the project is not a released one
	// If it is the case, suggest to use the `rp-cli projects releases get <prj-id>` command
	tmpProject, err := getProjectFromCache(prjUUID)
	if err != nil {
		return err
	}
	if tmpProject.Type == rpapiv1.RPProjectTypeReleased {
		errMsg := fmt.Sprintf(prjGetReleasedErrorMsg, tmpProject.Slug)
		return errors.New(errMsg)
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the repository file flag
	repoFlag, err := cmd.Flags().GetBool(prjsGetCmdRepoFlag)
	if err != nil {
		logFlagError(prjsGetCmdRepoFlag)
		return err
	}

	// Get the roles flag
	rolesFlag, err := cmd.Flags().GetBool(prjsGetCmdRoleFlag)
	if err != nil {
		logFlagError(prjsGetCmdRoleFlag)
		return err
	}

	// Get information for this project
	prj := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectGet(prjUUID.String(), &prj); err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(prj, verbose, displayOneProject)
	if err != nil {
		return err
	}

	// Display the repository file if requested
	if repoFlag {
		// Retrieve the access token
		jwtToken := ""
		if len(prj.ProjectAccessTokens) > 0 {
			jwtToken = prj.ProjectAccessTokens[0].JWT
		}

		// Print the repositories
		displayProjectRepo(&prj.Repos, jwtToken)
	}

	// Get and display the available roles for this project if requested
	if rolesFlag {
		// Get all the roles available for this project
		rolesAvailable := rpapiv1.RPRoleProjects{}
		if err := bckAPI.ProjectGetMembersRoles(prjUUID.String(), &rolesAvailable); err != nil {
			return err
		}

		// Print them properly (label + name)
		RPPrinter.CustomPrintLn()
		RPPrinter.CustomPrintLn(prjsGetCmdAnsLine1)
		RPPrinter.CustomPrintLn(prjsGetCmdAnsLineDash)
		RPPrinter.CustomPrintLn()
		RPPrinter.CustomPrintLn(prjsGetCmdAnsLine2)
		for _, role := range rolesAvailable {
			RPPrinter.CustomPrintLn(role.RoleLabel, "\t", role.RoleName)
		}
	}

	return nil
}

// prepareProjectStruct prepare a project structure to send to the backend
// to create or update a project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param prj: Project structure to fill
// :type prj: *rpapiv1.RPProjectFull
// :param isCreation: Indicate if we are creating the project
// :type isCreation: bool
// :returns: error -- an error if a problem has occurred
func prepareProjectStruct(cmd *cobra.Command, args []string, prj *rpapiv1.RPProjectFull, isCreation bool) error {

	// First, take care of Name, alias and description
	// ------------------------------------------------------------------------

	// Name of the project
	prjName, err := cmd.Flags().GetString(prjsAddCmdNameFlag)
	if err != nil {
		logFlagError(prjsAddCmdNameFlag)
		return err
	}
	if prjName == "" {
		Log.Error(prjNameFlagNotSetShortMsg)
		return errors.New(prjNameFlagNotSetLongMsg)
	}

	// Alias, or slug
	var prjAlias string
	if isCreation {
		// Get the alias of the project wanted by the user (optional)
		prjAlias, err = cmd.Flags().GetString(prjsAddCmdAliasFlag)
		if err != nil {
			logFlagError(prjsAddCmdAliasFlag)
			return err
		}
	}

	// Get the short release naming option of the project wanted by the user (optional)
	prj.ShortReleaseNaming, err = genericGetBoolFlagValue(cmd, isCreation, prj.ShortReleaseNaming, prjsAddCmdShortReleaseFlag,
		fmt.Sprintf("%s \"%s\" flag", pbFlagMainPrefix, prjsAddCmdShortReleaseFlag))
	if err != nil {
		return err
	}

	// Description
	prjDesc, err := cmd.Flags().GetString(prjsAddCmdDescFlag)
	if err != nil {
		logFlagError(prjsAddCmdDescFlag)
		return err
	}

	// Fill the "supported object" structure
	// App
	prj.SupportedObjects.Applications, err = genericGetBoolFlagValue(cmd, isCreation, prj.SupportedObjects.Applications, prjsAddCmdAppSupportFlag,
		fmt.Sprintf("%s \"%s\" flag", pbFlagMainPrefix, prjsAddCmdAppSupportFlag))
	if err != nil {
		return err
	}
	// Image
	prj.SupportedObjects.Images, err = genericGetBoolFlagValue(cmd, isCreation, prj.SupportedObjects.Images, prjsAddCmdImgSupportFlag,
		fmt.Sprintf("%s \"%s\" flag", pbFlagMainPrefix, prjsAddCmdImgSupportFlag))
	if err != nil {
		return err
	}
	// Container
	prj.SupportedObjects.Containers, err = genericGetBoolFlagValue(cmd, isCreation, prj.SupportedObjects.Containers, prjsAddCmdCtnSupportFlag,
		fmt.Sprintf("%s \"%s\" flag", pbFlagMainPrefix, prjsAddCmdCtnSupportFlag))
	if err != nil {
		return err
	}

	// Fill the project structure
	if isCreation {
		prj.Name = prjName
		prj.Slug = prjAlias
		prj.Description = prjDesc
	} else {
		if prjName != flagUpdateKeepCurrentValue {
			prj.Name = prjName
		}
		if prjDesc != flagUpdateKeepCurrentValue {
			prj.Description = prjDesc
		}
	}
	// ------------------------------------------------------------------------

	// Then, take care of inheritance, distros and arch
	// ------------------------------------------------------------------------

	// Get the possible inheritances wanted by the user
	prjInheritances, err := cmd.Flags().GetStringSlice(prjsAddCmdInheritFlag)
	if err != nil {
		logFlagError(prjsAddCmdInheritFlag)
		return err
	}

	// Get the list of all the mandatory architectures
	prjMandArch, err := cmd.Flags().GetStringSlice(prjsAddCmdMandArchFlag)
	if err != nil {
		logFlagError(prjsAddCmdMandArchFlag)
		return err
	}

	// Get the list of all the optional architectures
	prjOptArch, err := cmd.Flags().GetStringSlice(prjsAddCmdOptArchFlag)
	if err != nil {
		logFlagError(prjsAddCmdOptArchFlag)
		return err
	}

	// Get the list of all the mandatory distributions
	prjMandDistros, err := cmd.Flags().GetStringSlice(prjsAddCmdMandDistFlag)
	if err != nil {
		logFlagError(prjsAddCmdMandDistFlag)
		return err
	}

	// Get the list of all the optional distributions
	prjOptDistros, err := cmd.Flags().GetStringSlice(prjsAddCmdOptDistFlag)
	if err != nil {
		logFlagError(prjsAddCmdOptDistFlag)
		return err
	}

	// Set the structure we will use later
	projectArchsAndDistStruct := RPProjectArchDist{
		MandDistros: prjMandDistros,
		OptDistros:  prjOptDistros,
		MandArches:  prjMandArch,
		OptArches:   prjOptArch,
	}

	if !isCreation {
		// Get the arches and distros structure from the already existing project
		tmpProjectArchsAndDistros := fillArchDistFromProject(*prj)

		// Get the inheritances project from the already existing project
		tmpInheritances := []string{}
		for _, inheritance := range prj.Inheritances {
			tmpInheritances = append(tmpInheritances, inheritance.ID.String())
		}

		// If the user keeps the old mandatory distros, overwrite it in projectArchsAndDistStruct
		if utils.ContainsString(prjMandDistros, flagUpdateKeepCurrentValue) {
			// Reset the destination slice in case of an empty slice in the source
			projectArchsAndDistStruct.MandDistros = make([]string, len(tmpProjectArchsAndDistros.MandDistros))
			copy(projectArchsAndDistStruct.MandDistros, tmpProjectArchsAndDistros.MandDistros)
		}

		// If the user keeps the old optional distros, overwrite it in projectArchsAndDistStruct
		if utils.ContainsString(prjOptDistros, flagUpdateKeepCurrentValue) {
			// Reset the destination slice in case of an empty slice in the source
			projectArchsAndDistStruct.OptDistros = make([]string, len(tmpProjectArchsAndDistros.OptDistros))
			copy(projectArchsAndDistStruct.OptDistros, tmpProjectArchsAndDistros.OptDistros)
		}

		// If the user keeps the old mandatory arches, overwrite it in projectArchsAndDistStruct
		if utils.ContainsString(prjMandArch, flagUpdateKeepCurrentValue) {
			// Reset the destination slice in case of an empty slice in the source
			projectArchsAndDistStruct.MandArches = make([]string, len(tmpProjectArchsAndDistros.MandArches))
			copy(projectArchsAndDistStruct.MandArches, tmpProjectArchsAndDistros.MandArches)
		}

		// If the user keeps the old optional arches, overwrite it in projectArchsAndDistStruct
		if utils.ContainsString(prjOptArch, flagUpdateKeepCurrentValue) {
			// Reset the destination slice in case of an empty slice in the source
			projectArchsAndDistStruct.OptArches = make([]string, len(tmpProjectArchsAndDistros.OptArches))
			copy(projectArchsAndDistStruct.OptArches, tmpProjectArchsAndDistros.OptArches)
		}

		// If the user keeps the old inheritances, overwrite it
		if utils.ContainsString(prjInheritances, flagUpdateKeepCurrentValue) {
			// Reset the destination slice in case of an empty slice in the source
			prjInheritances = make([]string, len(tmpInheritances))
			copy(prjInheritances, tmpInheritances)
		}

		// Reset the values of tabs that will be filled after
		prj.Inheritances = rpapiv1.RPInheritances{}
		prj.Architectures = []rpapiv1.RPArch{}
		prj.Distributions = []rpapiv1.RPDistro{}
	}

	// Switch case according to the number of inheritance
	archNb := len(projectArchsAndDistStruct.MandArches) + len(projectArchsAndDistStruct.OptArches)
	distroNb := len(projectArchsAndDistStruct.MandDistros) + len(projectArchsAndDistStruct.OptDistros)
	switch inheritNb := len(prjInheritances); inheritNb {
	case 0:
		// The user did not specify any inheritance for the current project

		// Check that we have at least one arch (optional or mandatory)
		RPPrinter.CustomPrint(prjsAddCmdAnsLine1)
		if archNb == 0 {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			Log.Error(prjArchNotSetShortMsg)
			return errors.New(prjArchNotSetLongMsg)
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

		if prj.Type != rpapiv1.RPProjectTypeExternal {
			// Check that we have at least one distro (optional or mandatory)
			RPPrinter.CustomPrint(prjsAddCmdAnsLine2)
			if distroNb == 0 {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				Log.Error(prjDistNotSetShortMsg)
				return errors.New(prjDistNotSetLongMsg)
			}
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		}

	case 1:
		// The user specified one project to inherit from

		if (archNb + distroNb) == 0 {
			// The user did not specify anything, get the distributions and arches from the parent project
			distsArches, err := getDistAndArchFromPrj(prjInheritances[0])
			if err != nil {
				return err
			}

			// Set the distributions and architectures wanted
			projectArchsAndDistStruct.MandDistros = make([]string, len(distsArches.MandDistros))
			copy(projectArchsAndDistStruct.MandDistros, distsArches.MandDistros)

			projectArchsAndDistStruct.OptDistros = make([]string, len(distsArches.OptDistros))
			copy(projectArchsAndDistStruct.OptDistros, distsArches.OptDistros)

			projectArchsAndDistStruct.MandArches = make([]string, len(distsArches.MandArches))
			copy(projectArchsAndDistStruct.MandArches, distsArches.MandArches)

			projectArchsAndDistStruct.OptArches = make([]string, len(distsArches.OptArches))
			copy(projectArchsAndDistStruct.OptArches, distsArches.OptArches)

		} else if (archNb == 0) || (distroNb == 0) {
			// The user only partially specified what he wanted. He either did not put arch or distro
			// Return an error
			Log.Error(prjPartialDistArchInheritShortMsg)
			return errors.New(prjPartialDistArchInheritLongMsg)
		}

	default:
		// The user specified more than 1 project, print a warning and don't check anything
		RPPrinter.CustomPrintLn(prjInheritanceMoreThanOne)
	}

	// Add the inheritance for the project to create/update
	for _, inherit := range prjInheritances {
		inheritUUID, err := getProjectUUID(inherit)
		if err != nil {
			// The project ID has not been found, try to find a released project
			inheritUUID, err = getRelProjectUUID(inherit)
			if err != nil {
				return err
			}
		}
		// Create a temporary inheritance structure
		tmpInheritance := rpapiv1.RPInheritance{
			ID: inheritUUID,
		}
		// Append the inheritance to the slice of it
		prj.Inheritances = append(prj.Inheritances, tmpInheritance)
	}

	// Check if the distribution and the architectures specified are available
	availableArch := []rpapiv1.RPArch{}
	availableDist := rpapiv1.RPDistros{}
	// Get the list of availables architecture
	err = bckAPI.RPArchGet(&availableArch, false)
	if err != nil {
		return err
	}
	// Get the list of availables distributions
	err = bckAPI.RPDistrosGet(&availableDist, false)
	if err != nil {
		return err
	}

	// Go through the available architectures and enable what is needed
	for index, arch := range availableArch {
		// Check if the current arch is in the list of mandatory arch
		if utils.ContainsString(projectArchsAndDistStruct.MandArches, arch.ID) {
			// The current arch is mandatory
			availableArch[index].Mandatory = true
			availableArch[index].Set = true
		} else if utils.ContainsString(projectArchsAndDistStruct.OptArches, arch.ID) {
			// The current arch is optional
			availableArch[index].Mandatory = false
			availableArch[index].Set = true
		} else {
			// The current arch is not wanted
			availableArch[index].Mandatory = false
			availableArch[index].Set = false
		}
		// Add the Arch for the project
		prj.Architectures = append(prj.Architectures, availableArch[index])
	}

	// Go through the available distributions and enable what is needed
	for index, distro := range availableDist.Distros {
		// Check if the current distro is in the list of mandatory distros
		if utils.ContainsString(projectArchsAndDistStruct.MandDistros, distro.Slug) {
			// The current distro is mandatory
			availableDist.Distros[index].Mandatory = true
			availableDist.Distros[index].Set = true
		} else if utils.ContainsString(projectArchsAndDistStruct.OptDistros, distro.Slug) {
			// The current distro is optional
			availableDist.Distros[index].Mandatory = false
			availableDist.Distros[index].Set = true
		} else {
			// The current distro is not wanted
			availableDist.Distros[index].Mandatory = false
			availableDist.Distros[index].Set = false
		}
		// Add the Dist for the project
		prj.Distributions = append(prj.Distributions, availableDist.Distros[index])
	}

	// ------------------------------------------------------------------------

	// Everything went well, return nil
	return nil
}

// createProject creates a project as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func createProject(cmd *cobra.Command, args []string) error {

	// Display the first line to the user
	RPPrinter.CustomPrintLn(prjsAddCmdAnsLine0)

	// Create the structure to fill
	prj := rpapiv1.RPProjectFull{}

	// Fill the structure with the right values (given in command line)
	if err := prepareProjectStruct(cmd, args, &prj, true); err != nil {
		return err
	}

	// Log the project creation
	Log.Infof(projectCreationInfoMsg, prj.Name)
	Log.Debugf(projectCreationDebugMsg, prj)

	// Tell the user that the project is currently being created
	RPPrinter.CustomPrintF(prjsAddCmdAnsLine3, prj.Name)

	// Request the creation of this project to the backend
	prjAns := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectCreate(&prj, &prjAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(prjAns, false, displayNothing)

	// Create the sentence to display when the object will be added
	slugSentence := fmt.Sprintf(objectSlugDisplayMsg, "project", prjAns.Slug)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		// Display the slug of the created object
		RPPrinter.CustomPrintLn(slugSentence)
		return nil
	}

	// Block until we receive a corresponding websocket event
	err := waitForStableProject(prjAns.ID, rpapiv1.EVTProjectAdd)
	if err != nil {
		return err
	}

	// Display the slug of the object created
	RPPrinter.CustomPrintLn(slugSentence)

	return nil
}

// updateProject update a project as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func updateProject(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(prjsUpdateCmdAnsLine0)

	// Get the project UUID from what the user gave us
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Get the current project
	currPrj := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectGet(prjUUID.String(), &currPrj); err != nil {
		return err
	}

	// Reset the state of the project in order to not lead to any problem
	// cf. bug #5015 in Redmine
	currPrj.State = ""

	// Fill the structure with the right values (given in command line)
	if err := prepareProjectStruct(cmd, args, &currPrj, false); err != nil {
		return err
	}

	// Get the "force-stable-state" flag
	forceStable, err := cmd.Flags().GetBool(prjsForceStateFlag)
	if err != nil {
		return err
	}
	// If the user wants it, force the project state at "stable"
	if forceStable {
		currPrj.State = rpapiv1.RPPrjStableState
	}

	// Get the "build-type" flag
	bldType, err := cmd.Flags().GetString(prjsUpdateCmdBldTypeFlag)
	if err != nil {
		return err
	}

	// Get the "bld-type-apps" flag
	bldTypeApps, err := cmd.Flags().GetStringSlice(prjsUpdateCmdBldTypeAppsFlag)
	if err != nil {
		return err
	}

	// Get which kind of update is needed
	classicUpdateNeeded, bldTypeUpdateNeeded := prjGetUpdatesNeeded(cmd)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)

	prjAns := rpapiv1.RPProjectFull{}

	if bldTypeUpdateNeeded {
		// Tell the user that the build type is being updated
		RPPrinter.CustomPrintF(prjsUpdateCmdAnsLine2, currPrj.Slug)

		// Update the build type of all the apps or just some
		err = changeBuildType(bldType, prjUUID, bldTypeApps, &prjAns)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}

		// Display something to the user if in raw mode
		RPPrinter.CustomPrintObject(prjAns, false, displayNothing)

		if nonBlocking {
			// The user do not want blocking functions
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		} else {
			// Block until we receive a corresponding websocket event
			err := waitForStableProject(prjAns.ID, rpapiv1.EVTProjectStateChange)
			if err != nil {
				return err
			}
		}
	}

	// If we do 2 operations, we wait a bit here
	if classicUpdateNeeded && bldTypeUpdateNeeded {
		time.Sleep(500 * time.Millisecond)
	}

	if classicUpdateNeeded {
		// Tell the user that the project is currently being updated
		RPPrinter.CustomPrintF(prjsUpdateCmdAnsLine1, currPrj.Slug)

		// Request the update of this project to the backend
		if err := bckAPI.ProjectUpdate(prjUUID.String(), &currPrj, &prjAns); err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}

		// Display something to the user if in raw mode
		RPPrinter.CustomPrintObject(prjAns, false, displayNothing)

		if nonBlocking {
			// The user do not want blocking functions
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		} else {
			// Block until we receive a corresponding websocket event
			err := waitForStableProject(prjAns.ID, rpapiv1.EVTProjectStateChange)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// deleteProject deletes an existing project as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func deleteProject(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(prjsDelCmdAnsLine0)

	// Get the project UUID from what the user gave us
	prjUUID, prjSlug, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Tell the user that the project is currently being deleted
	RPPrinter.CustomPrintF(prjsDelCmdAnsLine1, prjSlug)

	// Ask the backend for deletion
	prjAns := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectDelete(prjUUID.String(), &prjAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(prjAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	err = waitForStableProject(prjAns.ID, rpapiv1.EVTProjectDelete)
	if err != nil {
		return err
	}

	return nil
}

// bookmarkProject bookmarks an existing project as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func bookmarkProject(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(prjsBookmarkCmdAnsLine0)

	// Get the project UUID from what the user gave us
	prjUUID, prjSlug, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Get the reset flag
	reset, err := cmd.Flags().GetBool(prjsBookmarkCmdResetFlag)
	if err != nil {
		logFlagError(prjsBookmarkCmdResetFlag)
		return err
	}

	// Get the project concerned by the bookmarking
	currPrj := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectGet(prjUUID.String(), &currPrj); err != nil {
		return err
	}

	// Do the appropriate action according to the variables
	resetDone := false
	setDone := false
	bookmarkObj := rpapiv1.RPProjectBookmark{}
	switch {
	case (reset && !currPrj.Bookmarked):
		// Reset is requested but the project is not bookmarked, nothing to do
		RPPrinter.CustomPrintF(prjsBookmarkCmdAnsLine3, prjSlug)
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	case (reset && currPrj.Bookmarked):
		// Reset is requested and the project is bookmarked
		RPPrinter.CustomPrintF(prjsBookmarkCmdAnsLine2, prjSlug)
		if err := bckAPI.ProjectBookmark(prjUUID.String(), &bookmarkObj); err != nil {
			return err
		}
		resetDone = true

	case (!reset && currPrj.Bookmarked):
		// Bookmark is requested but the project is already bookmarked, nothing to do
		RPPrinter.CustomPrintF(prjsBookmarkCmdAnsLine4, prjSlug)
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	case (!reset && !currPrj.Bookmarked):
		// Bookmark is requested and the project is not bookmarked
		RPPrinter.CustomPrintF(prjsBookmarkCmdAnsLine1, prjSlug)
		if err := bckAPI.ProjectBookmark(prjUUID.String(), &bookmarkObj); err != nil {
			return err
		}
		setDone = true
	}

	if setDone || resetDone {

		// Display something to the user if in raw mode
		RPPrinter.CustomPrintObject(bookmarkObj, false, displayNothing)

		// Check from viper the non-blocking flag to see if we wait
		nonBlocking := viper.GetBool(nonBlockingFlag)
		if nonBlocking {
			// The user do not want blocking functions
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
			return nil
		}

		// Block until we receive a corresponding websocket event
		for {
			rEvent := <-cmdEventChannel

			// Check that the received event is for a project addition
			if rEvent.Type == rpapiv1.EVTProjectStateChange {
				// Function handling the events
				rProject, err := rEvent.DecodeProject()
				if err != nil {
					return err
				}

				if utils.UUIDEqual(rProject.ID, bookmarkObj.ProjectID) {
					if rProject.Bookmarked != reset {
						// Display OK if not in raw mode
						RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
						// Display something to the user if in raw mode
						RPPrinter.CustomPrintObject(rProject, false, displayNothing)
						break
					}
				}
			}
		}
	}
	return nil
}

// getProjectMembers gets all the projects members in a project
// and prints it
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getProjectMembers(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the project UUID from what the user gave us
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Fill the roles cache for this project
	err = populateRolesCache(prjUUID.String())
	if err != nil {
		return err
	}

	// Get project members list
	prjMembers := rpapiv1.RPMembers{}
	if err := bckAPI.ProjectMembersGet(prjUUID.String(), &prjMembers); err != nil {
		return err
	}

	// Display the members
	RPPrinter.CustomPrintObject(prjMembers, verbose, displayMembers)

	return nil
}

// prepareProjectMember prepare a project member structure to send
// to the backend to add or update a project member
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param mbr: Member structure to fill
// :type mbr: *rpapiv1.RPMember
// :param isAddition: Indicate if we are adding the member
// :type isAddition: bool
// :param isDeletion: Indicate if we are deleting the member
// :type isDeletion: bool
// :returns: string, error -- the prjID passed by the user and an error if a problem has occurred
func prepareProjectMember(cmd *cobra.Command, args []string, mbr *rpapiv1.RPMember, isAddition bool, isDeletion bool) (string, error) {
	result := ""

	// Get the project UUID from what the user gave us
	prjUUID, prjSlug, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return result, err
	}

	var member rpapiv1.RPMember
	if isAddition {
		// Get the guest flag
		guestID, err := cmd.Flags().GetString(prjsMembersAddCmdGuestFlag)
		if err != nil {
			logFlagError(prjsMembersAddCmdGuestFlag)
			return result, err
		}
		if guestID == "" {
			Log.Error(prjGuestNotSetMsg)
			return result, errors.New(prjGuestRequiredMsg)
		}

		// Get all the guest IDs needed
		member, err = getGuestInfo(guestID, prjUUID)
		if err != nil {
			return result, err
		}

	} else {
		// Get the member flag
		memberID, err := cmd.Flags().GetString(prjsMembersUpdateCmdMemberFlag)
		if err != nil {
			logFlagError(prjsMembersUpdateCmdMemberFlag)
			return result, err
		}
		if memberID == "" {
			Log.Error(prjMemberNotSetMsg)
			return result, errors.New(prjMemberRequiredMsg)
		}

		// Get all the guest IDs needed
		member, err = getMemberInfo(memberID, prjUUID)
		if err != nil {
			return result, err
		}
	}

	if !isDeletion {
		// Get the role flag
		roles, err := cmd.Flags().GetStringSlice(prjsMembersAddCmdRoleFlag)
		if err != nil {
			logFlagError(prjsMembersAddCmdRoleFlag)
			return result, err
		}
		if len(roles) == 0 {
			Log.Error(prjRoleNotSetMsg)
			return result, errors.New(prjRoleRequiredMsg)
		}

		// Fill the roles cache for this project
		err = populateRolesCache(prjUUID.String())
		if err != nil {
			return result, err
		}

		// Get the roles IDs for the current project, from what the user gave
		mbr.Roles, err = getRolesStruct(roles)
		if err != nil {
			return result, err
		}
	}

	// Fill the member structure
	mbr.ID = member.ID
	mbr.Pseudo = member.Pseudo
	mbr.ObjectID = prjUUID

	return prjSlug, nil
}

// addProjectMembers add a new project member in a project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func addProjectMembers(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(prjsMembersAddCmdAnsLine0)

	// Prepare the structure to send to the backend to add a member
	memberToAdd := rpapiv1.RPMember{}
	prjID, err := prepareProjectMember(cmd, args, &memberToAdd, true, false)
	if err != nil {
		return err
	}

	// Create the roles string
	rolesStr := ""
	for index, role := range memberToAdd.Roles {
		if index > 0 {
			rolesStr += ", "
		}
		rolesStr += role.Label
	}
	// Print an information line to the user
	RPPrinter.CustomPrintF(prjsMembersAddCmdAnsLine1, memberToAdd.Pseudo, rolesStr, prjID)

	// Request to add a new member
	ansMember := rpapiv1.RPMember{}
	if err := bckAPI.ProjectAddMember(memberToAdd.ObjectID.String(), &memberToAdd, &ansMember); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(ansMember, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project addition
		if rEvent.Type == rpapiv1.EVTMemberAdd {
			// Function handling the events
			rMember, err := rEvent.DecodeMember()
			if err != nil {
				return err
			}

			if utils.UUIDEqual(rMember.ID, memberToAdd.ID) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rMember, false, displayNothing)
				break
			}
		}
	}

	return nil
}

// updateProjectMembers update a project member in the given project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func updateProjectMembers(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(prjsMembersUpdateCmdAnsLine0)

	// Prepare the structure to send to the backend to add a member
	memberToUpdate := rpapiv1.RPMember{}
	prjID, err := prepareProjectMember(cmd, args, &memberToUpdate, false, false)
	if err != nil {
		return err
	}

	// Create the roles string
	rolesStr := ""
	for index, role := range memberToUpdate.Roles {
		if index > 0 {
			rolesStr += ", "
		}
		rolesStr += role.Label
	}
	// Print an information line to the user
	RPPrinter.CustomPrintF(prjsMembersUpdateCmdAnsLine1, memberToUpdate.Pseudo, rolesStr, prjID)

	// Request to update an existing member
	ansMember := rpapiv1.RPMember{}
	if err := bckAPI.ProjectUpdateMember(memberToUpdate.ObjectID.String(), &memberToUpdate, &ansMember); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(ansMember, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project addition
		if rEvent.Type == rpapiv1.EVTMemberStateChange {
			// Function handling the events
			rMember, err := rEvent.DecodeMember()
			if err != nil {
				return err
			}

			if utils.UUIDEqual(rMember.ID, memberToUpdate.ID) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rMember, false, displayNothing)
				break
			}
		}
	}

	return nil
}

// deleteProjectMembers deletes a member from the given project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func deleteProjectMembers(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(prjsMembersDeleteCmdAnsLine0)

	// Prepare the structure to send to the backend to add a member
	memberToDelete := rpapiv1.RPMember{}
	prjID, err := prepareProjectMember(cmd, args, &memberToDelete, false, true)
	if err != nil {
		return err
	}

	// Print an information line to the user
	RPPrinter.CustomPrintF(prjsMembersDeleteCmdAnsLine1, memberToDelete.Pseudo, prjID)

	// Request to delete an existing member
	ansMember := rpapiv1.RPMember{}
	if err := bckAPI.ProjectDeleteMember(memberToDelete.ObjectID.String(), &memberToDelete, &ansMember); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(ansMember, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project addition
		if rEvent.Type == rpapiv1.EVTMemberDelete {
			// Function handling the events
			rMember, err := rEvent.DecodeMember()
			if err != nil {
				return err
			}

			if utils.UUIDEqual(rMember.ID, memberToDelete.ID) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rMember, false, displayNothing)
				break
			}
		}
	}

	return nil
}

// startExportProject starts the export of a project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func startExportProject(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(prjsExportStartCmdAnsLine0)

	// Get the project UUID from what the user gave us
	prjUUID, prjSlug, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Request the backend to start the export of the project
	RPPrinter.CustomPrintF(prjsExportStartCmdAnsLine1, prjSlug)
	prjExportState := rpapiv1.RPExportProjectState{}
	if err = bckAPI.ProjectExport(prjUUID.String(), &prjExportState); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Request went OK
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	RPPrinter.ForceFlush()

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		return nil
	}

	RPPrinter.CustomPrint(prjsExportStartCmdAnsLine2)

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project export
		if rEvent.Type == rpapiv1.EVTProjectExport {
			// Decode the event data
			currentProjectState, err := rEvent.DecodeExportProject()
			if err != nil {
				return err
			}

			// Check that this is our project
			if !utils.UUIDEqual(currentProjectState.ProjectID, prjExportState.ProjectID) {
				continue
			}

			if currentProjectState.Status == rpapiv1.RPStateExportProjectDone {

				// The export is done
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(currentProjectState, false, displayNothing)

				break

			} else if currentProjectState.Status == rpapiv1.RPStateExportProjectFAILED {

				// The export has failed
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(currentProjectState, false, displayNothing)

				return errors.New(currentProjectState.Error)
			}

		} else if rEvent.Type == rpapiv1.EVTError {
			// Function handling the events
			err = handleErrorEvent(prjExportState.ProjectID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return downloadExportedPrj(cmd, prjUUID)
}

// getExportProject get the status of a project export
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getExportProject(cmd *cobra.Command, args []string) error {
	// Get the project UUID from what the user gave us
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Request the state to the backend
	prjExportState := rpapiv1.RPExportProjectState{}
	if err = bckAPI.ProjectGetExportState(prjUUID.String(), &prjExportState); err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(prjExportState, verbose, displayExportState)
	if err != nil {
		return err
	}

	return nil
}

// downloadExportProject download the archive corresponding to an
// exported project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func downloadExportProject(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(prjsExportDownloadCmdAnsLine0)

	// Get the project UUID from what the user gave us
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	return downloadExportedPrj(cmd, prjUUID)
}

// startImportProject starts the import of a project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func startImportProject(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(prjsImportStartCmdAnsLine0)

	// Get the file flag, giving the archive to use to import the project
	filepath, err := cmd.Flags().GetString(prjsImportCmdFileFlag)
	if err != nil {
		logFlagError(prjsImportCmdFileFlag)
		return err
	}

	// Check that the file exists
	_, err = os.Stat(filepath)
	if errors.Is(err, os.ErrNotExist) {
		errMsg := fmt.Sprintf(prjImportFileNotExisting, filepath)
		return errors.New(errMsg)
	}

	// Get the interactive flag, to see if the user request an interactive import
	interactive, err := cmd.Flags().GetBool(prjsImportCmdInteractiveFlag)
	if err != nil {
		logFlagError(prjsImportCmdInteractiveFlag)
		return err
	}

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)

	// Start uploading the archive to the backend
	RPPrinter.CustomPrint(prjsImportStartCmdAnsLine1)
	prjImportState := rpapiv1.RPImportProjectState{}
	if err = bckAPI.ProjectImportUpload(filepath, &prjImportState); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Wait for upload to finish completely
	if !nonBlocking {
		err = waitForImportEvent(rpapiv1.RPStateImportProjectUploadManageInfoDone, rpapiv1.RPStateImportProjectUploadManageInfoFailed, prjImportState.FuturePrjID)
		if err != nil {
			return err
		}
	}

	if interactive && !nonBlocking {
		var answer string
		// If the interactive flag is set, wait for the user input
		RPPrinter.CustomPrintLn(prjsImportStartCmdAnsLine4)
		RPPrinter.ForceFlush()
		// Get the input
		fmt.Scanln(&answer)
		// Add a newline
		RPPrinter.CustomPrintLn()
	}

	// Request the backend to start importing the project from the archive previously uploaded
	RPPrinter.CustomPrint(prjsImportStartCmdAnsLine2)
	if err = bckAPI.ProjectImport(prjImportState.FuturePrjID.String(), &prjImportState); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Request went OK
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	RPPrinter.ForceFlush()

	if nonBlocking {
		return nil
	}

	RPPrinter.CustomPrint(prjsImportStartCmdAnsLine3)

	// Block until we receive a corresponding websocket event
	err = waitForImportEvent(rpapiv1.RPStateImportProjectDone, rpapiv1.RPStateImportProjectFailed, prjImportState.FuturePrjID)

	return err
}

// waitForImportEvent is a function allowing to test if we received the right
// event at the right moment
//
// :param successStatus: Status that will be considered as successfull for the function
// :type successStatus: rpapiv1.RPStateImportProject
// :param failStatus: Status that will be considered as fail for the function
// :type failStatus: rpapiv1.RPStateImportProject
// :param targetUUID: UUID of the project that is interesting us
// :type targetUUID: uuid.UUID
// :returns: error -- an error if a problem has occurred
func waitForImportEvent(successStatus, failStatus rpapiv1.RPStateImportProject, targetUUID uuid.UUID) error {
	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project export
		if rEvent.Type == rpapiv1.EVTProjectImport {
			// Decode the event data
			currentProjectState, err := rEvent.DecodeImportProject()
			if err != nil {
				return err
			}

			// Check that this is our project
			if !utils.UUIDEqual(currentProjectState.FuturePrjID, targetUUID) {
				continue
			}

			if currentProjectState.Status == successStatus {

				// The import is done
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(currentProjectState, false, displayNothing)

				break

			} else if currentProjectState.Status == failStatus {

				// The export has failed
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(currentProjectState, false, displayNothing)

				return errors.New(currentProjectState.Error)
			}

		} else if rEvent.Type == rpapiv1.EVTError {
			// Function handling the events
			err := handleErrorEvent(targetUUID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// getImportProject get the status of a project import
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getImportProject(cmd *cobra.Command, args []string) error {
	// Get the project UUID from what the user gave us
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Request the state to the backend
	prjImportState := rpapiv1.RPImportProjectState{}
	if err = bckAPI.ProjectGetImportState(prjUUID.String(), &prjImportState); err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(prjImportState, verbose, displayImportState)
	if err != nil {
		return err
	}

	return nil
}

// prepareRelProjectStruct prepare the rpapiv1.RPReleaseProject structure needed to release a project.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param relPrj: Release Project structure to fill
// :type relPrj: *rpapiv1.RPReleaseProject
// :param isCreation: Indicate if we are creating the project
// :type isCreation: bool
// :returns: error -- an error if a problem has occurred
func prepareRelProjectStruct(cmd *cobra.Command, relPrj *rpapiv1.RPReleaseProject, isCreation bool) error {

	// First, take care of name and description
	// ----------------------------------------

	// Name of the project release
	relPrjName, err := cmd.Flags().GetString(prjsReleaseCmdNameFlag)
	if err != nil {
		logFlagError(prjsReleaseCmdNameFlag)
		return err
	}

	// Get the description of the released project
	relPrjDesc, err := cmd.Flags().GetString(prjsAddCmdDescFlag)
	if err != nil {
		logFlagError(prjsAddCmdDescFlag)
		return err
	}

	// Then fill the information we got in the expected structure
	// ----------------------------------------------------------

	if isCreation {
		// Name and description management
		relPrj.Name = relPrjName
		relPrj.Description = relPrjDesc

	} else {
		if relPrjName != flagUpdateKeepCurrentValue {
			relPrj.Name = relPrjName
		}
		if relPrjDesc != flagUpdateKeepCurrentValue {
			relPrj.Description = relPrjDesc
		}
	}

	return nil
}

// prepareRPObjectRolesStruct prepare the rpapiv1.RPObjectRoles structure needed to change the audience on a released object
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param objRoles: Input object roles structure to fill and containing the availables roles
// :type objRoles: *rpapiv1.RPObjectRoles
// :param isCreation: Indicate if we are creating the project
// :type isCreation: bool
// :returns: bool, error -- true if the user gave an audience parameter and an error if a problem has occurred
func prepareRPObjectRolesStruct(cmd *cobra.Command, objRoles *rpapiv1.RPObjectRoles, isCreation bool) (bool, error) {

	// First, take care of teams, users and factory flags
	// ------------------------------------------------

	// Get the teams to potentially add in the audience
	relPrjTeams, err := cmd.Flags().GetStringSlice(prjsReleaseCmdTeamFlag)
	if err != nil {
		logFlagError(prjsReleaseCmdTeamFlag)
		return false, err
	}

	// Get the users to potentially add in the audience
	relPrjUsers, err := cmd.Flags().GetStringSlice(prjsReleaseCmdUserFlag)
	if err != nil {
		logFlagError(prjsReleaseCmdUserFlag)
		return false, err
	}

	// Get the factory flag for the audience
	objRoles.Stack, err = genericGetBoolFlagValue(cmd, isCreation, objRoles.Stack, prjsReleaseCmdAllUsersFlag,
		fmt.Sprintf("%s \"%s\" flag", pbFlagMainPrefix, prjsReleaseCmdAllUsersFlag))
	if err != nil {
		return false, err
	}

	// If in update mode check if the user wants to reset the users or the teams
	resetUsers := false
	resetTeams := false
	if !isCreation {
		// Get the reset users flag
		resetUsers, err = cmd.Flags().GetBool(prjsUpdateReleaseCmdResetUsersFlag)
		if err != nil {
			logFlagError(prjsUpdateReleaseCmdResetUsersFlag)
			return false, err
		}
		// Get the reset teams flag
		resetTeams, err = cmd.Flags().GetBool(prjsUpdateReleaseCmdResetTeamsFlag)
		if err != nil {
			logFlagError(prjsUpdateReleaseCmdResetTeamsFlag)
			return false, err
		}
		// If requested, reset the users
		if resetUsers {
			resetSetRolesStruct(objRoles, rpapiv1.RPMemberUserType)
		}
		// If requested, reset the teams
		if resetTeams {
			resetSetRolesStruct(objRoles, rpapiv1.RPMemberTeamType)
		}
	}

	// Delete the potential artificial values from update mode
	utils.DeleteStringFromSlice(&relPrjTeams, flagUpdateKeepCurrentValue)
	utils.DeleteStringFromSlice(&relPrjUsers, flagUpdateKeepCurrentValue)

	// Detect the type of release done here (projects or images)
	roleWanted := "release-project"
	cmd.VisitParents(func(currCmd *cobra.Command) {
		if currCmd.Use == imagesCmdName {
			roleWanted = "release-image"
		}
	})

	// Manage the filling of the 'rpapiv1.RPObjectRoles' structure
	err = fillReleaseObjectRoles(relPrjTeams, relPrjUsers, objRoles, roleWanted, isCreation)
	if err != nil {
		return false, err
	}

	// Evaluate the boolean to return
	result := false
	if objRoles.Stack || (len(relPrjTeams) > 0) || (len(relPrjUsers) > 0) || resetUsers || resetTeams {
		result = true
	}

	return result, nil
}

// releaseProject release a standard project on the factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func releaseProject(cmd *cobra.Command, args []string) error {
	// Print the header
	RPPrinter.CustomPrintLn(prjsReleaseCmdAnsLine0)

	// Get the slug and the UUID of the project given by the user
	prjUUID, prjSlug, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Get the distribution
	distro, err := cmd.Flags().GetString(prjsReleaseCmdDistroFlag)
	if err != nil {
		logFlagError(prjsReleaseCmdDistroFlag)
		return err
	}

	// Get the version
	version, err := cmd.Flags().GetString(prjsReleaseCmdVersionFlag)
	if err != nil {
		logFlagError(prjsReleaseCmdVersionFlag)
		return err
	}

	// Check that we can find the distribution in projects
	RPPrinter.CustomPrint(prjsReleaseCmdAnsLine1)
	distroUUID, err := getProjectUUID(distro)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Prepare the structures to send to the backend
	prjRelease := rpapiv1.RPReleaseProject{
		DistributionID: distroUUID,
		Version:        version,
	}

	// Fill the project release structure with other value
	err = prepareRelProjectStruct(cmd, &prjRelease, true)
	if err != nil {
		return err
	}

	// Prepare the answer we will receive
	ansProject := rpapiv1.RPReleaseProjectFull{}

	// Send the request to the backend to release the project
	RPPrinter.CustomPrintF(prjsReleaseCmdAnsLine2, prjSlug)
	if err = bckAPI.ProjectRelease(prjUUID.String(), &prjRelease, &ansProject); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(ansProject, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		return nil
	}

	RPPrinter.CustomPrintF(prjsReleaseCmdAnsLine3, prjSlug)
	RPPrinter.ForceFlush()

	// Block until we receive a corresponding websocket event
	refProjectStable := false
	relProjectStable := false
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project addition
		if rEvent.Type == rpapiv1.EVTProjectStateChange {

			// Function handling the events
			rProject, err := rEvent.DecodeProject()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			// Check if this is one of the project we monitoring and if there are stable
			if utils.UUIDEqual(rProject.ID, prjUUID) && (rProject.State == rpapiv1.RPPrjStableState) {
				refProjectStable = true
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rProject, false, displayNothing)
			}

		} else if rEvent.Type == rpapiv1.EVTReleaseProjectStateChange {
			// Function handling the events
			rProjectRel, err := rEvent.DecodeReleaseProject()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			// Check if it is our release project
			if utils.UUIDEqual(rProjectRel.ID, ansProject.ID) && (rProjectRel.State == rpapiv1.RPPrjStableState) {
				relProjectStable = true
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rProjectRel, false, displayNothing)
			}

		} else if rEvent.Type == rpapiv1.EVTError {
			// Function handling the error, try with the reference project
			err = handleErrorEvent(prjUUID, rEvent, true)
			if err != nil {
				return err
			}

			// Function handling the error, try with the released project
			err = handleErrorEvent(ansProject.ID, rEvent, true)
			if err != nil {
				return err
			}
		}

		// Check if both projects are stable
		if refProjectStable && relProjectStable {
			// The release is done
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
			break
		}
	}

	// Manage the audience for the released project
	infoMessage := fmt.Sprintf(prjsReleaseCmdAnsLine4, prjSlug)
	err = manageAudienceUpdate(cmd, ansProject.ID, true, infoMessage, bckAPI.GetReleasedProjectAudience, bckAPI.UpdateReleasedProjectAudience)

	return err
}

// listReleasedProjects list the released projects from the factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func listReleasedProjects(cmd *cobra.Command, args []string) error {
	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Prepare the function needed by the printer to get Released Projects
	getRelProjectsCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		// Get projects list
		prjs := rpapiv1.RPReleaseProjects{}
		if err := bckAPI.ListReleasedProjectByPage(&prjs, pageNb, perPage); err != nil {
			return prjs.ReleaseProjects, printer.RPPrinterPaging{}, err
		}
		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&prjs)

		return prjs.ReleaseProjects, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: projectRelDisplayVerboseLightLineNb,
		GetObjCB:      getRelProjectsCB,
		PrintFancyCB:  displayRelProjects,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct)
}

// getReleasedProject display one released project from the factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getReleasedProject(cmd *cobra.Command, args []string) error {
	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the audience flag
	audience, err := cmd.Flags().GetBool(prjsGetReleaseCmdAudienceFlag)
	if err != nil {
		logFlagError(prjsGetReleaseCmdAudienceFlag)
		return err
	}

	// Get the UUID of the project given by the user
	prjUUID, _, err := genericGetRelPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Get the released project
	relPrj := rpapiv1.RPReleaseProjectFull{}
	if err := bckAPI.GetReleasedProject(prjUUID.String(), &relPrj); err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(relPrj, verbose, displayOneRelProject)
	if err != nil {
		return err
	}

	// Check if we need to display the audience
	if !audience {
		return nil
	}

	// Retrieve the audience for this released project
	relPrjAudience := rpapiv1.RPObjectRoles{}
	err = bckAPI.GetReleasedProjectAudience(prjUUID.String(), &relPrjAudience)
	if err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(relPrjAudience, verbose, printRelAudience)

	return err
}

// updateReleasedProject update a released project settings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func updateReleasedProject(cmd *cobra.Command, args []string) error {
	// Print the header
	RPPrinter.CustomPrintLn(prjsReleaseUpdateCmdAnsLine0)

	// Get the UUID of the released project given by the user
	prjUUID, prjSlug, err := genericGetRelPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Get the current values for the fields of the released project (directly from cache)
	currReleasedPrj := cachedRPRelProjectsByUUID[prjUUID]

	// Fill the project release structure with other value
	err = prepareRelProjectStruct(cmd, &currReleasedPrj, false)
	if err != nil {
		return err
	}

	// Get the "force-stable-state" flag
	forceStable, err := cmd.Flags().GetBool(prjsForceStateFlag)
	if err != nil {
		return err
	}
	// If the user wants it, force the project state at "stable"
	if forceStable {
		currReleasedPrj.State = rpapiv1.RPPrjStableState
	}

	// Prepare the answer we will receive
	ansProject := &rpapiv1.RPReleaseProjectFull{}

	// Send the request to the backend to release the project
	RPPrinter.CustomPrintF(prjsReleaseUpdateCmdAnsLine1, prjSlug)
	if err = bckAPI.UpdateReleasedProject(prjUUID.String(), &currReleasedPrj, ansProject); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(ansProject, false, displayNothing)

	// No need to wait for event for the moment, no async operation

	// Manage the audience for the released project
	infoMessage := fmt.Sprintf(prjsReleaseUpdateCmdAnsLine2, prjSlug)
	err = manageAudienceUpdate(cmd, ansProject.ID, false, infoMessage, bckAPI.GetReleasedProjectAudience, bckAPI.UpdateReleasedProjectAudience)

	return err
}

// deleteReleasedProject delete a released project from the factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func deleteReleasedProject(cmd *cobra.Command, args []string) error {
	// Print the header
	RPPrinter.CustomPrintLn(prjsDeleteReleaseCmdAnsLine0)

	// Get the UUID of the released project given by the user
	prjUUID, prjSlug, err := genericGetRelPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Request to delete the released project
	RPPrinter.CustomPrintF(prjsDeleteReleaseCmdAnsLine1, prjSlug)
	ansProject := &rpapiv1.RPReleaseProjectFull{}
	if err = bckAPI.DeleteReleasedProject(prjUUID.String(), ansProject); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(ansProject, false, displayNothing)

	return nil
}

// deployReleasedProject deploy a released project on boards
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func deployReleasedProject(cmd *cobra.Command, args []string) error {
	// Print the header
	RPPrinter.CustomPrintLn(prjsDeployReleaseCmdAnsLine0)

	// Get the UUID of the released project given by the user
	prjUUID, prjSlug, err := genericGetRelPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Get the boards list given by the user
	boardSlugsList, err := cmd.Flags().GetStringSlice(prjsDeployReleaseCmdBoardsFlag)
	if err != nil {
		logFlagError(prjsDeployReleaseCmdBoardsFlag)
		return err
	}

	// Check that at least one board is given
	if len(boardSlugsList) <= 0 {
		return errors.New(prjRelDeployNoBoardsSpecified)
	}

	// From the list of slugs, get the list of boards UUID
	deployStruct := rpapiv1.RPDeploy{BoardsID: []uuid.UUID{}}
	for _, tmpBoard := range boardSlugsList {
		// Convert the ID given by the user to an UUID
		tmpUUID, err := getBoardUUID(tmpBoard)
		if err != nil {
			return err
		}
		// Add inside the deployment structure
		deployStruct.BoardsID = append(deployStruct.BoardsID, tmpUUID)
	}

	// Request to deploy the released project on the boards
	RPPrinter.CustomPrintF(prjsDeployReleaseCmdAnsLine1, prjSlug)
	deployAns := rpapiv1.RPDeploymentFull{}
	if err = bckAPI.DeployReleasedProject(prjUUID.String(), &deployStruct, &deployAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(deployAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		return nil
	}

	// Wait for the end of deployment
	RPPrinter.CustomPrintF(prjsDeployReleaseCmdAnsLine2, prjSlug)
	RPPrinter.ForceFlush()

	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a released project deployment state change
		if rEvent.Type == rpapiv1.EVTDeployReleasedProjectStateChange {

			// Decode the structure in order to process the event afterwards
			rDeployment, err := rEvent.DecodeRelPrjDeployment()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			// Check if this deployment is the one we are monitoring and if it has reached a 'finished' state
			if utils.UUIDEqual(rDeployment.ID, deployAns.ID) && (rDeployment.Status == rpapiv1.RPDeploymentGlobalStatusFinished) {
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rDeployment, false, displayNothing)
				// The deployment is done
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				break
			}
		}
	}

	return nil
}

// getProjectBuildsStats display the builds statistics for the given
// project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getProjectBuildsStats(cmd *cobra.Command, args []string) error {
	// Get the UUID of the project given by the user
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Now request and display builds stats for this project
	return getAndDisplayBuildsStats(cmd, prjUUID, nil)
}

// getBuildsMatrixCmd display a builds matrix for the given
// project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getBuildsMatrixCmd(cmd *cobra.Command, args []string) error {
	// Get the UUID of the project given by the user
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Now request and display builds stats for this project
	return getAndDisplayBuildsMatrix(prjUUID)
}

// getProjectTestsStats display the tests statistics for the given
// project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getProjectTestsStats(cmd *cobra.Command, args []string) error {
	// Get the UUID of the project given by the user
	prjUUID, _, err := genericGetPrjUUIDAndSlug(args)
	if err != nil {
		return err
	}

	// Now request and display builds stats for this project
	return getAndDisplayTestsStats(cmd, prjUUID, uuid.Nil, uuid.Nil)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// genericGetPrjUUID allows a function to retrieve the UUID of the
// project targetted by the user.
//
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: uuid, slug, error -- the UUID of the project and an error if something went wrong, nil in the other case
func genericGetPrjUUIDAndSlug(args []string) (uuid.UUID, string, error) {
	// Get the project ID and return an error if there is nothing
	prjID, err := getFirstArg(args, prjIDNotSetMsg, prjIDRequiredMsg)
	if err != nil {
		return uuid.Nil, "", err
	}

	// Get the UUID for the project
	prjUUID, err := getProjectUUID(prjID)
	if err != nil {
		return uuid.Nil, "", err
	}

	// Get the slug of the project
	prjSlug, err := getPrjSlugFromUUID(prjUUID)

	return prjUUID, prjSlug, err
}

// genericGetRelPrjUUIDAndSlug allows a function to retrieve the UUID of the
// released project targeted by the user and its slug
//
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: uuid, slug, error -- the UUID of the project and an error if something went wrong, nil in the other case
func genericGetRelPrjUUIDAndSlug(args []string) (uuid.UUID, string, error) {
	// Get the project ID and return an error if there is nothing
	prjID, err := getFirstArg(args, prjIDNotSetMsg, prjIDRequiredMsg)
	if err != nil {
		return uuid.Nil, "", err
	}

	// Get the UUID for the project
	prjUUID, err := getRelProjectUUID(prjID)
	if err != nil {
		return uuid.Nil, "", err
	}

	// Get the slug of the project
	prjSlug, err := getRelPrjSlugFromUUID(prjUUID)

	return prjUUID, prjSlug, err
}

// getProjectFromCache allows a function to retrieve a project from our cache
//
// :param prjID: UUID of the project to retrieve
// :type prjID: uuid.UUID
// :returns: rpapiv1.RPProject, error -- RPProject found and an error if something went bad
func getProjectFromCache(prjID uuid.UUID) (rpapiv1.RPProject, error) {
	// Fill the cache if it is empty
	if (cachedRPProjectsByUUID == nil) || (runMode == "testing") {
		err := populateProjectsCache(nil)
		if err != nil {
			return rpapiv1.RPProject{}, err
		}
	}

	tmpProject, ok := cachedRPProjectsByUUID[prjID]
	if ok {
		// We found the project
		return tmpProject, nil
	}

	// If we reached this part of code, we did not find anything
	return rpapiv1.RPProject{}, fmt.Errorf(prjIdentifierNotFound, prjID.String())
}

// getProjectUUID is a private function allowing to retrieve the
// UUID of a project from a given identifier. This identifier can
// be the UUID itself as a string representation or the Slug of the
// project
//
// :param identifier: the identifier of the project
// :type identifier: string
// :returns: (uuid.UUID, error) -- the UUID of the project and an error if something went bad
func getProjectUUID(identifier string) (uuid.UUID, error) {

	// Optimization: pass through the cache in order to avoid several requests for Project list
	if (cachedRPProjectsByUUID == nil) || (cachedRPProjectsBySlug == nil) || (runMode == "testing") {
		err := populateProjectsCache(nil)
		if err != nil {
			return uuid.Nil, err
		}
	}

	// Check if we can find the identifier in the different map
	prjUUID, ok := cachedRPProjectsBySlug[identifier]
	if ok {
		// The identifier is the slug and we found the uuid !
		return prjUUID, nil
	}

	tmpUUID, err := uuid.FromString(identifier)
	if err == nil {
		_, ok = cachedRPProjectsByUUID[tmpUUID]
		if ok {
			// The identifier is the UUID and we found it !
			return tmpUUID, nil
		}
	}

	// If we reached this part of code, we did not find anything
	return uuid.Nil, fmt.Errorf(prjIdentifierNotFound, identifier)
}

// getRelProjectUUID is a private function allowing to retrieve the
// UUID of a released project from a given identifier. This identifier can
// be the UUID itself as a string representation or the Slug of the
// released project
//
// :param identifier: the identifier of the released project
// :type identifier: string
// :returns: (uuid.UUID, error) -- the UUID of the project and an error if something went bad
func getRelProjectUUID(identifier string) (uuid.UUID, error) {

	// Optimization: pass through the cache in order to avoid several requests for Project list
	if (cachedRPRelProjectsByUUID == nil) || (cachedRPRelProjectsBySlug == nil) || (runMode == "testing") {
		err := populateRelProjectsCache()
		if err != nil {
			return uuid.Nil, err
		}
	}

	// Check if we can find the identifier in the different map
	prjUUID, ok := cachedRPRelProjectsBySlug[identifier]
	if ok {
		// The identifier is the slug and we found the uuid !
		return prjUUID, nil
	}

	tmpUUID, err := uuid.FromString(identifier)
	if err == nil {
		_, ok = cachedRPRelProjectsByUUID[tmpUUID]
		if ok {
			// The identifier is the UUID and we found it !
			return tmpUUID, nil
		}
	}

	// If we reached this part of code, we did not find anything
	return uuid.Nil, fmt.Errorf(prjIdentifierNotFound, identifier)
}

// getPrjSlugFromUUID allows to get the slug of a project from its UUID
//
// :param prjID: ID of the project wanted
// :type prjID: uuid.UUID
// :returns: string, error -- the app's slug and an error if something went wrong, nil in the other case
func getPrjSlugFromUUID(prjID uuid.UUID) (string, error) {

	// Optimization: pass through the cache in order to avoid several requests for Project list
	if (cachedRPProjectsByUUID == nil) || (runMode == "testing") {
		err := populateProjectsCache(nil)
		if err != nil {
			return "", err
		}
	}

	// Check if we find the UUID
	tmpPrj, ok := cachedRPProjectsByUUID[prjID]
	if ok {
		// We found the project!
		return tmpPrj.Slug, nil
	}

	return "", fmt.Errorf(prjIdentifierNotFound, prjID.String())
}

// getRelPrjSlugFromUUID allows to get the slug of a released project from its UUID
//
// :param prjID: ID of the project wanted
// :type prjID: uuid.UUID
// :returns: string, error -- the project slug and an error if something went wrong, nil in the other case
func getRelPrjSlugFromUUID(prjID uuid.UUID) (string, error) {
	// Initialize the variable that we will use later
	slugFound := ""

	if cachedRPRelProjectsByUUID != nil {
		// Check if we find the UUID
		tmpPrj, ok := cachedRPRelProjectsByUUID[prjID]
		if ok {
			// We found the project!
			slugFound = tmpPrj.Slug
		}
	}

	// If the slug is still not found, request directly the backend
	if slugFound == "" {
		relProject := rpapiv1.RPReleaseProjectFull{}
		if err := bckAPI.GetReleasedProject(prjID.String(), &relProject); err != nil {
			return slugFound, err
		}
		slugFound = relProject.Slug
	}

	return slugFound, nil
}

// getDistAndArchFromPrj is a private function allowing to retrieve all the
// distributions and architectures from a given project and to return it in a
// RPProjectArchDist structure.
//
// :param projectIdentifier: the identifier of the project
// :type projectIdentifier: string
// :returns: (RPProjectArchDist, error) -- the structure containing arch and distros and an error if something went bad
func getDistAndArchFromPrj(projectIdentifier string) (RPProjectArchDist, error) {
	resStruct := RPProjectArchDist{}

	// Try to get the project UUID from the identifier
	prjUUID, err := getProjectUUID(projectIdentifier)
	if err != nil {
		// The project ID has not been found, try to find a released project
		relPrjUUID, err := getRelProjectUUID(projectIdentifier)
		if err != nil {
			return resStruct, err
		}
		// Get the corresponding referent project UUID
		prjUUID = cachedRPRelProjectsByUUID[relPrjUUID].ProjectReferenceID
	}

	// Retrieve the corresponding project
	project, err := getProjectFromCache(prjUUID)
	if err != nil {
		return resStruct, err
	}

	// Fill the distributions and architecture
	resStruct = fillArchDistFromProject(project)

	return resStruct, nil
}

// fillArchDistFromProject is a simple function that returns a RPProjectArchDist structure
// from a rpapiv1.RPProjectFull structure
//
// :param project: structure containing the arches and distributions to extract
// :type project: rpapiv1.RPProject
// :returns: RPProjectArchDist -- the structure containing arch and distros
func fillArchDistFromProject(project interface{}) RPProjectArchDist {
	// Declare the return value
	resStruct := RPProjectArchDist{}

	// Declare the structure
	prj := rpapiv1.RPProjectFull{}

	// Cast to the full type in all case but use "full" parameters as indicated by the user
	// We have to pass through the Marshal/Unmarshal tricks to not have problem during conversion from
	// RPProject -> RPProjectFull at runtime
	dataBytes, _ := json.Marshal(project)
	json.Unmarshal(dataBytes, &prj)

	// Fill the distributions
	for _, distro := range prj.Distributions {
		if distro.Set && distro.Mandatory {
			resStruct.MandDistros = append(resStruct.MandDistros, distro.Slug)
		} else if distro.Set && !distro.Mandatory {
			resStruct.OptDistros = append(resStruct.OptDistros, distro.Slug)
		}
	}

	// Fill the architectures
	for _, arch := range prj.Architectures {
		if arch.Set && arch.Mandatory {
			resStruct.MandArches = append(resStruct.MandArches, arch.ID)
		} else if arch.Set && !arch.Mandatory {
			resStruct.OptArches = append(resStruct.OptArches, arch.ID)
		}
	}

	return resStruct
}

// populateProjectsCache is a function that needs to be called when
// projects will be used in the user request. The function fill the
// global variable "cachedRPProjectsByUUID" and "cachedRPProjectsBySlug"
// from a request to the Redpesk backend.
//
// :param prjList: A pointer to the list of projects if the request
// has been already done. Nil if the request needs to be done to the backend
// :type prjList: *rpapiv1.RPProjects
// :returns: error -- An error if something went bad
func populateProjectsCache(prjList *rpapiv1.RPProjects) error {

	var tmpList *rpapiv1.RPProjects
	if prjList == nil {
		tmpList = &rpapiv1.RPProjects{}
		// Get the projects of all types
		if err := bckAPI.ProjectsListGetOfTypes(tmpList, []string{"standard", "external"}); err != nil {
			return err
		}
	} else {
		tmpList = prjList
	}

	// Create the map if it does not exists
	if (cachedRPProjectsByUUID == nil) || (runMode == "testing") {
		cachedRPProjectsByUUID = make(map[uuid.UUID]rpapiv1.RPProject)
	}

	// Create the map if it does not exists
	if (cachedRPProjectsBySlug == nil) || (runMode == "testing") {
		cachedRPProjectsBySlug = make(map[string]uuid.UUID)
	}

	// We got correctly the projects list, fill the cache now
	for _, project := range tmpList.Projects {

		// Copy each project in the cache
		tmpProject := rpapiv1.RPProject{}
		copier.Copy(&tmpProject, &project)
		cachedRPProjectsByUUID[project.ID] = tmpProject

		// Fill the other cache
		cachedRPProjectsBySlug[project.Slug] = project.ID
	}

	return nil
}

// populateRelProjectsCache is a function that needs to be called when
// released projects will be used in the user request. The function fill the
// global variables "cachedRPRelProjectsByUUID" and "cachedRPRelProjectsBySlug"
// from a request to the Redpesk backend.
//
// :returns: error -- An error if something went bad
func populateRelProjectsCache() error {
	// Get the release projects
	tmpList := &rpapiv1.RPReleaseProjects{}
	if err := bckAPI.ListReleasedProject(tmpList); err != nil {
		return err
	}

	// Create the map if it does not exists
	if (cachedRPRelProjectsByUUID == nil) || (runMode == "testing") {
		cachedRPRelProjectsByUUID = make(map[uuid.UUID]rpapiv1.RPReleaseProject)
	}

	// Create the map if it does not exists
	if (cachedRPRelProjectsBySlug == nil) || (runMode == "testing") {
		cachedRPRelProjectsBySlug = make(map[string]uuid.UUID)
	}

	// We got correctly the projects list, fill the cache now
	for _, project := range tmpList.ReleaseProjects {
		// Copy each project in the cache
		tmpProject := rpapiv1.RPReleaseProject{}
		copier.Copy(&tmpProject, &project)
		cachedRPRelProjectsByUUID[project.ID] = tmpProject
		// Fill the other cache
		cachedRPRelProjectsBySlug[project.Slug] = project.ID
	}

	return nil
}

// downloadExportedPrj download the archive corresponding to an
// exported project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param prjUUID: UUID of the project as a string
// :type prjUUID: uuid.UUID
// :returns: error -- an error if a problem has occurred
func downloadExportedPrj(cmd *cobra.Command, prjUUID uuid.UUID) error {
	// Retrieve the directory flag
	downloadDir, err := cmd.Flags().GetString(prjsExportCmdDirFlag)
	if err != nil {
		logFlagError(prjsExportCmdDirFlag)
		return err
	}

	// If we have reached here, the export went well, now download the archive
	RPPrinter.CustomPrint(prjsExportStartCmdAnsLine3)

	fullFilename, err := bckAPI.ProjectExportDownload(prjUUID.String(), downloadDir)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// The download went well
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	RPPrinter.CustomColorPrintF(printer.ColorCyan, prjsExportStartCmdAnsLine4, fullFilename)

	return nil
}

// populateRolesCache is a function that needs to be called when
// roles will be used in the user request. The function fill the
// global variable "cachedRedpeskRoles" from a request to the
// Redpesk backend.
//
// :param prjUUID: the UUID of the concerned project as a string
// :type prjUUID: string
// :returns: error -- An error if something went bad
func populateRolesCache(prjUUID string) error {

	// Retrieve the roles available for the given project
	rolesAvailable := rpapiv1.RPRoleProjects{}
	if err := bckAPI.ProjectGetMembersRoles(prjUUID, &rolesAvailable); err != nil {
		return err
	}

	// We got correctly the roles list, fill the cache now
	for _, role := range rolesAvailable {
		tmpRole := rpapiv1.RPRole{
			ID:    role.RoleID,
			Label: role.RoleLabel,
		}
		cachedRPPrjRoles = append(cachedRPPrjRoles, tmpRole)
	}

	return nil
}

// getRolesStruct is a function allowing to get all the structure of
// several roles. The cache needs to be populated in order to use this function!
//
// :param identifiers: Array of identifiers that is interesting for the calling function
// :type identifiers: []string
// :returns: []rpapiv1.RPRole, error -- slice of RPRole wanted and an error if something went wrong
func getRolesStruct(identifiers []string) ([]rpapiv1.RPRole, error) {
	var foundIndex int

	// Initialize roles tab
	resRoles := []rpapiv1.RPRole{}

	// Go through the list of identifiers
	for _, identifier := range identifiers {
		foundIndex = -1

		// Go through the list of roles in the cache
		for index, rpRole := range cachedRPPrjRoles {
			// Check if the current identifier is the one we are looking for
			if (rpRole.ID.String() == identifier) || (rpRole.Label == identifier) {
				foundIndex = index
				break
			}
		}

		// Check that we found the identifier
		if foundIndex == -1 {
			return resRoles, fmt.Errorf(prjRoleIdentifierNotFound, identifier)
		}

		// If we get here, the identifier has been found, fill the slice !
		resRoles = append(resRoles, cachedRPPrjRoles[foundIndex])
	}

	return resRoles, nil
}

// getGuestIDInfo is a function allowing to get all the info about
// a guest (especially UUID and pseudo)
//
// :param identifier: an identifier of the guest
// :type identifier: string
// :param prjUUID: UUID of the project related to the request
// :type prjUUID: uuid.UUID
// :returns: rpapiv1.RPMember, error -- the member's structure and an error if something went wrong
func getGuestInfo(identifier string, prjUUID uuid.UUID) (rpapiv1.RPMember, error) {
	uuidIndex := -1
	memberResult := rpapiv1.RPMember{}

	// Get the list of the guest available on the Redpesk instance
	invitList := rpapiv1.RPMembers{}
	if err := bckAPI.ProjectGetMembersToInvite(prjUUID.String(), &invitList); err != nil {
		return memberResult, err
	}

	// Goes through the list to find the right UUID
	for index, guest := range invitList {
		// Check if the UUID or the Pseudo is the one we are looking for
		if (guest.ID.String() == identifier) || (guest.Pseudo == identifier) {
			uuidIndex = index
			break
		}
	}

	if uuidIndex == -1 {
		return memberResult, fmt.Errorf(prjGuestIdentifierNotFound, identifier)
	}

	return invitList[uuidIndex], nil
}

// getMemberInfo is a function allowing to get all the info about
// a project member (especially UUID and pseudo)
//
// :param identifier: an identifier of the guest
// :type identifier: string
// :param prjUUID: UUID of the project related to the request
// :type prjUUID: uuid.UUID
// :returns: rpapiv1.RPMember, error -- the member's structure and an error if something went wrong
func getMemberInfo(identifier string, prjUUID uuid.UUID) (rpapiv1.RPMember, error) {
	uuidIndex := -1
	memberResult := rpapiv1.RPMember{}

	// Get the list of the project members
	membersAvailable := rpapiv1.RPMembers{}
	if err := bckAPI.ProjectMembersGet(prjUUID.String(), &membersAvailable); err != nil {
		return memberResult, err
	}

	// Goes through the list to find the right UUID
	for index, member := range membersAvailable {
		// Check if the UUID or the Pseudo is the one we are looking for
		if (member.ID.String() == identifier) || (member.Pseudo == identifier) {
			uuidIndex = index
			break
		}
	}

	if uuidIndex == -1 {
		return memberResult, fmt.Errorf(prjMemberIdentifierNotFound, identifier)
	}

	return membersAvailable[uuidIndex], nil
}

// fillReleaseObjectRoles is a function allowing to fill all the object roles
// needed to change the audience of a project release
//
// :param teams: List of team slugs given by the user
// :type teams: []string
// :param users: List of user slugs given by the user
// :type users: []string
// :param objRoles: RPObjectRoles structure to fill
// :type objRoles: *rpapiv1.RPObjectRoles
// :param roleWanted: Type of role wanted for this project
// :type roleWanted: string
// :param isCreation: Boolean allowing to manage specific action according to creation/update mode
// :type isCreation: bool
// :returns:error -- an error if something went wrong
func fillReleaseObjectRoles(teams, users []string, objRoles *rpapiv1.RPObjectRoles, roleWanted string, isCreation bool) error {

	// Firstly, retrieve all the UUID for given users and teams and return an error if we cannot find one
	identitiesUUID := []string{}

	// First, the teams
	for _, team := range teams {
		tmpUUID, err := getTeamUUID(team)
		if err != nil {
			return err
		}
		identitiesUUID = append(identitiesUUID, tmpUUID.String())
	}

	// Then, the users
	for _, user := range users {
		tmpUUID, err := getUserUUID(user)
		if err != nil {
			return err
		}
		identitiesUUID = append(identitiesUUID, tmpUUID.String())
	}

	// Then check that each given identity ID exists in the object roles structure
	for _, identity := range identitiesUUID {
		idRolesList, ok := objRoles.IdentityRolesLists[identity]
		if !ok {
			// Return an error
			errMsg := fmt.Sprintf(prjRelAudienceNotAvailableMsg, identity)
			return errors.New(errMsg)
		}

		// The identity has been retrieved, look in the available role to be sure it has the wanted role ("release-project" for example)
		indexRoleWanted := -1
		for index, role := range *(idRolesList.AvailableRoles) {
			if role.Object == roleWanted {
				indexRoleWanted = index
				break
			}
		}
		if indexRoleWanted < 0 {
			errMsg := fmt.Sprintf(prjRelRoleNotAvailableMsg, roleWanted, identity)
			return errors.New(errMsg)
		}

		// If we reach here, the wanted role is available
		*idRolesList.SetRoles = append(*idRolesList.SetRoles, (*idRolesList.AvailableRoles)[indexRoleWanted])
	}

	return nil
}

// resetSetRolesStruct is a function allowing to reset the roles set for the audience
// in a rpapiv1.RPObjectRoles structure
//
// :param objRoles: Pointer to the rpapiv1.RPObjectRoles structure to modify
// :type objRoles: *rpapiv1.RPObjectRoles
func resetSetRolesStruct(objRoles *rpapiv1.RPObjectRoles, typeToReset rpapiv1.RPMemberType) {
	for _, rolesList := range objRoles.IdentityRolesLists {
		if (rolesList.Type == typeToReset) && (len(*rolesList.SetRoles) > 0) {
			rolesList.SetRoles = &([]rpapiv1.RPRole{})
		}
	}
}

// manageAudienceUpdate is a function allowing to manage the update of the audience of a
// released project/image
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param objectID: UUID of the released project/image
// :type objectID: uuid.UUID
// :param isCreation: Indicate if we are creating the object
// :type isCreation: bool
// :param infoMessage: Message to display to the user when managing the audience
// :type infoMessage: string
// :returns: error -- an error if something went wrong
func manageAudienceUpdate(cmd *cobra.Command, objectID uuid.UUID, isCreation bool, infoMessage string,
	getAudienceCB func(string, *rpapiv1.RPObjectRoles) error,
	updateAudienceCB func(string, *rpapiv1.RPObjectRoles, *rpapiv1.RPObjectRoles) error) error {

	// Get the available audience
	relPrjAudience := rpapiv1.RPObjectRoles{}
	err := getAudienceCB(objectID.String(), &relPrjAudience)
	if err != nil {
		return err
	}

	// Do the audience releasing if the user wants so
	audienceGiven, err := prepareRPObjectRolesStruct(cmd, &relPrjAudience, isCreation)
	if err != nil {
		return err
	}

	if audienceGiven {
		RPPrinter.CustomPrint(infoMessage)

		// Launch the request
		answer := rpapiv1.RPObjectRoles{}
		err = updateAudienceCB(objectID.String(), &relPrjAudience, &answer)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}
		// Display something to the user if in raw mode
		RPPrinter.CustomPrintObject(answer, false, displayNothing)

		// Display OK for the other mode
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}

	return nil
}

// prjGetUpdatesNeeded is a function allowing to check what kind of update is needed
// for the project: either a classic update, or a build type update, or both
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :returns: bool, bool -- classicUpdateNeeded, bldTypeUpdateNeeded
func prjGetUpdatesNeeded(cmd *cobra.Command) (bool, bool) {
	classicUpdateNeeded := false
	bldTypeUpdateNeeded := false

	// Function saving flag set
	activatedFlags := []string{}
	saveFlag := func(f *pflag.Flag) {
		activatedFlags = append(activatedFlags, f.Name)
	}

	// Get the flags activated by the user
	cmd.Flags().Visit(saveFlag)

	// Loop over the args
	for _, flag := range activatedFlags {
		if (flag == prjsUpdateCmdBldTypeFlag) || (flag == prjsUpdateCmdBldTypeFlagAlias) || (flag == prjsUpdateCmdBldTypeAppsFlag) {
			bldTypeUpdateNeeded = true
		} else {
			classicUpdateNeeded = true
		}
	}

	return classicUpdateNeeded, bldTypeUpdateNeeded
}

// waitForStableProject is a blocking function that wait for an event
// about the given project, with a stable status
//
// :param prjUUID: UUID of the project we want to monitor
// :type prjUUID: uuid.UUID
// :param evtToWatch: Type of event to watch
// :type evtToWatch: string
// :returns: error -- an error if something went wrong
func waitForStableProject(prjUUID uuid.UUID, evtToWatch string) error {
	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project addition
		if rEvent.Type == evtToWatch {
			// Function handling the events
			rProject, err := rEvent.DecodeProject()
			if err != nil {
				return err
			}

			if utils.UUIDEqual(rProject.ID, prjUUID) && (rProject.State == rpapiv1.RPPrjStableState) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rProject, false, displayNothing)
				break
			}
		}
	}

	return nil
}

// changeBuildType is a function that requests the backend to change the build type
// for some apps or for all of them
//
// :param bldType: New build type to set
// :type bldType: string
// :param prjUUID: UUID of the project containing the apps to change
// :type prjUUID: uuid.UUID
// :param appIDs: List of applications ID we want to update, if empty, all the apps are updated
// :type appIDs: []string
// :param ans: rpapiv1.RPProjectFull structure to fill with the backend answer
// :type ans: *rpapiv1.RPProjectFull
// :returns: error -- an error if something went wrong
func changeBuildType(bldType string, prjUUID uuid.UUID, appIDs []string, prjAns *rpapiv1.RPProjectFull) error {
	// Initialize the structure to send to the backend
	chgBldTypeStruct := rpapiv1.RPChangeBuildType{
		BuildType: rpapiv1.RPAppBuildMethod(bldType),
	}

	// Get the list of applications for the given project
	apps := rpapiv1.RPApplications{}
	if err := bckAPI.ApplicationsListGetOfTypes(&apps, prjUUID.String(), []string{}); err != nil {
		return err
	}

	// Check if the user wants to update all the apps or only few of them
	if len(appIDs) > 0 {
		// Initialize the sub-structure
		chgBldTypeStruct.Applications = []rpapiv1.RPAppChangeBuildType{}

		// Now, loop on all the apps
		for _, app := range apps.Apps {
			if utils.ContainsString(appIDs, app.Slug) || utils.ContainsString(appIDs, app.ID.String()) {
				// Add the UUID of the app to update
				chgBldTypeStruct.Applications = append(chgBldTypeStruct.Applications, rpapiv1.RPAppChangeBuildType{ApplicationID: app.ID})
			}
		}
	}

	// Now request the backend to change the build type for the project apps
	return bckAPI.ProjectChangeBuildType(prjUUID.String(), chgBldTypeStruct, prjAns)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// displayProjects is a private function allowing to display
// a slice of projects in the console
//
// :param projects: tab of projects to display
// :type projects: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayProjects(projects interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	prjs := projects.([]rpapiv1.RPProject)

	// Display result
	for index, project := range prjs {
		if verbose && (index > 0) {
			fmt.Fprintln(writer)
		} else if !verbose && index == 0 {
			manageHeaderPrint(prjsListCmdHeader, writer)
		}
		// Print this project
		_printOneProject(project, verbose, false, false, writer)
	}
	return nil
}

// displayOneProject is a private function allowing to display
// one project in the console
//
// :param prj: Project to display
// :type prj: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :returns: error -- an error if something went bad
func displayOneProject(prj interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Print the project
	_printOneProject(prj, verbose, true, true, writer)
	return nil
}

// _printOneProject is a private function allowing to print one project in the console
//
// :param prj: project to print (can be rpapiv1.RPProjectFull or rpapiv1.RPProject)
// :type prj: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isAlone: True if there is only one project to display in the context where this function is called
// :type isAlone: bool
// :param isFull: True if we need to display the Full object, false in the other case
// :type isFull: bool
// :param writer: Writer to used to print the project information
// :type writer: *tabwriter.Writer
func _printOneProject(prj interface{}, verbose, isAlone, isFull bool, writer *tabwriter.Writer) {
	var typeStr string
	prjToPrint := rpapiv1.RPProjectFull{}

	// Cast to the full type in all case but use "full" parameters as indicated by the user
	// We have to pass through the Marshal/Unmarshal tricks to not have problem during conversion from
	// RPProject -> RPProjectFull at runtime
	dataBytes, _ := json.Marshal(prj)
	json.Unmarshal(dataBytes, &prjToPrint)

	// Create the string for type
	// Including the supported types objs in project and the main type
	supportedTypesTab := []string{}
	supportedTypesTab = append(supportedTypesTab, string(prjToPrint.Type))

	if prjToPrint.Type == "external" {
		supportedTypesTab[len(supportedTypesTab)-1] = "distro (external)"
	}

	if prjToPrint.SupportedObjects.Applications {
		supportedTypesTab = append(supportedTypesTab, "applications")
	}
	if prjToPrint.SupportedObjects.Images {
		supportedTypesTab = append(supportedTypesTab, "images")
	}
	if prjToPrint.SupportedObjects.Containers {
		supportedTypesTab = append(supportedTypesTab, "containers")
	}
	typeToDisplay := strings.Join(supportedTypesTab, "-")

	// Create the string for version
	versionToDisplay := prjToPrint.Version
	if versionToDisplay == "" {
		versionToDisplay = mainNotSetMsg
	}

	if verbose {
		// Create the string containing all the distributions
		distributions := ""
		for _, distro := range prjToPrint.Distributions {
			if distro.Set {
				if distributions != "" {
					distributions += ", "
				}
				if distro.Mandatory {
					typeStr = " [Mand]"
				} else {
					typeStr = " [Opt]"
				}
				distributions += (distro.Slug + typeStr)
			}
		}

		// Create the string containing all the architectures
		architectures := ""
		for _, arch := range prjToPrint.Architectures {
			if arch.Set {
				if architectures != "" {
					architectures += ", "
				}
				if arch.Mandatory {
					typeStr = " [Mand]"
				} else {
					typeStr = " [Opt]"
				}
				architectures += (arch.ID + typeStr)
			}
		}

		// Create the string containing the inheritances (if full)
		inheritances := ""
		accessTokensStr := ""
		if isFull {
			// Manage the inheritance
			for _, prjInherit := range prjToPrint.Inheritances {
				// Initialize local variable
				tmpUUID := prjInherit.ID
				slugToAdd := ""

				// Check that the UUID exists in the cache (standard project)
				if prj, ok := cachedRPProjectsByUUID[tmpUUID]; ok {
					// The UUID exists!
					slugToAdd = prj.Slug
				} else {
					// Check if the inherited project is a released project
					relPrjSlug, err := getRelPrjSlugFromUUID(tmpUUID)
					if err == nil {
						slugToAdd = relPrjSlug
					}
					// SILENT ERROR here
				}

				// Add the parent project if it exists
				if slugToAdd != "" {
					if inheritances != "" {
						inheritances += ", "
					}
					inheritances += fmt.Sprintf(prjInheritanceStr, slugToAdd, prjInherit.Priority)
				}
			}
			// Manage the token(s) lines
			for index, token := range prjToPrint.ProjectAccessTokens {
				if index > 0 {
					accessTokensStr += ", "
				}
				accessTokensStr += token.JWT
			}
		}

		// Create the string that can potentially be empty
		description := prjToPrint.Description
		if description == "" {
			description = mainNotSetMsg
		}

		// Create the string for supported objects
		supportedObjSlice := []string{}
		if prjToPrint.SupportedObjects.Applications {
			supportedObjSlice = append(supportedObjSlice, "applications")
		}
		if prjToPrint.SupportedObjects.Images {
			supportedObjSlice = append(supportedObjSlice, "images")
		}
		if prjToPrint.SupportedObjects.Containers {
			supportedObjSlice = append(supportedObjSlice, "containers")
		}
		supportedObj := strings.Join(supportedObjSlice, ", ")

		// Printing the result on stdout
		fmt.Fprintln(writer, prjsListCmdVerboseLine0, prjToPrint.ID.String())
		fmt.Fprintln(writer, prjsListCmdVerboseLine1, prjToPrint.Name)
		fmt.Fprintln(writer, prjsListCmdVerboseLine6, prjToPrint.Slug)
		fmt.Fprintln(writer, prjsListCmdVerboseLine11, versionToDisplay)
		fmt.Fprintln(writer, prjsListCmdVerboseLine10, typeToDisplay)
		fmt.Fprintln(writer, prjsListCmdVerboseLine14, supportedObj)
		fmt.Fprintln(writer, prjsListCmdVerboseLine8, description)
		// If the project inherit from others, print them there
		if inheritances != "" {
			fmt.Fprintln(writer, prjsListCmdVerboseLine7, inheritances)
		}
		fmt.Fprintln(writer, prjsListCmdVerboseLine4, distributions)
		fmt.Fprintln(writer, prjsListCmdVerboseLine5, architectures)
		fmt.Fprintln(writer, prjsListCmdVerboseLine2, prjToPrint.CreatedAt.String())
		fmt.Fprintln(writer, prjsListCmdVerboseLine3, prjToPrint.UpdatedAt.String())
		fmt.Fprintln(writer, prjsListCmdVerboseLine9, prjToPrint.Bookmarked)
		fmt.Fprintln(writer, prjsListCmdVerboseLine12, prjToPrint.ShortReleaseNaming)
		// If we have an access token, print it here
		if accessTokensStr != "" {
			fmt.Fprintln(writer, prjsListCmdVerboseLine15, accessTokensStr)
		}

	} else {
		if isAlone {
			fmt.Fprintln(writer, prjsListCmdHeader)
		}
		fmt.Fprintln(writer, prjToPrint.ID.String(), "\t", prjToPrint.Slug, "\t", prjToPrint.Name, "\t", versionToDisplay, "\t", typeToDisplay)
	}
}

// displayProjectRepo is a private function allowing to display
// the repos for a project per distribution
//
// :param repos: Slice of available repositories for the project
// :type repos: *rpapiv1.RPRepos
// :param jwtAccessToken: Access token needed to access the repos
// :type jwtAccessToken: string
func displayProjectRepo(repos *rpapiv1.RPRepos, jwtAccessToken string) {
	reposURLStr := "\n\n" + prjsGetCmdAnsRepoURL + "\n" + prjsGetCmdAnsLineDash + "\n"
	reposCurlStr := "\n\n" + prjsGetCmdAnsLine0 + "\n" + prjsGetCmdAnsLineDash + "\n"

	// Loop over the available repositories
	for _, repo := range *repos {
		// Get the URL
		repoURL := repo.Build

		// Delete the "basearch" since we don't need them
		repoURL = strings.ReplaceAll(repoURL, "$basearch", "")
		repoURL = strings.ReplaceAll(repoURL, "\\$basearch", "")

		// Create the repoURL with token
		repoURLWithToken := repoURL
		if len(jwtAccessToken) > 0 {
			repoURLWithToken = fmt.Sprintf("%s?token=%s", repoURL, jwtAccessToken)
		}

		// Update the repos URL string
		reposURLStr = reposURLStr + "\n" + repo.Distro.Slug + "\n  -> " + repoURLWithToken + "\n"

		// Create the curl command
		repoURL = strings.TrimSuffix(repoURL, "/")
		curlCmd := fmt.Sprintf(repoCurlAddTemplate, repoURL, jwtAccessToken)

		// Update the repos curl string
		reposCurlStr = reposCurlStr + "\n" + repo.Distro.Slug + "\n  -> " + curlCmd + "\n"
	}

	// Now print the strings
	RPPrinter.CustomPrint(reposURLStr)
	RPPrinter.CustomPrint(reposCurlStr)
	RPPrinter.CustomPrintLn()
}

// displayMembers is a private function allowing to display
// a slice of project members in the console
//
// :param members: tab of members to display
// :type members: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayMembers(members interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	mbrs := members.(rpapiv1.RPMembers)

	// Display result
	first := true

	for _, member := range mbrs {
		if verbose {
			if !first {
				fmt.Fprintln(writer, prjMembersGetCmdAnsLine0)
			}

			// Print this member
			_printOneMember(&member, verbose, false, writer)

		} else {
			if first {
				fmt.Fprintln(writer, prjMembersGetCmdHeader)
			}

			// Print this member
			_printOneMember(&member, verbose, false, writer)
		}
		first = false
	}

	return nil
}

// _printOneMember is a private function allowing to print one member in the console
//
// :param member: pointer to the member to print
// :type member: *rpapiv1.RPMember
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isAlone: True if there is only one member to display in the context where this function is called
// :type isAlone: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
func _printOneMember(member *rpapiv1.RPMember, verbose bool, isAlone bool, writer *tabwriter.Writer) {

	// Aggregate the roles
	rolesStr := ""
	for index, role := range member.Roles {
		// Check that this role is known
		if role.Label != "" {
			// Add a comma only if it is not the first iteration
			if index > 0 {
				rolesStr = rolesStr + ", "
			}
			rolesStr = rolesStr + role.Label
		}
	}

	if verbose {

		// Create the string containing the results
		fmt.Fprintln(writer, prjMembersGetCmdAnsLine1, member.ID.String())
		fmt.Fprintln(writer, prjMembersGetCmdAnsLine2, member.Pseudo)

		// Display the roles
		fmt.Fprintln(writer, prjMembersGetCmdAnsLine3, rolesStr)

	} else {
		if isAlone {
			fmt.Fprintln(writer, prjMembersGetCmdHeader)
		}
		fmt.Fprintln(writer, member.Pseudo, "\t", rolesStr)
	}
}

// displayExportState is a function allowing to print the state of a project export
//
// :param prjState: State to print
// :type prjState: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayExportState(prjState interface{}, verbose bool, writer *tabwriter.Writer) error {
	localState := prjState.(rpapiv1.RPExportProjectState)
	if verbose {
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine5, localState.ProjectID.String())
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine6, localState.StartDate.String())
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine7, localState.EndDate.String())
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine8, string(localState.Status))
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine9, localState.ElapsedAt.String())
		errToPrint := "[Not set]"
		if localState.Error != "" {
			errToPrint = localState.Error
		}
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine10, errToPrint)
	} else {
		// Uniquely print the status
		fmt.Fprintln(writer, string(localState.Status))
	}
	return nil
}

// displayImportState is a function allowing to print the state of a project import
//
// :param prjState: State to print
// :type prjState: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayImportState(prjState interface{}, verbose bool, writer *tabwriter.Writer) error {
	localState := prjState.(rpapiv1.RPImportProjectState)
	if verbose {
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine5, localState.FuturePrjID.String())
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine6, localState.StartDate.String())
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine7, localState.EndDate.String())
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine8, string(localState.Status))
		errToPrint := "[Not set]"
		if localState.Error != "" {
			errToPrint = localState.Error
		}
		fmt.Fprintln(writer, prjsExportStartCmdAnsLine10, errToPrint)
	} else {
		// Uniquely print the status
		fmt.Fprintln(writer, string(localState.Status))
	}
	return nil
}

// displayRelProjects is a private function allowing to display
// a slice of released projects in the console
//
// :param projects: tab of projects to display
// :type projects: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayRelProjects(projects interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	relPrjs := projects.([]rpapiv1.RPReleaseProject)

	// Display result
	for index, project := range relPrjs {
		if verbose && (index > 0) {
			fmt.Fprintln(writer)
		} else if !verbose && index == 0 {
			manageHeaderPrint(prjsListReleasedCmdHeader, writer)
		}
		// Print this project
		_printOneProjectReleased(project, verbose, false, writer)
	}
	return nil
}

// displayOneRelProject is a private function allowing to display
// one released project in the console
//
// :param prj: Project to display
// :type prj: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :returns: error -- an error if something went bad
func displayOneRelProject(prj interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Print the project
	_printOneProjectReleased(prj, verbose, true, writer)
	return nil
}

// _printOneProjectReleased is a private function allowing to print one released project in the console
//
// :param prj: project to print (can be rpapiv1.RPProjectFull or rpapiv1.RPProject)
// :type prj: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isAlone: True if there is only one project to display in the context where this function is called
// :type isAlone: bool
// :param writer: Writer to used to print the project information
// :type writer: *tabwriter.Writer
func _printOneProjectReleased(prjRel interface{}, verbose, isAlone bool, writer *tabwriter.Writer) {
	releasedPrj := rpapiv1.RPReleaseProjectFull{}

	// Cast to the full type in all case but use "full" parameters as indicated by the user
	// We have to pass through the Marshal/Unmarshal tricks to not have problem during conversion from
	// RPProject -> RPProjectFull at runtime
	dataBytes, _ := json.Marshal(prjRel)
	json.Unmarshal(dataBytes, &releasedPrj)

	if verbose {
		description := releasedPrj.Description
		if description == "" {
			description = mainNotSetMsg
		}
		// Display the project in verbose mode

		// Display the release project ID details
		fmt.Fprintln(writer, prjsListCmdVerboseLine0, releasedPrj.ID.String())
		fmt.Fprintln(writer, prjsListCmdVerboseLine1, releasedPrj.Name)
		fmt.Fprintln(writer, prjsListCmdVerboseLine6, releasedPrj.Slug)

		// Then, display the project reference
		fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine0, releasedPrj.ProjectReferenceID.String())
		fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine1, releasedPrj.ProjectReferenceSlug)
		fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine2, releasedPrj.DistributionID.String())

		// After, display more details
		fmt.Fprintln(writer, prjsListCmdVerboseLine11, releasedPrj.Version)
		fmt.Fprintln(writer, prjsListCmdVerboseLine8, description)
		fmt.Fprintln(writer, prjsListCmdVerboseLine2, releasedPrj.CreatedAt.String())
		fmt.Fprintln(writer, prjsListCmdVerboseLine3, releasedPrj.UpdatedAt.String())
		fmt.Fprintln(writer, prjsListCmdVerboseLine12, releasedPrj.ShortReleaseNaming)

		if isAlone {
			// Even more details  if full data
			if releasedPrj.FullData {
				// Manage the token(s) lines
				accessTokensStr := ""
				for index, token := range releasedPrj.ProjectAccessTokens {
					if index > 0 {
						accessTokensStr += ", "
					}
					accessTokensStr += token.JWT
				}
				// If we have an access token, print it here
				if accessTokensStr != "" {
					fmt.Fprintln(writer, prjsListCmdVerboseLine15, accessTokensStr)
				}
			}

			// Finally, display the apps
			fmt.Fprintln(writer)
			fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine3)

			for _, app := range releasedPrj.Apps {
				fmt.Fprintln(writer)
				fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine4, app.ApplicationID.String())
				fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine5, app.PackageName)
				fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine6, app.ShaGitPkg)
				fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine7, app.Branch)
				fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine8, app.BuildID)
			}

			if releasedPrj.FullData {
				fmt.Fprintln(writer)
				fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine9)
				fmt.Fprintln(writer)
				// Create the json string
				jsonObj, _ := json.MarshalIndent(releasedPrj.ReleaseInfo, "", "    ")
				// Print the pretty json string
				fmt.Fprintln(writer, string(jsonObj))
			}
		}

	} else {
		// Display essential info
		if isAlone {
			fmt.Fprintln(writer, prjsListReleasedCmdHeader)
		}
		fmt.Fprintln(writer, releasedPrj.ID.String(), "\t", releasedPrj.Slug, "\t", releasedPrj.Version, "\t", releasedPrj.ProjectReferenceSlug)
	}
}

// printRelAudience is a private function allowing to print the audience of a released project
//
// :param audience: audience to print (rpapiv1.RPObjectRoles)
// :type audience: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the project information
// :type writer: *tabwriter.Writer
func printRelAudience(audience interface{}, verbose bool, writer *tabwriter.Writer) error {
	audienceObj := audience.(rpapiv1.RPObjectRoles)
	if verbose {
		// Work on the values we will need to display
		factoryStr := "no"
		teamsSlice := []string{}
		usersSlice := []string{}

		// Check if the release has been done for the factory
		if audienceObj.Stack {
			factoryStr = "yes"
		}

		// Loop on the possible audience (teams or users)
		for _, rolesList := range audienceObj.IdentityRolesLists {
			if (len(*rolesList.SetRoles) > 0) && (rolesList.Type == rpapiv1.RPMemberTeamType) {
				teamsSlice = append(teamsSlice, rolesList.PseudoOrSlug)
			} else if (len(*rolesList.SetRoles) > 0) && (rolesList.Type == rpapiv1.RPMemberUserType) {
				usersSlice = append(usersSlice, rolesList.PseudoOrSlug)
			}
		}

		// Create the string to display for teams and users
		teamsString := strings.Join(teamsSlice, ", ")
		usersString := strings.Join(usersSlice, ", ")
		if teamsString == "" {
			teamsString = "[Not set]"
		}
		if usersString == "" {
			usersString = "[Not set]"
		}

		fmt.Fprintln(writer)
		fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine10)
		fmt.Fprintln(writer)
		fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine11, factoryStr)
		fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine12, teamsString)
		fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine13, usersString)
	}
	return nil
}
