/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"os"
	"path"
	"regexp"
	"strings"
	"testing"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpbckmock"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/stretchr/testify/assert"
)

const (
	// Projects
	projectDisplayNoVerboseColNb    = 5
	projectDisplayVerboseFullLineNb = 14

	// Projects member
	prjMembersDisplayNoVerboseColNb = 2
	prjMembersDisplayVerboseLineNb  = 3

	// Projects export
	prjExportDisplayVerboseLineNb = 6

	// Projects import
	prjImportDisplayVerboseLineNb = 5

	// Projects released
	projectRelDisplayNoVerboseColNb            = 4
	projectRelDisplayVerboseFullLineNb         = 45
	projectRelDisplayVerboseFullAudienceLineNb = projectRelDisplayVerboseFullLineNb + 6

	// Projects build
	projectBuildsStatsVerboseColNb    = 7
	projectBuildsStatsNoVerboseLineNb = 5

	// Projects test
	projectTestsStatsVerboseColNb    = 7
	projectTestsStatsNoVerboseLineNb = 5

	// Export project directory
	exportImportProjectsDirectory = "/tmp/rp-cli-export-import-project/"

	// Import archive content
	importArchiveContent = "THIS IS A PROJECT TO IMPORT!!!!"
)

// TestProjectsAddNoBlock is the function allowing to test if
// the command "rp-cli projects add --nonblocking ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsAddNoBlock(test *testing.T) {

	// Create the command
	cmdToRun := getMainAddPrjCmd("Test Project", "redpesk-devel-arz")
	// Add the non-blocking flag
	cmdToRun = append(cmdToRun, "--nonblocking")

	// Create the project if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The project has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":    `^[a-z0-9-]+$`,
		"Name":    `.+`,
		"Version": `[Not set]`,
		"Type":    `^standard(-applications|-images|-containers)?$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, projectDisplayNoVerboseColNb, regexpMap, prjsCommand, "get", "test-project")
}

// TestProjectsAddBlock is the function allowing to test if
// the command "rp-cli projects add ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsAddBlock(test *testing.T) {

	// Create the command
	cmdToRun := getMainAddPrjCmd("Test Project 2", "redpesk-devel-arz")

	// Create the project if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project-2", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The project has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":    `^[a-z0-9-]+$`,
		"Name":    `.+`,
		"Version": `[Not set]`,
		"Type":    `^standard(-applications|-images|-containers)?$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, projectDisplayNoVerboseColNb, regexpMap, prjsCommand, "get", "test-project-2")
}

// TestProjectsAddContainers is the function allowing to test if
// the command "rp-cli projects add ..." is working correctly when the
// project created is supporting containers also
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsAddContainers(test *testing.T) {

	// Create the command
	cmdToRun := getMainAddPrjCmd("Test Project Containers", "redpesk-devel-arz")
	cmdToRun = append(cmdToRun, "--containers")

	// Create the project if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project-containers", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The project has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                  `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                 `^Test Project Containers`,
		"Slug":                 `^test-project-containers$`,
		"Version":              `[Not set]`,
		"Type":                 `^standard(-applications|-images|-containers)?$`,
		"Supported Objects":    `^containers$`,
		"Description":          `.+`,
		"Distributions":        `^redpesk-devel-arz.*`,
		"Architectures":        `(x86_64|aarch64)`,
		"Creation time":        `.+`,
		"Update time":          `.+`,
		"Bookmarked":           `^false$`,
		"Short release naming": `^false$`,
		"Access token(s)":      `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectDisplayVerboseFullLineNb, regexpMap, prjsCommand, "get", "-v", "test-project-containers")
}

// TestProjectsInheritance is the function allowing to test if
// the command "rp-cli projects add ..." is working correctly
// when a project inherit from another one
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsInheritance(test *testing.T) {

	// Create the command
	cmdToRun := getMainAddPrjCmd("Test Project", "redpesk-devel-arz")

	// Create the project if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the project that inherit from the previously created project
	cmdToRun1 := getMainAddPrjCmd("Inheriting Project", "redpesk-devel-arz")
	// Add the inherit flag
	cmdToRun1 = append(cmdToRun1, "--inherit", "test-project")

	// Create the project if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, prjsCommand, "inheriting-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// The project has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `^Inheriting Project`,
		"Slug":              `^inheriting-project$`,
		"Version":           `[Not set]`,
		"Type":              `^standard(-applications|-images|-containers)?$`,
		"Supported Objects": `^applications$`,
		"Description":       `.+`,
		"Inherits from":     `^test-project`,
		"Distributions":     `^redpesk-devel-arz.*`,
		"Architectures":     `(x86_64|aarch64)`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^false$`,
		"Access token(s)":   `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectDisplayVerboseFullLineNb+1, regexpMap2, prjsCommand, "get", "-v", "inheriting-project")

	// Delete the inheriting project
	checkActionCmdAns(test, prjsCommand, "delete", "inheriting-project")
}

// TestProjectsInheritFromReleased is the function allowing to test if
// the command "rp-cli projects add ..." is working correctly
// when a project inherit from another one that has been released
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsInheritFromReleased(test *testing.T) {

	// Do a first creation/release
	createAndReleaseProject(test, "Lorenzo", []string{}, []string{}, false)

	// Create the project that inherit from the previously created project
	cmdToRun1 := getMainAddPrjCmd("Inheriting Lorenzo", "redpesk-devel-arz")
	// Add the inherit flag
	cmdToRun1 = append(cmdToRun1, "--inherit", "lorenzo-1.0")

	// Create the project if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "inheriting-lorenzo", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// The project has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `^Inheriting Lorenzo`,
		"Slug":              `^inheriting-lorenzo$`,
		"Version":           `[Not set]`,
		"Type":              `^standard(-applications|-images|-containers)?$`,
		"Supported Objects": `^applications$`,
		"Description":       `.+`,
		"Inherits from":     `^lorenzo-1.0 \[Priority: [0-9]\]$`,
		"Distributions":     `^redpesk-devel-arz.*`,
		"Architectures":     `^(x86_64|aarch64)`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^false$`,
		"Access token(s)":   `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectDisplayVerboseFullLineNb+1, regexpMap2, prjsCommand, "get", "-v", "inheriting-lorenzo")

	// Delete the inheriting project
	checkActionCmdAns(test, prjsCommand, "delete", "inheriting-lorenzo")

	// Delete the released project and check that it went [OK]
	checkActionCmdAns(test, prjsCommand, "releases", "delete", "lorenzo-1.0")
}

// TestProjectsList is the function allowing to test if
// the command "rp-cli projects list ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsList(test *testing.T) {

	// Create the commands
	cmdToRun1 := getMainAddPrjCmd("Test Project", "redpesk-devel-arz")
	cmdToRun2 := getMainAddPrjCmd("Test Project 2", "redpesk-devel-arz")
	cmdToRun3 := getMainAddPrjCmd("Test Project 3", "redpesk-devel-arz")

	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, prjsCommand, "test-project-2", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, prjsCommand, "test-project-3", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// The projects have been created, check it is now in backend

	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":    `^[a-z0-9-]+$`,
		"Name":    `.+`,
		"Version": `[Not set]`,
		"Type":    `^standard(-applications|-images|-containers)?$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, projectDisplayNoVerboseColNb, regexpMap, prjsCommand, "list")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `.+`,
		"Slug":              `^[a-z0-9-]+$`,
		"Version":           `[Not set]`,
		"Type":              `^standard(-applications|-images|-containers)?$`,
		"Supported Objects": `^(applications|images|containers)$`,
		"Description":       `.+`,
		"Distributions":     `^redpesk-(devel|lts)-(arz|33).*`,
		"Architectures":     `(x86_64|aarch64)`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^(true|false)$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, projectDisplayVerboseLightLineNb, "", regexpMap2, prjsCommand, "list", "-v")
}

// TestProjectsListTypeSpecified is the function allowing to test if
// the command "rp-cli projects list ..." is working correctly when
// the flag `--type` is specified
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsListTypeSpecified(test *testing.T) {

	// Create the commands
	cmdToRun1 := getMainAddPrjCmd("Nouveau Monde", "redpesk-devel-arz")
	cmdToRun2 := getMainAddPrjCmd("Piany pianino", "redpesk-devel-arz")
	cmdToRun3 := getMainAddPrjCmd("Chapters", "redpesk-devel-arz")

	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "nouveau-monde", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, prjsCommand, "piany-pianino", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, prjsCommand, "chapters", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// FIXME: TO REWORK WITH NEW TESTS ON PROJECT RELEASE (ARMAND)

	// // The projects have been created, now release two of them and check that all is [OK]
	// checkActionCmdAns(test, adminCommand, "release-project", "piany-pianino")
	// checkActionCmdAns(test, adminCommand, "release-project", "chapters")

	// // Not verbose mode
	// // Create the map containing the regexp we want to match in the answer
	// regexpMap := map[string]string{
	// 	"ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
	// 	"Slug":    `^[a-z0-9-]+$`,
	// 	"Name":    `.+`,
	// 	"Version": `[Not set]`,
	// 	"Type":    `^released$`,
	// }

	// // Run the command and the tests (no verbose and verbose, since this is the same result)
	// checkCmdAnsKeysRegexpCol(test, projectDisplayNoVerboseColNb, regexpMap, prjsCommand, "list", "--type", "released")

	// // Verbose mode
	// // Create the map containing the regexp we want to match in the answer
	// regexpMap2 := map[string]string{
	// 	"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
	// 	"Name":              `.+`,
	// 	"Slug":              `^[a-z0-9-]+$`,
	// 	"Version":           `[Not set]`,
	// 	"Type":              `^released$`,
	// 	"Supported Objects": `^(applications|images)$`,
	// 	"Description":       `.+`,
	// 	"Distributions":     `^redpesk-devel-arz.*`,
	// 	"Architectures":     `(x86_64|aarch64)`,
	// 	"Creation time":     `.+`,
	// 	"Update time":       `.+`,
	// 	"Bookmarked":        `^(true|false)$`,
	// }

	// // Run the command and the tests
	// checkCmdAnswerListKeysRegexp(test, projectDisplayVerboseLineNb, "", regexpMap2, prjsCommand, "list", "-v", "--type", "released")

	// // Now request type "standard" and "released" and check that we have it all
	// regexpMap["Type"] = `^(released|standard)$`
	// checkCmdAnsKeysRegexpCol(test, projectDisplayNoVerboseColNb, regexpMap, prjsCommand, "list", "--type", "released", "--type", "standard")
}

// TestProjectsGet is the function allowing to test if
// the command "rp-cli projects get ..." is working correctly
// with any flag
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsGet(test *testing.T) {

	// Create the command
	cmdToRun := getMainAddPrjCmd("Test Project", "redpesk-devel-arz")

	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Test the repos option ------------------------------
	output, err := executeCommand(prjsCommand, "get", "-r", "test-project")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer contains the right title
	assert.Contains(test, output, prjsGetCmdAnsLine0)

	// Get the second part of the answer
	splitAns := strings.Split(output, prjsGetCmdAnsLine0)

	// Check that the second part contains the slug of the project and an URL
	assert.Contains(test, splitAns[1], "test-project")
	assert.Regexp(test, regexp.MustCompile(`http(s)*://`), splitAns[1])
	// ----------------------------------------------------

	// Test the roles option ------------------------------
	output, err = executeCommand(prjsCommand, "get", "--roles", "test-project")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer contains the right title
	assert.Contains(test, output, prjsGetCmdAnsLine1)

	// Get the second part of the answer
	splitAns = strings.Split(output, prjsGetCmdAnsLine1)

	// Check that the second part contains the slug of the project and an URL
	assert.Contains(test, splitAns[1], "manager")
	assert.Contains(test, splitAns[1], "tester")
	assert.Contains(test, splitAns[1], "developer")
	// ----------------------------------------------------
}

// TestProjectsUpdateDesc is the function allowing to test if
// the command "rp-cli projects update ..." is working correctly
// when the user is updating the description
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsUpdateDesc(test *testing.T) {

	// Create the command
	cmdToRun := getMainAddPrjCmd("Test Project", "redpesk-devel-arz")

	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateDescPrjCmd("test-project", "Updated description")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the description has been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `.+`,
		"Version":           `[Not set]`,
		"Slug":              `^[a-z0-9-]+$`,
		"Type":              `^standard(-applications|-images|-containers)?$`,
		"Supported Objects": `^applications$`,
		"Description":       `^Updated description`,
		"Distributions":     `^redpesk-devel-arz \[Mand\]$`,
		"Architectures":     `(x86_64|aarch64)`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^(true|false)$`,
		"Access token(s)":   `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectDisplayVerboseFullLineNb, regexpMap2, prjsCommand, "get", "-v", "test-project")
}

// TestProjectsUpdateDistro is the function allowing to test if
// the command "rp-cli projects update ..." is working correctly
// when the user is updating the distribution
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsUpdateDistro(test *testing.T) {

	// Create the command
	cmdToRun := getMainAddPrjCmd("Test Project Update distro", "redpesk-devel-arz")

	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project-update-distro", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateDistrPrjCmd("test-project-update-distro")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the description has been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `.+`,
		"Version":           `[Not set]`,
		"Slug":              `^[a-z0-9-]+$`,
		"Type":              `^standard(-applications|-images|-containers)?$`,
		"Supported Objects": `^applications$`,
		"Description":       `^This is a test project`,
		"Distributions":     `^redpesk-devel-arz \[Opt\]$`,
		"Architectures":     `(x86_64|aarch64)`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^(true|false)$`,
		"Access token(s)":   `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectDisplayVerboseFullLineNb, regexpMap2, prjsCommand, "get", "-v", "test-project-update-distro")
}

// TestProjectsUpdateArch is the function allowing to test if
// the command "rp-cli projects update ..." is working correctly
// when the user is updating the architecture
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsUpdateArch(test *testing.T) {

	// Create the command
	cmdToRun := getMainAddPrjCmd("Test Project Update arch", "redpesk-devel-arz")

	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project-update-arch", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateArchPrjCmd("test-project-update-arch")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the architecture has been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `.+`,
		"Version":           `[Not set]`,
		"Slug":              `^[a-z0-9-]+$`,
		"Type":              `^standard(-applications|-images|-containers)?$`,
		"Supported Objects": `^applications$`,
		"Description":       `^This is a test project`,
		"Distributions":     `^redpesk-devel-arz \[Mand\]$`,
		"Architectures":     `^x86_64 \[Mand\]$`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^(true|false)$`,
		"Access token(s)":   `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectDisplayVerboseFullLineNb, regexpMap2, prjsCommand, "get", "-v", "test-project-update-arch")
}

// TestProjectsUpdateExternalProject is the function allowing to test if
// the command "rp-cli projects update ..." is working correctly
// when the user is updating an external project (also called "distro")
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsUpdateExternalProject(test *testing.T) {
	// Firstly, import a project in order to have a "distro"
	err := addObjectIfNotExists(test, prjsCommand, "redpesk-distro-project", "admin", "import-project", "-i", "redpesk-distro-project", "-v", "1.0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the project exists in the backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `^redpesk-distro-project$`,
		"Slug":              `^redpesk-distro-project$`,
		"Version":           `^1\.0$`,
		"Type":              `^distro \(external\)-applications$`,
		"Supported Objects": `^applications$`,
		"Description":       `^\[Not set\]$`,
		"Distributions":     `.*`,
		"Architectures":     `.*`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^(true|false)$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, projectDisplayVerboseLightLineNb, "", regexpMap, prjsCommand, "get", "redpesk-distro-project", "-v")

	// Try to update the "distro" project
	updateCmd := getUpdateDescPrjCmd("redpesk-distro-project", "Updated description")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the description has been updated
	regexpMap["Description"] = `^Updated description`

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectDisplayVerboseLightLineNb, regexpMap, prjsCommand, "get", "-v", "redpesk-distro-project")
}

// TestProjectsDelete is the function allowing to test if
// the command "rp-cli projects delete ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsDelete(test *testing.T) {

	// Create the commands
	cmdToRun1 := getMainAddPrjCmd("Test Project", "redpesk-devel-arz")
	cmdToRun2 := getMainAddPrjCmd("Test Project 2", "redpesk-devel-arz")
	cmdToRun3 := getMainAddPrjCmd("Test Project 3", "redpesk-devel-arz")

	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, prjsCommand, "test-project-2", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, prjsCommand, "test-project-3", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// The projects have been created, now try to delete one

	// Run the delete command and check that all is [OK]
	checkActionCmdAns(test, prjsCommand, "delete", "test-project-2")

	// Check that the project deleted does not exist anymore
	_, err = executeCommand(prjsCommand, "get", "test-project-2")
	assert.NotNil(test, err)
}

// TestProjectsBookmark is the function allowing to test if
// the command "rp-cli projects bookmark ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsBookmark(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Run the bookmark command and check that all is [OK]
	checkActionCmdAns(test, prjsCommand, "bookmark", "test-project")

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `^Test Project$`,
		"Slug":              `^[a-z0-9-]+$`,
		"Version":           `[Not set]`,
		"Type":              `^standard(-applications|-images|-containers)?$`,
		"Supported Objects": `^applications$`,
		"Description":       `.*`,
		"Distributions":     `^redpesk-devel-arz.*`,
		"Architectures":     `(x86_64|aarch64)`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^true$`,
		"Access token(s)":   `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectDisplayVerboseFullLineNb, regexpMap, prjsCommand, "get", "-v", "test-project")

	// Run the reset bookmark command and check that all is [OK]
	checkActionCmdAns(test, prjsCommand, "bookmark", "test-project", "--reset")

	// Change slightly the regexpMap in order to check if the bookmark has been reset
	regexpMap["Bookmarked"] = `^false$`

	// Check if the last command has been taken into account
	checkCmdAnswerKeysRegexp(test, projectDisplayVerboseFullLineNb, regexpMap, prjsCommand, "get", "-v", "test-project")
}

// TestProjectsDisplayBookmarked is the function allowing to test if
// the command "rp-cli projects list --bookmarked" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsDisplayBookmarked(test *testing.T) {
	// Create 3 projects
	cmdToRun1 := getMainAddPrjCmd("Test Project", "redpesk-devel-arz")
	cmdToRun2 := getMainAddPrjCmd("Test Project 2", "redpesk-devel-arz")
	cmdToRun3 := getMainAddPrjCmd("Test Project 3", "redpesk-devel-arz")

	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "test-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, prjsCommand, "test-project-2", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, prjsCommand, "test-project-3", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// Bookmark 2 out of 3
	checkActionCmdAns(test, prjsCommand, "bookmark", "test-project-2")
	checkActionCmdAns(test, prjsCommand, "bookmark", "test-project-3")

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                  `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                 `.+`,
		"Slug":                 `^test-project-[0-9]{1}$`,
		"Version":              `[Not set]`,
		"Type":                 `^standard(-applications|-images|-containers)?$`,
		"Supported Objects":    `^applications$`,
		"Description":          `.+`,
		"Distributions":        `^redpesk-devel-arz.*`,
		"Architectures":        `(x86_64|aarch64)`,
		"Creation time":        `.+`,
		"Update time":          `.+`,
		"Bookmarked":           `true`,
		"Short release naming": `^false$`,
	}

	// Run the command and tests that only bookmarked projects are displayed
	checkCmdAnswerListKeysRegexp(test, projectDisplayVerboseLightLineNb, "", regexpMap, prjsCommand, "list", "-v", "-b")

	// Reset the bookmarks on the projects
	checkActionCmdAns(test, prjsCommand, "bookmark", "test-project-2", "--reset")
	checkActionCmdAns(test, prjsCommand, "bookmark", "test-project-3", "--reset")
}

// TestProjectsMemberAddNoBlock is the function allowing to test if
// the command "rp-cli projects members add <prjID> ... --nonblocking" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsMemberAddNoBlock(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then creates a test user if it does not exists
	cmdToRun2 := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")
	err = addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Add the user as member of the project
	memberRoles := []string{"tester"}
	cmdToRun3 := getMainAddMemberCmd("test-project", "tiramisu", memberRoles)
	cmdToRun3 = append(cmdToRun3, "--nonblocking")
	checkActionCmdAns(test, cmdToRun3...)

	// Check that the member is in the list
	memberInPrj := checkMemberIsInPrj(test, "test-project", "tiramisu", "")

	// Delete the member newly added
	if memberInPrj {
		checkActionCmdAns(test, prjsCommand, "members", "delete", "test-project", "-m", "tiramisu")
	}
}

// TestProjectsMemberAddBlock is the function allowing to test if
// the command "rp-cli projects members add <prjID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsMemberAddBlock(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project 2", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-2", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then creates a test team if it does not exists
	cmdToRun2 := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", true)
	err = addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Add the team as member of the project
	memberRoles := []string{"developer", "tester"}
	cmdToRun3 := getMainAddMemberCmd("test-project-2", "redpesk-team", memberRoles)
	checkActionCmdAns(test, cmdToRun3...)

	// Check that the member is in the list
	memberInPrj := checkMemberIsInPrj(test, "test-project-2", "redpesk-team", "")

	// Delete the member newly added
	if memberInPrj {
		checkActionCmdAns(test, prjsCommand, "members", "delete", "test-project-2", "-m", "redpesk-team")
	}
}

// TestProjectsMembersGet is the function allowing to test if
// the command "rp-cli projects members get <prjID> -v ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsMembersGet(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project 3", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-3", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then creates a test user if it does not exists
	cmdToRun2 := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")
	err = addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create another user
	cmdToRun3 := getMainAddUserCmd("kravmager", "krav@iot.bzh", "Imi", "Lichtenfeld", "0666999949")
	err = addObjectIfNotExists(test, usersCommand, "kravmager", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// Creates a test team if it does not exists
	cmdToRun4 := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", true)
	err = addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// Add everything as members of the project
	cmdToRun5 := getMainAddMemberCmd("test-project-3", "tiramisu", []string{"tester"})
	checkActionCmdAns(test, cmdToRun5...)
	cmdToRun6 := getMainAddMemberCmd("test-project-3", "kravmager", []string{"developer", "tester"})
	checkActionCmdAns(test, cmdToRun6...)
	cmdToRun7 := getMainAddMemberCmd("test-project-3", "redpesk-team", []string{"developer"})
	checkActionCmdAns(test, cmdToRun7...)

	// Verbose only, since the no verbose option is tested in almost all other tests

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Member ID": `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Pseudo":    `^[a-z0-9-_]+$`,
		"Role(s)":   `(manager|tester|developer)`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, prjMembersDisplayVerboseLineNb, "", regexpMap, prjsCommand, "members", "get", "test-project-3", "-v")

	// Delete all the added members
	checkActionCmdAns(test, prjsCommand, "members", "delete", "test-project-3", "-m", "tiramisu")
	checkActionCmdAns(test, prjsCommand, "members", "delete", "test-project-3", "-m", "kravmager")
	checkActionCmdAns(test, prjsCommand, "members", "delete", "test-project-3", "-m", "redpesk-team")
}

// TestProjectsMemberUpdate is the function allowing to test if
// the command "rp-cli projects members update <prjID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsMemberUpdate(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then creates a test user if it does not exists
	cmdToRun2 := getMainAddUserCmd("kravmager", "krav@iot.bzh", "Imi", "Lichtenfeld", "0666999949")
	err = addObjectIfNotExists(test, usersCommand, "kravmager", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Add the user as member of the project
	cmdToRun3 := getMainAddMemberCmd("test-project", "kravmager", []string{"developer"})
	checkActionCmdAns(test, cmdToRun3...)

	// Check that the member is in the list
	memberInPrj := checkMemberIsInPrj(test, "test-project", "kravmager", `^developer$`)

	// Update the member
	cmdToRun4 := getMainUpdateMemberCmd("test-project", "kravmager", []string{"developer", "tester"})
	checkActionCmdAns(test, cmdToRun4...)

	// Check that the member has been updated is in the list
	checkMemberIsInPrj(test, "test-project", "kravmager", `^(developer|tester), (developer|tester)$`)

	// Delete the member newly added
	if memberInPrj {
		checkActionCmdAns(test, prjsCommand, "members", "delete", "test-project", "-m", "kravmager")
	}
}

// TestProjectsMemberDelete is the function allowing to test if
// the command "rp-cli projects members delete <prjID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsMemberDelete(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project 2", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-2", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then creates a test team if it does not exists
	cmdToRun2 := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", true)
	err = addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Add the team as member of the project
	cmdToRun3 := getMainAddMemberCmd("test-project-2", "redpesk-team", []string{"developer", "tester"})
	checkActionCmdAns(test, cmdToRun3...)

	// Check that the member is in the list
	checkMemberIsInPrj(test, "test-project-2", "redpesk-team", "")

	// Delete the member newly added
	checkActionCmdAns(test, prjsCommand, "members", "delete", "test-project-2", "-m", "redpesk-team")

	// Check that the member is not in the Project anymore
	isCorrectlyDeleted := checkMemberIsNotInPrj(test, "test-project-2", "redpesk-team")
	assert.True(test, isCorrectlyDeleted)
}

// TestProjectsExportStartSuccess is the function allowing to test if
// the command "rp-cli projects export start <prjID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsExportStartSuccess(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project Export", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-export", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Request the export of this project, and check that everything is OK
	commandOutput, err := executeCommand(prjsCommand, prjsExportCmd, "start", "test-project-export", "-d", exportImportProjectsDirectory)
	if err != nil {
		test.Fatal(err)
	}
	assert.NotContains(test, commandOutput, "[KO]")

	// Check that the output contains the path to the file
	assert.Contains(test, commandOutput, "Path to the exported project archive")
	assert.Contains(test, commandOutput, (exportImportProjectsDirectory + "mocked_project_export.zip"))

	// Check that the file exists
	assert.FileExists(test, (exportImportProjectsDirectory + "mocked_project_export.zip"))
}

// TestProjectsExportStartFailed is the function allowing to test if
// the command "rp-cli projects export start <prjID> ..." is working correctly
// when the export fails on the server point of view
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsExportStartFailed(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project Export", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-export", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Request the mocked backend to fail
	rpbckmock.RequestPrjExportFails()

	// Request the export of this project, and check that everything is OK
	_, err = executeCommand(prjsCommand, prjsExportCmd, "start", "test-project-export", "-d", exportImportProjectsDirectory)
	if assert.Error(test, err) {
		assert.Equal(test, "an error occurred during project export", err.Error())
	}
}

// TestProjectsExportGet is the function allowing to test if
// the command "rp-cli projects export get <prjID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsExportGet(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project Export Get", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-export-get", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Request the export of this project, and check that everything is OK
	commandOutput, err := executeCommand(prjsCommand, prjsExportCmd, "start", "test-project-export-get", "-d", exportImportProjectsDirectory)
	if err != nil {
		test.Fatal(err)
	}
	assert.NotContains(test, commandOutput, "[KO]")

	// Check that the output contains the path to the file
	assert.Contains(test, commandOutput, "Path to the exported project archive")
	assert.Contains(test, commandOutput, (exportImportProjectsDirectory + "mocked_project_export.zip"))

	// Check that the file exists
	assert.FileExists(test, (exportImportProjectsDirectory + "mocked_project_export.zip"))

	// Not verbose mode
	commandOutput, err = executeCommand(prjsCommand, prjsExportCmd, "get", "test-project-export-get")
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, commandOutput, string(rpapiv1.RPStateExportProjectDone)+"\n")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Project ID:": `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Start date":  `.+`,
		"End date":    `.+`,
		"Status":      `^done$`,
		"Elapsed at":  `.+`,
		"Error":       `^\[Not set\]$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, prjExportDisplayVerboseLineNb, "", regexpMap, prjsCommand, prjsExportCmd, "get", "test-project-export-get", "-v")
}

// TestProjectsExportDownload is the function allowing to test if
// the command "rp-cli projects export download <prjID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsExportDownload(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project Export Dld", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-export-dld", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Request the export of this project, and check that everything is OK
	commandOutput, err := executeCommand(prjsCommand, prjsExportCmd, "start", "test-project-export-dld", "--nonblocking")
	if err != nil {
		test.Fatal(err)
	}
	assert.NotContains(test, commandOutput, "[KO]")

	// Delete the archive if it has already been downloaded previously
	exportArchive := (exportImportProjectsDirectory + "mocked_project_export.zip")
	_, err = os.Stat(exportArchive)
	if !os.IsNotExist(err) {
		err = os.Remove(exportArchive)
		if err != nil {
			test.Fatal(err)
		}
	}

	// Check that the file has been correctly deleted
	assert.NoFileExists(test, exportArchive)

	// Request to download the archive
	commandOutput, err = executeCommand(prjsCommand, prjsExportCmd, "download", "test-project-export-dld", "--dir", exportImportProjectsDirectory)
	if err != nil {
		test.Fatal(err)
	}
	assert.NotContains(test, commandOutput, "[KO]")

	// Check that the output contains the path to the file
	assert.Contains(test, commandOutput, "Path to the exported project archive")
	assert.Contains(test, commandOutput, exportArchive)

	// Check that the file exists
	assert.FileExists(test, exportArchive)
}

// TestProjectsImportStartSuccess is the function allowing to test if
// the command "rp-cli projects import start <prjID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsImportStartSuccess(test *testing.T) {
	// Create a fake archive file
	importArchive := path.Join(exportImportProjectsDirectory, "project-to-import-success.zip")
	err := createFalseFile(importArchive, importArchiveContent)
	if err != nil {
		test.Fatal(err)
	}

	// Defer the deletion of the file
	defer deleteFalseFile(importArchive)

	// Request to upload and import the project
	checkActionCmdAns(test, prjsCommand, prjsImportCmd, prjsImportStartCmd, "-f", importArchive)
}

// TestProjectsImportStartFail is the function allowing to test if
// the command "rp-cli projects import start <prjID> ..." is working correctly
// when the import fails from the server point of view
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsImportStartFail(test *testing.T) {
	// Create a fake archive file
	importArchive := path.Join(exportImportProjectsDirectory, "project-to-import-fail.zip")
	err := createFalseFile(importArchive, importArchiveContent)
	if err != nil {
		test.Fatal(err)
	}

	// Defer the deletion of the file
	defer deleteFalseFile(importArchive)

	// Request the mocked backend to fails on import
	rpbckmock.RequestPrjImportFails()

	// Request to upload and import the project
	_, err = executeCommand(prjsCommand, prjsImportCmd, prjsImportStartCmd, "-f", importArchive)
	if assert.Error(test, err) {
		assert.Equal(test, "an error occurred during project upload", err.Error())
	}
}

// TestProjectsImportStartInteractive is the function allowing to test if
// the command "rp-cli projects import start <prjID> --interactive ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsImportStartInteractive(test *testing.T) {
	// Create a fake archive file
	importArchive := path.Join(exportImportProjectsDirectory, "project-to-import-interactive.zip")
	err := createFalseFile(importArchive, importArchiveContent)
	if err != nil {
		test.Fatal(err)
	}

	// Defer the deletion of the file
	defer deleteFalseFile(importArchive)

	// Create a pipe
	readPipe, writePipe, err := os.Pipe()
	if err != nil {
		test.Fatal(err)
	}

	// Save the original *os.File
	origStdin := os.Stdin

	// Set os.Stdin with our fake one
	os.Stdin = readPipe

	// Start a go routine that will press enter for us
	go func() {
		// Wait a bit, in order for the command to be executed
		time.Sleep(250 * time.Millisecond)

		// Say that we want to configure the backend connection
		writePipe.Write([]byte("\n"))
	}()

	// Request to upload and import the project in interactive mode
	_, err = executeCommand(prjsCommand, prjsImportCmd, prjsImportStartCmd, "-f", importArchive, "--interactive")

	// Restore Stdin
	os.Stdin = origStdin

	if err != nil {
		test.Fatal(err)
	}

	// Get the latest imported project UUID
	wantedUUID := rpbckmock.GetLatestImportedProjectUUID()
	commandOutput, err := executeCommand(prjsCommand, prjsImportCmd, "get", wantedUUID.String())
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, commandOutput, string(rpapiv1.RPStateImportProjectDone)+"\n")
}

// TestProjectsImportGet is the function allowing to test if
// the command "rp-cli projects import get <prjID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsImportGet(test *testing.T) {
	// Create a fake archive file
	importArchive := path.Join(exportImportProjectsDirectory, "project-to-import-success-get.zip")
	err := createFalseFile(importArchive, importArchiveContent)
	if err != nil {
		test.Fatal(err)
	}

	// Defer the deletion of the file
	defer deleteFalseFile(importArchive)

	// Request to upload and import the project
	checkActionCmdAns(test, prjsCommand, prjsImportCmd, prjsImportStartCmd, "-f", importArchive)

	// Get the latest imported project UUID
	wantedUUID := rpbckmock.GetLatestImportedProjectUUID()

	// Try to get the project import state

	// Not verbose mode
	commandOutput, err := executeCommand(prjsCommand, prjsImportCmd, "get", wantedUUID.String())
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, commandOutput, string(rpapiv1.RPStateImportProjectDone)+"\n")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Project ID:": `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Start date":  `.+`,
		"End date":    `.+`,
		"Status":      `^done$`,
		"Error":       `^\[Not set\]$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, prjImportDisplayVerboseLineNb, "", regexpMap, prjsCommand, prjsImportCmd, "get", wantedUUID.String(), "-v")
}

// TestProjectsUpdateAppsBuildType is the function allowing to test if
// the command "rp-cli projects update ... --build-type ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsUpdateAppsBuildType(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("led zep project", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "led-zep-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Now create two applications
	app1Cmd := getMainAddAppCmd("led-zep-app-1", "led-zep-app-1", "led-zep-project")
	err = addObjectIfNotExists(test, appsCommand, "led-zep-app-1", app1Cmd...)
	if err != nil {
		test.Fatal(err)
	}
	app2Cmd := getMainAddAppCmd("led-zep-app-2", "led-zep-app-2", "led-zep-project")
	err = addObjectIfNotExists(test, appsCommand, "led-zep-app-2", app2Cmd...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdatePrjCmd("led-zep-project", "--build-type", string(rpapiv1.RPAppStandardBuildCross))
	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the apps description has been updated for both of them

	regexpMap := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `^led-zep-app`,
		"Package Name":      `^led-zep-app`,
		"Slug":              `^led-zep-app`,
		"Type":              `^standard$`,
		"Parent project":    `^led-zep-project$`,
		"Description":       `.*`,
		"Source URL":        `http(s)*://`,
		"Source Revision":   `^master$`,
		"External Specfile": `(\[Not set\]|(http(s)*://))`,
		"Distributions":     `^redpesk-devel-arz.*`,
		"Autotests":         `^(ON|OFF)$`,
		"Build type":        `^standard_build_cross$`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^false$`,
		"Services":          `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "led-zep-app-1", "-v")
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "led-zep-app-2", "-v")

	// Now update back only the second app
	updateCmd = getUpdatePrjCmd("led-zep-project", "--build-type", string(rpapiv1.RPAppEmulatedBuild))
	updateCmd = append(updateCmd, "--bld-type-apps", "led-zep-app-2")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the app description has been updated correctly
	regexpMap["Build type"] = "^emulated_build$"
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "led-zep-app-2", "-v")
}

// TestProjectsRelease is the function allowing to test if
// the command "rp-cli projects release <prjID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsRelease(test *testing.T) {
	// Add a user if it does not exists
	cmdUserToRun := getMainAddUserCmd("the_banano", "banano@iot.bzh", "Armand", "Bénéteau", "0666989945")
	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "the_banano", cmdUserToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a team if it does not exists
	cmdToRun := getMainAddTeamCmd("Bruit", "La meilleure des teams", true)
	// Create the team if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, teamsCommand, "bruit", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Do the creation/release
	createAndReleaseProject(test, "Banano Industrial", []string{"the_banano"}, []string{"bruit"}, false)
}

// TestProjectsReleasesGet is the function allowing to test if
// the command "rp-cli projects releases get <prjID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsReleasesGet(test *testing.T) {
	// Do a first creation/release
	createAndReleaseProject(test, "Banano Industrial 0", []string{}, []string{}, false)

	// Check the content of the "not verbose" output
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":      `^banano-industrial-0-1.0$`,
		"Version":   `^1.0$`,
		"Reference": `^banano-industrial-0$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, projectRelDisplayNoVerboseColNb, regexpMap, prjsCommand, "releases", "get", "banano-industrial-0-1.0")

	// Check the content of the "verbose" output
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":                     `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                    `^banano-industrial-0-1.0$`,
		"Slug":                    `^banano-industrial-0-1.0$`,
		"Reference project ID":    `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Reference project slug":  `^banano-industrial-0$`,
		"Distribution ID":         `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Version":                 `^1.0$`,
		"Description":             `.+`,
		"Creation time":           `.+`,
		"Update time":             `.+`,
		"Short release naming":    `^false$`,
		"Access token(s)":         `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"[Released applications]": `.*`,
		"Application ID":          `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Package name":            `[a-z0-9-]+`,
		"SHA Gitpkg":              `[a-f0-9]{40}`,
		"Branch":                  `[a-z0-9-_\.]+`,
		"Build ID":                `[0-9]+`,
		"[Detailed info]":         `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectRelDisplayVerboseFullLineNb, regexpMap2, prjsCommand, "releases", "get", "-v", "banano-industrial-0-1.0")
}

// TestProjectsReleasesList is the function allowing to test if
// the command "rp-cli projects releases list <prjID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsReleasesList(test *testing.T) {
	// Do a first creation/release
	createAndReleaseProject(test, "Banano Industrial 1", []string{}, []string{}, false)

	// Then release several times the project
	checkActionCmdAns(test, prjsCommand, "release", "banano-industrial-1", "--distribution", "redpesk-lts-arz", "-v", "1.1")
	checkActionCmdAns(test, prjsCommand, "release", "banano-industrial-1", "--distribution", "redpesk-lts-arz", "-v", "1.2")
	checkActionCmdAns(test, prjsCommand, "release", "banano-industrial-1", "--distribution", "redpesk-lts-arz", "-v", "1.3")

	// Check the content of the output (no verbose)
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":      `^(banano-industrial|project-to-deploy)-*[0-9]*-[12].[0123]$`,
		"Version":   `^[12].[0123]$`,
		"Reference": `^(banano-industrial|project-to-deploy)-*[0-9]*$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, projectRelDisplayNoVerboseColNb, regexpMap, prjsCommand, "releases", "list")

	// Get the total number of released projects
	relPrjNb := getObjNumber(test, prjsCommand, "releases", "list")

	// Check the content of the output (verbose)
	regexpMap2 := map[string]string{
		"ID:":                    `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                   `^(banano-industrial|project-to-deploy)-*[0-9]*-[12].[0123]$`,
		"Slug":                   `^(banano-industrial|project-to-deploy)-*[0-9]*-[12].[0123]$`,
		"Reference project ID":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Reference project slug": `^(banano-industrial|project-to-deploy)-*[0-9]*$`,
		"Distribution ID":        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Version":                `^[12].[0123]$`,
		"Description":            `.+`,
		"Creation time":          `.+`,
		"Update time":            `.+`,
		"Short release naming":   `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, ((projectRelDisplayVerboseLightLineNb+1)*relPrjNb)-1, regexpMap2, prjsCommand, "releases", "list", "-v")
}

// TestProjectsReleasesDelete is the function allowing to test if
// the command "rp-cli projects releases delete <prjID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsReleasesDelete(test *testing.T) {
	// Do a first creation/release
	createAndReleaseProject(test, "Banano Industrial 2", []string{}, []string{}, false)

	// Delete the released project and check that it went [OK]
	checkActionCmdAns(test, prjsCommand, "releases", "delete", "banano-industrial-2-1.0")
}

// TestProjectsReleasesUpdate is the function allowing to test if
// the command "rp-cli projects releases update <prjID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsReleasesUpdate(test *testing.T) {
	// Do a first creation/release
	createAndReleaseProject(test, "Banano Industrial 3", []string{}, []string{}, false)

	// Update the released project
	checkActionCmdAns(test, prjsCommand, "releases", "update", "banano-industrial-3-1.0", "-d", "A new description")

	// Check the content of the "verbose" output
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":                     `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                    `^banano-industrial-3-1.0$`,
		"Slug":                    `^banano-industrial-3-1.0$`,
		"Reference project ID":    `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Reference project slug":  `^banano-industrial-3$`,
		"Distribution ID":         `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Version":                 `^1.0$`,
		"Description":             `^A new description$`,
		"Creation time":           `.+`,
		"Update time":             `.+`,
		"Short release naming":    `^false$`,
		"Access token(s)":         `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"[Released applications]": `.*`,
		"Application ID":          `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Package name":            `[a-z0-9-]+`,
		"SHA Gitpkg":              `[a-f0-9]{40}`,
		"Branch":                  `[a-z0-9-_\.]+`,
		"Build ID":                `[0-9]+`,
		"[Detailed info]":         `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectRelDisplayVerboseFullLineNb, regexpMap2, prjsCommand, "releases", "get", "-v", "banano-industrial-3-1.0")
}

// TestProjectsReleasesUpdateAudience is the function allowing to test if
// the command "rp-cli projects releases update <prjID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsReleasesUpdateAudience(test *testing.T) {
	// Add a user if it does not exists
	cmdUserToRun := getMainAddUserCmd("the_banano", "banano@iot.bzh", "Armand", "Bénéteau", "0666989945")
	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "the_banano", cmdUserToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a team if it does not exists
	cmdToRun := getMainAddTeamCmd("Private Bruit", "La meilleure des teams", false)
	// Create the team if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, teamsCommand, "private-bruit", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Do a first creation/release
	createAndReleaseProject(test, "Banano Industrial 4", []string{"the_banano"}, []string{"private-bruit"}, false)

	// Try to update the audience of the release and check that it goes right!
	checkActionCmdAns(test, prjsCommand, "releases", "update", "banano-industrial-4-1.0", "--reset-teams", "--reset-users", "-a", "true")
}

// TestProjectsReleasesGetAudience is the function allowing to test if
// the command "rp-cli projects releases get <prjID> -av" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsReleasesGetAudience(test *testing.T) {
	// Add a user if it does not exists
	cmdUserToRun := getMainAddUserCmd("the_banano", "banano@iot.bzh", "Armand", "Bénéteau", "0666989945")
	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "the_banano", cmdUserToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a team if it does not exists
	cmdToRun := getMainAddTeamCmd("Private Bruit", "La meilleure des teams", false)
	// Create the team if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, teamsCommand, "private-bruit", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Do a first creation/release
	createAndReleaseProject(test, "Banano Industrial 5", []string{"the_banano"}, []string{"private-bruit"}, false)

	// Check the content of the "verbose" output
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":                     `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                    `^banano-industrial-5-1.0$`,
		"Slug":                    `^banano-industrial-5-1.0$`,
		"Reference project ID":    `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Reference project slug":  `^banano-industrial-5$`,
		"Distribution ID":         `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Version":                 `^1.0$`,
		"Description":             `.+`,
		"Creation time":           `.+`,
		"Update time":             `.+`,
		"Short release naming":    `^false$`,
		"Access token(s)":         `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"[Released applications]": `.*`,
		"Application ID":          `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Package name":            `[a-z0-9-]+`,
		"SHA Gitpkg":              `[a-f0-9]{40}`,
		"Branch":                  `[a-z0-9-_\.]+`,
		"Build ID":                `[0-9]+`,
		"[Detailed info]":         `.*`,
		"[Audience]":              `.*`,
		"All Users":               `^no$`,
		"Team(s)":                 `^private-bruit$`,
		"User(s)":                 `^the_banano$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectRelDisplayVerboseFullAudienceLineNb, regexpMap2, prjsCommand, "releases", "get", "-av", "banano-industrial-5-1.0")
}

// TestProjectsBuildsStats is the function allowing to test if
// the command "rp-cli projects builds stats <prjID>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsBuildsStats(test *testing.T) {
	nbApps := 3
	// First create all the projects, apps and builds
	err := createFullPrjAppsBuildsTests(test, "Eminem", "redpesk-lts-arz", "Recovery", nbApps, 3, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Check the content of the "not verbose" output
	regexpMap := map[string]string{
		"Project ID:": `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Done:":       `^[0-12]+$`,
		"Failed:":     `^[0-12]+$`,
		"Cancelled:":  `^[0-12]+$`,
		"Total:":      `^12$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectBuildsStatsNoVerboseLineNb, regexpMap, prjsCommand, prjsBuildsCmd, "stats", "eminem")

	// Now, let's go for the verbose mode
	verboseAns := checkCmdAnswerKeysRegexp(test, projectBuildsStatsNoVerboseLineNb+4+nbApps, regexpMap, prjsCommand, prjsBuildsCmd, "stats", "eminem", "-v")

	// Create the regexp map we will use
	regexpMapCol := map[string]string{
		"ID":        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":      `^recovery-[0-9]+$`,
		"Package":   `^recovery-[0-9]+$`,
		"Done":      `^[0-9]+$`,
		"Failed":    `^[0-9]+$`,
		"Cancelled": `^[0-9]+$`,
		"Total":     `^4$`,
	}

	// Get the second part of the answer
	columnsAns := strings.Split(verboseAns, prjsSubBuildsStatsCmdAnsLine5)[1]
	// Trim it
	columnsAns = strings.TrimSpace(columnsAns)

	// Check the answer of the command
	checkObjectKeysRegexpCol(test, columnsAns, projectBuildsStatsVerboseColNb, regexpMapCol)
}

// TestProjectsBuildsMatrix is the function allowing to test if
// the command "rp-cli projects builds matrix <prjID>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsBuildsMatrix(test *testing.T) {
	nbApps := 3

	// First create all the projects, apps and builds
	err := createFullPrjAppsBuildsTests(test, "Build Matrix", "redpesk-lts-arz", "Matrix App", nbApps, 3, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Application":     `^matrix-app-[0-9]+$`,
		"redpesk-lts-arz": `^(done|failed)$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, 2, regexpMap, prjsCommand, prjsBuildsCmd, "matrix", "build-matrix")
}

// TestProjectsTestsStats is the function allowing to test if
// the command "rp-cli projects tests stats <prjID>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestProjectsTestsStats(test *testing.T) {
	nbApps := 3
	// First create all the projects, apps and builds
	err := createFullPrjAppsBuildsTests(test, "MPL", "redpesk-lts-arz", "Mystère abyssal", nbApps, 3, 5)
	if err != nil {
		test.Fatal(err)
	}

	// Check the content of the "not verbose" output
	regexpMap := map[string]string{
		"Project ID:": `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Succeeded:":  `^[0-9]+$`,
		"Failed:":     `^[0-9]+$`,
		"Cancelled:":  `^[0-9]+$`,
		"Total:":      `^15$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, projectTestsStatsNoVerboseLineNb, regexpMap, prjsCommand, prjsTestsCmd, "stats", "mpl")

	// Now, let's go for the verbose mode
	verboseAns := checkCmdAnswerKeysRegexp(test, projectTestsStatsNoVerboseLineNb+4+nbApps, regexpMap, prjsCommand, prjsTestsCmd, "stats", "mpl", "-v")

	// Create the regexp map we will use
	regexpMapCol := map[string]string{
		"ID":        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":      `^mystère-abyssal-[0-9]+$`,
		"Package":   `^mystère-abyssal-[0-9]+$`,
		"Succeeded": `^[0-9]+$`,
		"Failed":    `^[0-9]+$`,
		"Cancelled": `^[0-9]+$`,
		"Total":     `^5$`,
	}

	// Get the second part of the answer
	columnsAns := strings.Split(verboseAns, prjsTestsStatsCmdAnsLine5)[1]
	// Trim it
	columnsAns = strings.TrimSpace(columnsAns)

	// Check the answer of the command
	checkObjectKeysRegexpCol(test, columnsAns, projectTestsStatsVerboseColNb, regexpMapCol)
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// createAndReleaseProject is a function allowing to create and release a project easily
// The calling function needs to specify the name of the reference project
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param prjName: Name of the reference project
// :type prjName: string
// :param usersTab: Tab of users wanted as audience for the release
// :type usersTab: []string
// :param teamsTab: Tab of teams wanted as audience for the release
// :type teamsTab: []string
// :param factory: True if the release is for the factory
// :type factory: bool
func createAndReleaseProject(test *testing.T, prjName string, usersTab, teamsTab []string, factory bool) {
	// From the name, create the slug
	prjSlug := strings.ReplaceAll(strings.ToLower(prjName), " ", "-")

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd(prjName, "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, prjSlug, cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command
	cmdToRun2 := getMainAddAppCmd((prjSlug + "-application"), (prjSlug + "-application"), prjSlug)

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, (prjSlug + "-application"), cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the basic command line
	cmdLine := []string{prjsCommand, "release", prjSlug, "--distribution", "redpesk-lts-arz", "-v", "1.0"}

	// Complete with potential audience
	for _, user := range usersTab {
		cmdLine = append(cmdLine, "-u", user)
	}
	for _, team := range teamsTab {
		cmdLine = append(cmdLine, "--team", team)
	}
	if factory {
		cmdLine = append(cmdLine, "-s")
	}

	// Try to release the project
	checkActionCmdAns(test, cmdLine...)
}

// getMainAddPrjCmd is an initializer function to get the mainly
// used command to create a project
//
// :param name: Name of the project to create
// :type name: string
// :param distro: Name of the distribution for the project
// :type distro: string
// :returns: []string -- the value of the wanted constant
func getMainAddPrjCmd(name, distro string) []string {
	return []string{
		prjsCommand,
		prjsAddCmd,
		"-n",
		name,
		"-d",
		"This is a test project",
		"--mandatory-arch",
		"aarch64",
		"--optional-arch",
		"x86_64",
		"--mandatory-distro",
		distro,
	}
}

// used command to create a project
//
// :param name: Name of the project to create
// :type name: string
// :param distro: Name of the distribution for the project
// :type distro: string
// :returns: []string -- the value of the wanted constant
func getAddPrjCmdOptDistro(name, distro string) []string {
	return []string{
		prjsCommand,
		prjsAddCmd,
		"-n",
		name,
		"-d",
		"This is a test project",
		"--optional-arch",
		"x86_64",
		"--optional-distro",
		distro,
	}
}

// getUpdateArchPrjCmd is an initializer function to get the mainly
// used command to update the architecture of a project
//
// :param slug: Slug of the project to update
// :type slug: string
// :returns: []string -- the value of the wanted constant
func getUpdatePrjCmd(slug string, flag string, value string) []string {
	return []string{
		prjsCommand,
		"update",
		slug,
		flag,
		value,
	}
}

// getUpdateDescPrjCmd is an initializer function to get the mainly
// used command to update the description of a project
//
// :param slug: Slug of the project to update
// :type slug: string
// :param desc: Description of the project to update
// :type desc: string
// :returns: []string -- the value of the wanted constant
func getUpdateDescPrjCmd(slug string, desc string) []string {
	return []string{
		prjsCommand,
		"update",
		slug,
		"-d",
		desc,
	}
}

// getUpdateDistrPrjCmd is an initializer function to get the mainly
// used command to update the distribution of a project
//
// :param slug: Slug of the project to update
// :type slug: string
// :returns: []string -- the value of the wanted constant
func getUpdateDistrPrjCmd(slug string) []string {
	return []string{
		prjsCommand,
		"update",
		slug,
		"--mandatory-distro",
		"",
		"--optional-distro",
		"redpesk-devel-arz",
	}
}

// getUpdateArchPrjCmd is an initializer function to get the mainly
// used command to update the architecture of a project
//
// :param slug: Slug of the project to update
// :type slug: string
// :returns: []string -- the value of the wanted constant
func getUpdateArchPrjCmd(slug string) []string {
	return []string{
		prjsCommand,
		"update",
		slug,
		"--mandatory-arch",
		"x86_64",
	}
}

// getMainAddMemberCmd is an initializer function to get the mainly
// used command to add a member in a project
//
// :param project: Name of the project
// :type project: string
// :param guest: Guest to add
// :type guest: string
// :param roles: Roles for the new members
// :type roles: []string
// :returns: []string -- the value of the wanted constant
func getMainAddMemberCmd(project string, guest string, roles []string) []string {
	resultStr := []string{
		prjsCommand,
		"members",
		"add",
		project,
		"-g",
		guest,
	}

	for _, role := range roles {
		resultStr = append(resultStr, "-r", role)
	}

	return resultStr
}

// getMainUpdateMemberCmd is an initializer function to get the mainly
// used command to update a member in a project
//
// :param project: Name of the project
// :type project: string
// :param member: Member to modify
// :type member: string
// :param roles: Roles for the new members
// :type roles: []string
// :returns: []string -- the value of the wanted constant
func getMainUpdateMemberCmd(project string, member string, roles []string) []string {
	resultStr := []string{
		prjsCommand,
		"members",
		"update",
		project,
		"-m",
		member,
	}

	for _, role := range roles {
		resultStr = append(resultStr, "-r", role)
	}

	return resultStr
}

// checkMemberIsInPrj is a function allowing to check if a given member
// is in a given project
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param project: Slug of the project
// :type project: string
// :param member: Pseudo of the member
// :type member: string
// :param rolesToCheck: Roles to check for the member, can be empty
// :type rolesToCheck: string
// :returns: bool -- True if the member is in the project, False in the other case
func checkMemberIsInPrj(test *testing.T, project string, member string, rolesToCheck string) bool {
	// Create the command to run
	cmdToRun := []string{prjsCommand, "members", "get", project}

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Pseudo": `^[a-z0-9-_]+$`,
		"Roles":  `(manager|tester|developer)`,
	}

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(cmdToRun...)
	if err != nil {
		test.Fatal(err)
		return false
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check the answer of the command
	checkObjectKeysRegexpCol(test, cmdOutputTrimmed, prjMembersDisplayNoVerboseColNb, regexpMap)

	// Check that in the answer, the member exists, and that he has the right roles
	result := strings.Contains(cmdOutputTrimmed, member)
	if result && (rolesToCheck != "") {
		allLines := strings.Split(cmdOutputTrimmed, "\n")
		for _, line := range allLines {
			if strings.HasPrefix(line, member) {
				// Get each column of the current line
				columnsValues := getEachColumnFromLine(line)
				// Check that the value match the regexp
				result = result && assert.Regexp(test, regexp.MustCompile(rolesToCheck), columnsValues[1])
				break
			}
		}
	}

	return result
}

// checkMemberIsNotInPrj is a function allowing to check that a given member
// is not in a given project
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param project: Slug of the project
// :type project: string
// :param member: Pseudo of the member
// :type member: string
// :param rolesToCheck: Roles to check for the member, can be empty
// :type rolesToCheck: string
// :returns: bool -- True if the member is in the project, False in the other case
func checkMemberIsNotInPrj(test *testing.T, project string, member string) bool {
	// Create the command to run
	cmdToRun := []string{prjsCommand, "members", "get", project}

	// Check that the answer from rp-cli is correct

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(cmdToRun...)
	if err != nil {
		test.Fatal(err)
		return false
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check that in the answer, the member exists
	result := strings.Contains(cmdOutputTrimmed, member)

	return !result
}

// --------------------------------------------------------------
// --------------------------------------------------------------
