/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/
package cmd

import (
	"errors"
	"fmt"
	"sort"
	"text/tabwriter"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	"github.com/jinzhu/copier"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
)

// -------------------------------------------------
// Global variable (inside this package) -----------
// -------------------------------------------------

// Teams cached
var cachedRPTeamsByUUID map[uuid.UUID]rpapiv1.RPTeam
var cachedRPTeamsBySlug map[string]uuid.UUID

// -------------------------------------------------
// -------------------------------------------------

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

func initTeamsCmd(rootCmd *cobra.Command) {

	// teamsCmd represents the teams command
	var teamsCmd = &cobra.Command{
		Use:   teamsCommand,
		Short: teamsCmdShortDesc,
	}

	// teamsListsCmd represents the "list" command for users
	var teamsListCmd = &cobra.Command{
		Use:          teamsListCmdName,
		SilenceUsage: true,
		Aliases:      []string{teamsListCmdAlias},
		Short:        teamsListCmdShortUsage,
		RunE:         getTeamsList,
	}

	// teamsGetCmd represents the "get" command
	var teamsGetCmd = &cobra.Command{
		Use:               teamGetCmdName,
		Short:             teamGetCmdShortUsage,
		Long:              teamGetCmdLongUsage,
		SilenceUsage:      true,
		RunE:              getTeam,
		ValidArgsFunction: getListTeamsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// teamsAddCmd represents the "add" command
	var teamsAddCmd = &cobra.Command{
		Use:          teamAddCmdName,
		Short:        teamAddCmdShortUsage,
		SilenceUsage: true,
		PreRunE:      initEventsTeams,
		RunE:         createTeam,
		Annotations:  map[string]string{"__bck_ws_connect__": "TRUE"},
	}

	// teamsUpdateCmd represents the "update" command
	var teamsUpdateCmd = &cobra.Command{
		Use:               teamUpdateCmdName,
		Short:             teamUpdateCmdShortUsage,
		Long:              teamUpdateCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventsTeams,
		RunE:              updateTeam,
		ValidArgsFunction: getListTeamsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// teamsInviteCmd represents the "invite" command
	var teamsInviteCmd = &cobra.Command{
		Use:               teamInviteCmdName,
		Short:             teamInviteCmdShortUsage,
		Long:              teamInviteCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventInvitation,
		RunE:              InviteInTeam,
		ValidArgsFunction: getListTeamsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// teamsDelMemberCmd represents the "del-member" command
	var teamsDelMemberCmd = &cobra.Command{
		Use:               teamDelMemberCmdName,
		Short:             teamDelMemberCmdShortUsage,
		Long:              teamDelMemberCmdLongUsage,
		SilenceUsage:      true,
		RunE:              DelFromTeam,
		ValidArgsFunction: getListTeamsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	// Define verbose flag for all the commands that use it
	teamsListCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	teamsGetCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Add the flags necessary for the create/update command
	// Creation
	teamsAddCmd.Flags().StringP(teamAddCmdNameFlag, teamAddCmdNameFlagAlias, "", teamAddCmdNameFlagUsage)
	teamsAddCmd.Flags().StringP(teamAddCmdDescFlag, teamAddCmdDescFlagAlias, "", teamAddCmdDescFlagUsage)
	teamsAddCmd.Flags().Bool(teamAddCmdPublicFlag, false, teamAddCmdPublicFlagUsage)
	// Update
	teamsUpdateCmd.Flags().StringP(teamAddCmdNameFlag, teamAddCmdNameFlagAlias, flagUpdateKeepCurrentValue, teamAddCmdNameFlagUsage)
	teamsUpdateCmd.Flags().StringP(teamAddCmdDescFlag, teamAddCmdDescFlagAlias, flagUpdateKeepCurrentValue, teamAddCmdDescFlagUsage)
	teamsUpdateCmd.Flags().String(teamAddCmdPublicFlag, flagUpdateKeepCurrentValue, teamAddCmdPublicFlagUsage)

	// Add the guest flag to the invite command
	teamsInviteCmd.Flags().StringP(teamInviteCmdGuestFlag, teamInviteCmdGuestFlagAlias, "", teamInviteCmdGuestFlagUsage)

	// Add the member flag to the del-member command
	teamsDelMemberCmd.Flags().StringP(teamDelMemberCmdMemberFlag, teamDelMemberCmdMemberFlagAlias, "", teamDelMemberCmdMemberFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required
	teamsAddCmd.MarkFlagRequired(teamAddCmdNameFlag)
	teamsInviteCmd.MarkFlagRequired(teamInviteCmdGuestFlag)
	teamsDelMemberCmd.MarkFlagRequired(teamDelMemberCmdMemberFlag)

	// Specify dynamic flag completion
	teamsInviteCmd.RegisterFlagCompletionFunc(teamInviteCmdGuestFlag, getMembersToInviteAsTab)
	teamsDelMemberCmd.RegisterFlagCompletionFunc(teamDelMemberCmdMemberFlag, getTeamMembersAsTab)

	// Add the command "teams" as a subcommand of rp-cli
	rootCmd.AddCommand(teamsCmd)
	// Add the command "list" as a subcommand of teamsCmd
	teamsCmd.AddCommand(teamsListCmd)
	// Add the command "get" as a subcommand of teamsCmd
	teamsCmd.AddCommand(teamsGetCmd)
	// Add the command "add" as a subcommand of teamsCmd
	teamsCmd.AddCommand(teamsAddCmd)
	// Add the command "update" as a subcommand of teamsCmd
	teamsCmd.AddCommand(teamsUpdateCmd)
	// Add the command "invite" as a subcommand of teamsCmd
	teamsCmd.AddCommand(teamsInviteCmd)
	// Add the command "del-member" as a subcommand of teamsCmd
	teamsCmd.AddCommand(teamsDelMemberCmd)
}

// initEventsTeams inits everything that is needed for the team
// creation and update
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventsTeams(cmd *cobra.Command, args []string) error {
	// Init the events list
	eventsList := []string{
		rpapiv1.EVTTeamAdd,
		rpapiv1.EVTTeamStateChange,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// initEventInvitation inits everything that is needed for invitation
// in teams
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventInvitation(cmd *cobra.Command, args []string) error {
	return CLIRegisterToMultipleEvents([]string{rpapiv1.EVTInvitationSent})
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getListTeamsAsTab gets all the teams available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListTeamsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	teamsList := []string{}

	// Get all teams list (available for current user)
	allTeams := rpapiv1.RPTeams{}
	if err := bckAPI.TeamsGetList(&allTeams); err != nil {
		return teamsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the team lists, fill the teamsList tab
	for _, team := range allTeams.Teams {
		teamsList = append(teamsList, team.Slug)
	}

	// Finally sort the string tab
	sort.Strings(teamsList)

	return teamsList, cobra.ShellCompDirectiveNoFileComp
}

// getMembersToInviteAsTab is used when the user wants to send an
// invitation either to a user or to a public team. Therefore,
// it lists all the public teams and all the users available in the
// Redpesk instance.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getMembersToInviteAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	finalList := []string{}

	// Try to get the team id passed by the user
	teamID, err := getFirstArg(args, teamIDNotSetMsg, teamIDRequiredMsg)
	if err != nil {
		return finalList, cobra.ShellCompDirectiveNoFileComp
	}

	// Get the team UUID from the slug
	teamUUID, err := getTeamUUID(teamID)
	if err != nil {
		return finalList, cobra.ShellCompDirectiveNoFileComp
	}

	// Get all the possible members for the team
	possibleMembers := rpapiv1.RPMembers{}
	if err := bckAPI.TeamGetMembersToInvite(teamUUID.String(), &possibleMembers); err != nil {
		return finalList, cobra.ShellCompDirectiveNoFileComp
	}

	// Fill the slice to return
	for _, member := range possibleMembers {
		finalList = append(finalList, member.Pseudo)
	}

	return finalList, cobra.ShellCompDirectiveNoFileComp
}

// getTeamMembersAsTab is used to get the list of members of a team
// specified as first argument by the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getTeamMembersAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	finalList := []string{}

	// Try to get the team id passed by the user
	teamID, err := getFirstArg(args, teamIDNotSetMsg, teamIDRequiredMsg)
	if err != nil {
		return finalList, cobra.ShellCompDirectiveNoFileComp
	}

	// Get the team UUID from the slug
	teamUUID, err := getTeamUUID(teamID)
	if err != nil {
		return finalList, cobra.ShellCompDirectiveNoFileComp
	}

	// Get the team
	team := rpapiv1.RPTeam{}
	if err := bckAPI.TeamGet(teamUUID.String(), &team); err != nil {
		return finalList, cobra.ShellCompDirectiveNoFileComp
	}

	// Fill the slice to return
	for _, member := range team.Members {
		finalList = append(finalList, member.Pseudo)
	}

	return finalList, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// getTeamsList gets all the teams available in the Redpesk
// instance and prints it.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func getTeamsList(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Prepare the function needed by the printer to get the teams
	getTeamsCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		teams := rpapiv1.RPTeams{}
		if err := bckAPI.TeamsGetListByPage(&teams, pageNb, perPage); err != nil {
			return teams.Teams, printer.RPPrinterPaging{}, err
		}
		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&teams)

		return teams.Teams, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: teamsListVerboseLineNb,
		GetObjCB:      getTeamsCB,
		PrintFancyCB:  displayTeams,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct)
}

// getTeam gets information about a team in the Redpesk
// instance and prints it.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func getTeam(cmd *cobra.Command, args []string) error {

	// Get the team-ID passed as an argument and return an error if there is nothing
	teamID, err := getFirstArg(args, teamIDNotSetMsg, teamIDRequiredMsg)
	if err != nil {
		return err
	}

	// Get the team UUID
	teamUUID, err := getTeamUUID(teamID)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the team
	team := rpapiv1.RPTeam{}
	if err := bckAPI.TeamGet(teamUUID.String(), &team); err != nil {
		return err
	}

	// Display the team
	RPPrinter.CustomPrintObject(team, verbose, displayOneTeam)

	return nil
}

// createTeam creates a new team in the Redpesk instance
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func createTeam(cmd *cobra.Command, args []string) error {

	// Display the first line to the user
	RPPrinter.CustomPrintLn(teamAddCmdAnsLine0)

	// Create the structure to fill
	team := rpapiv1.RPTeam{}

	// Fill the structure with the right values (given in command line)
	if err := prepareTeamStruct(cmd, &team, true); err != nil {
		return err
	}

	// Log the team creation
	Log.Infof(teamCreationInfoMsg, team.Name)
	Log.Debugf(teamCreationDebugMsg, team)

	// Tell the user that the team is currently being created
	RPPrinter.CustomPrintF(teamAddCmdAnsLine1, team.Name)

	// Request the creation of this team to the backend
	teamAns := rpapiv1.RPTeam{}
	if err := bckAPI.TeamCreate(&team, &teamAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(teamAns, false, displayNothing)

	// Create the sentence to display when the object will be added
	slugSentence := fmt.Sprintf(objectSlugDisplayMsg, "team", teamAns.Slug)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		// Display the slug of the created object
		RPPrinter.CustomPrintLn(slugSentence)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a team creation
		if rEvent.Type == rpapiv1.EVTTeamAdd {
			// Function handling the events
			rTeam, err := rEvent.DecodeTeam()
			if err != nil {
				return err
			}

			if utils.UUIDEqual(rTeam.ID, teamAns.ID) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rTeam, false, displayNothing)
				// Display the slug of the created object
				RPPrinter.CustomPrintLn(slugSentence)
				break
			}
		}
	}

	return nil
}

// updateTeam update the information of a team in the Redpesk
// instance
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func updateTeam(cmd *cobra.Command, args []string) error {

	// Display the first line to the user
	RPPrinter.CustomPrintLn(teamUpdateCmdAnsLine0)

	// Get the team-ID passed as an argument and return an error if there is nothing
	teamID, err := getFirstArg(args, teamIDNotSetMsg, teamIDRequiredMsg)
	if err != nil {
		return err
	}

	// Get the team UUID
	teamUUID, err := getTeamUUID(teamID)
	if err != nil {
		return err
	}

	// Get the team that the user wants to modify
	team := rpapiv1.RPTeam{}
	if err := bckAPI.TeamGet(teamUUID.String(), &team); err != nil {
		return err
	}

	// Fill the structure with the right values (given in command line)
	if err := prepareTeamStruct(cmd, &team, false); err != nil {
		return err
	}

	// Log the team update
	Log.Infof(teamUpdateInfoMsg, team.Name)
	Log.Debugf(teamUpdateDebugMsg, team)

	// Tell the user that the team is currently being updated
	RPPrinter.CustomPrintF(teamUpdateCmdAnsLine1, team.Name)

	// Request the update of this team to the backend
	teamAns := rpapiv1.RPTeam{}
	if err := bckAPI.TeamUpdate(&team, &teamAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(teamAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a team creation
		if rEvent.Type == rpapiv1.EVTTeamStateChange {
			// Function handling the events
			rTeam, err := rEvent.DecodeTeam()
			if err != nil {
				return err
			}

			if utils.UUIDEqual(rTeam.ID, teamAns.ID) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rTeam, false, displayNothing)
				break
			}
		}
	}

	return nil
}

// InviteInTeam allows to invite an existing user or team
// in another team
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func InviteInTeam(cmd *cobra.Command, args []string) error {

	// Display the first line to the user
	RPPrinter.CustomPrintLn(teamInviteCmdAnsLine0)

	// Get the team-ID passed as an argument and return an error if there is nothing
	teamID, err := getFirstArg(args, teamIDNotSetMsg, teamIDRequiredMsg)
	if err != nil {
		return err
	}

	// Get the team UUID
	teamUUID, err := getTeamUUID(teamID)
	if err != nil {
		return err
	}

	// Get the guest flag value
	guest, err := cmd.Flags().GetString(teamInviteCmdGuestFlag)
	if err != nil {
		logFlagError(teamInviteCmdGuestFlag)
		return err
	}
	if guest == "" {
		Log.Error(teamGuessFlagNotSetShortMsg)
		return errors.New(teamGuessFlagNotSetLongMsg)
	}

	// Get the guest member structure
	guestMember, err := getGuestForTeam(teamUUID.String(), guest)
	if err != nil {
		return err
	}

	// Give details to the user about guest and host
	RPPrinter.CustomPrintF(teamInviteCmdAnsLine1, guest, teamID)

	// Request the backend to send the invitation
	invitAns := rpapiv1.RPInvitResult{}
	if err := bckAPI.TeamRequestInvitation(teamUUID.String(), &guestMember, &invitAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(invitAns, false, displayNothing)

	// Check that the invitation has been correctly sent
	mbrCorrectlyAdded := false
	for _, member := range invitAns.Team.Members {
		if utils.UUIDEqual(member.ID, guestMember.ID) {
			mbrCorrectlyAdded = true
			break
		}
	}
	if mbrCorrectlyAdded {
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	} else {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return errors.New(noUsersInvitedMsg)
	}

	return nil
}

// DelFromTeam allows to delete a member from a team.
// The member can be an user or a team
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func DelFromTeam(cmd *cobra.Command, args []string) error {
	// Display the first line to the user
	RPPrinter.CustomPrintLn(teamDelMemberCmdAnsLine0)

	// Get the team-ID passed as an argument and return an error if there is nothing
	teamID, err := getFirstArg(args, teamIDNotSetMsg, teamIDRequiredMsg)
	if err != nil {
		return err
	}

	// Get the team UUID
	teamUUID, err := getTeamUUID(teamID)
	if err != nil {
		return err
	}

	// Get the guest flag value
	member, err := cmd.Flags().GetString(teamDelMemberCmdMemberFlag)
	if err != nil {
		logFlagError(teamDelMemberCmdMemberFlag)
		return err
	}

	// Get the full team object
	teamObj := rpapiv1.RPTeam{}
	if err := bckAPI.TeamGet(teamUUID.String(), &teamObj); err != nil {
		return err
	}

	// Find the member UUID
	memberUUID := uuid.Nil
	for _, memberObj := range teamObj.Members {
		// Check if the current member is the one we are looking for
		if (memberObj.ID.String() == member) ||
			(memberObj.Pseudo == member) {
			// This is our member!
			memberUUID = memberObj.ID
			break
		}
	}

	// Check that we found the member UUID in the team
	if memberUUID == uuid.Nil {
		errorMsg := fmt.Sprintf(noMemberInTeamMsg, member, teamID)
		return errors.New(errorMsg)
	}

	// Give details to the user about member and team
	RPPrinter.CustomPrintF(teamDelMemberCmdAnsLine1, member, teamID)

	// Request the backend to delete the member the invitation
	ansTeamObj := rpapiv1.RPTeam{}
	if err := bckAPI.TeamDeleteMember(teamUUID.String(), memberUUID.String(), &ansTeamObj); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Check that the member is not in the team anymore
	memberFound := false
	for _, memberObj := range ansTeamObj.Members {
		// Check if the current member is the one we are looking for
		if utils.UUIDEqual(memberObj.ID, memberUUID) {
			memberFound = true
			break
		}
	}

	if memberFound {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return errors.New(teamMemberNotDeleted)
	} else {
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// prepareTeamStruct prepare a team structure to send to the backend
// to create or update a team
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param team: Team structure to fill
// :type team: *rpapiv1.RPTeam
// :param isCreation: Indicate if we are creating the team
// :type isCreation: bool
// :returns: pointer to the RPAPIInterface object
func prepareTeamStruct(cmd *cobra.Command, team *rpapiv1.RPTeam, isCreation bool) error {

	// Get the name of the team wanted by the user
	teamName, err := cmd.Flags().GetString(teamAddCmdNameFlag)
	if err != nil {
		logFlagError(teamAddCmdNameFlag)
		return err
	}
	if isCreation && teamName == "" {
		Log.Error(teamNameFlagNotSetShortMsg)
		return errors.New(teamNameFlagNotSetLongMsg)
	}

	// Get the description of the team wanted by the user (optional)
	teamDesc, err := cmd.Flags().GetString(teamAddCmdDescFlag)
	if err != nil {
		logFlagError(teamAddCmdDescFlag)
		return err
	}

	// Get the public flag for the team
	var isTeamPublic bool
	isTeamPublic, err = genericGetBoolFlagValue(cmd, isCreation, (team.Status == "public"), teamAddCmdPublicFlag,
		fmt.Sprintf("%s \"%s\" flag", pbFlagMainPrefix, teamAddCmdPublicFlag))
	if err != nil {
		return err
	}

	if isTeamPublic {
		team.Status = "public"
	} else {
		team.Status = "private"
	}

	// Fill the team structure
	if isCreation {
		team.Name = teamName
		team.Description = teamDesc
	} else {
		if teamName != flagUpdateKeepCurrentValue {
			team.Name = teamName
		}
		if teamDesc != flagUpdateKeepCurrentValue {
			team.Description = teamDesc
		}
	}

	// Everything went well, return nil
	return nil
}

// getTeamUUID is a private function allowing to retrieve the
// UUID of a team from a given identifier. This identifier can
// be the UUID itself as a string representation or the Slug of the
// team
//
// :param identifier: the identifier of the team
// :type identifier: string
// :returns: (uuid.UUID, error) -- the UUID of the team and an error if something went bad
func getTeamUUID(identifier string) (uuid.UUID, error) {

	// Optimization: pass through the cache in order to avoid several requests for Teams list
	if (cachedRPTeamsByUUID == nil) || (cachedRPTeamsBySlug == nil) || (runMode == "testing") {
		err := populateTeamsCache()
		if err != nil {
			return uuid.Nil, err
		}
	}

	// Check if we can find the identifier in the different map
	teamUUID, ok := cachedRPTeamsBySlug[identifier]
	if ok {
		// The identifier is the slug and we found the uuid !
		return teamUUID, nil
	}

	tmpUUID, err := uuid.FromString(identifier)
	if err == nil {
		_, ok = cachedRPTeamsByUUID[tmpUUID]
		if ok {
			// The identifier is the UUID and we found it !
			return tmpUUID, nil
		}
	}

	// If we reached this part of code, we did not find anything
	return uuid.Nil, fmt.Errorf(teamIdentifierNotFound, identifier)
}

// populateTeamsCache is a function that needs to be called when
// teams will be used in the user request. The function fills the
// global variable "cachedRPTeamsByUUID" and "cachedRPTeamsBySlug"
// from a request to the Redpesk backend.
//
// :returns: error -- An error if something went bad
func populateTeamsCache() error {

	// Get all teams list
	allTeams := rpapiv1.RPTeams{}
	if err := bckAPI.TeamsGetList(&allTeams); err != nil {
		return err
	}

	// Create the map if it does not exists
	if (cachedRPTeamsByUUID == nil) || (runMode == "testing") {
		cachedRPTeamsByUUID = make(map[uuid.UUID]rpapiv1.RPTeam)
	}

	// Create the map if it does not exists
	if (cachedRPTeamsBySlug == nil) || (runMode == "testing") {
		cachedRPTeamsBySlug = make(map[string]uuid.UUID)
	}

	// We got correctly the teams list, fill the cache now
	for _, team := range allTeams.Teams {

		// Copy each team in the cache
		tmpTeam := rpapiv1.RPTeam{}
		copier.Copy(&tmpTeam, &team)
		cachedRPTeamsByUUID[team.ID] = tmpTeam

		// Fill the other cache
		cachedRPTeamsBySlug[team.Slug] = team.ID
	}

	return nil
}

// getGuestForTeam is a private function allowing to retrieve the
// full RPMember structure of a guest from a given identifier.
// This identifier can be the UUID itself as a string representation or
// the pseudo of the guest
//
// :param teamUUID: UUID of the targetted Team
// :type teamUUID: string
// :param identifier: the identifier of the guest
// :type identifier: string
// :returns: (rpapiv1.RPMember, error) -- the full rpapiv1.RPMember to invite and an error if something went bad
func getGuestForTeam(teamUUID string, identifier string) (rpapiv1.RPMember, error) {
	uuidIndex := -1
	memberResult := rpapiv1.RPMember{}

	// Get all the possible members for the team
	possibleMembers := rpapiv1.RPMembers{}
	if err := bckAPI.TeamGetMembersToInvite(teamUUID, &possibleMembers); err != nil {
		return memberResult, err
	}

	// Search inside the possible members array
	for index, member := range possibleMembers {
		if (member.ID.String() == identifier) || (member.Pseudo == identifier) {
			uuidIndex = index
			break
		}
	}

	// Check that we found the member to invite
	if uuidIndex == -1 {
		return memberResult, fmt.Errorf(teamGuestIdentifierNotFound, identifier)
	}

	return possibleMembers[uuidIndex], nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// displayTeams is a private function allowing to display
// a slice of teams in the console
//
// :param tms: tab of teams to display
// :type tms: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayTeams(tms interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	teams := tms.([]rpapiv1.RPTeam)

	// Display result
	first := true

	for _, team := range teams {
		if verbose {
			if !first {
				fmt.Fprintln(writer)
			}

			// Print this member
			_printOneTeam(&team, verbose, false, writer)

		} else {
			if first {
				manageHeaderPrint(teamsListCmdHeader, writer)
			}

			// Print this member
			_printOneTeam(&team, verbose, false, writer)
		}
		first = false
	}
	return nil
}

// displayOneTeam is a private function allowing to display
// one team in the console
//
// :param tm: Pointer to the team to display
// :type tm: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayOneTeam(tm interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	team := tm.(rpapiv1.RPTeam)
	// Print the team
	_printOneTeam(&team, verbose, true, writer)
	return nil
}

// _printOneTeam is a private function allowing to print one team in the console
//
// :param team: pointer to the team to print
// :type team: *rpapiv1.RPTeam
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isAlone: True if there is only one team to display in the context where this function is called
// :type isAlone: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
func _printOneTeam(team *rpapiv1.RPTeam, verbose, isAlone bool, writer *tabwriter.Writer) {
	if verbose {
		// Aggregate the members
		membersStr := ""
		for index, member := range team.Members {
			// Add a comma only if it is not the first iteration
			if index > 0 {
				membersStr = membersStr + "\n"
			}
			membersStr = membersStr + (member.Pseudo + "\t " + member.Email + "\t " + string(member.InvitationStatus))
		}

		// Create the string that can potentially be empty
		description := team.Description
		if description == "" {
			description = mainNotSetMsg
		}

		// Display the different strings describing the team
		fmt.Fprintln(writer, teamsListCmdVerboseLine1, team.ID.String())
		fmt.Fprintln(writer, teamsListCmdVerboseLine7, team.Slug)
		fmt.Fprintln(writer, teamsListCmdVerboseLine2, team.Name)
		fmt.Fprintln(writer, teamsListCmdVerboseLine3, team.Status)
		fmt.Fprintln(writer, teamsListCmdVerboseLine4, description)
		fmt.Fprintln(writer, teamsListCmdVerboseLine5, team.UserOwnerPseudo)
		fmt.Fprintln(writer, teamsListCmdVerboseLine9, team.NbUserMembers)

		if isAlone {
			fmt.Fprintln(writer)
			fmt.Fprintln(writer, teamsListCmdVerboseLine6)
			fmt.Fprintln(writer)
			fmt.Fprintln(writer, teamsListCmdVerboseLine8)
			fmt.Fprintln(writer, membersStr)
		}

	} else {
		if isAlone {
			fmt.Fprintln(writer, teamsListCmdHeader)
		}
		fmt.Fprintln(writer, team.ID.String(), "\t", team.Slug, "\t", team.Name, "\t", team.Status, "\t", team.NbUserMembers)
	}
}

// --------------------------------------------------------------
// --------------------------------------------------------------
