/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"encoding/base64"
	"encoding/json"
	"errors"
	"fmt"
	"sort"
	"strconv"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
)

// -------------------------------------------------
// Types definition --------------------------------
// -------------------------------------------------

// Sort applications by Slug
type _AppByLabel []rpapiv1.RPApplication

// -------------------------------------------------
// -------------------------------------------------

// Applications tests cached
var cachedAppRPTests []rpapiv1.RPTest

func (s _AppByLabel) Len() int      { return len(s) }
func (s _AppByLabel) Swap(i, j int) { s[i], s[j] = s[j], s[i] }
func (s _AppByLabel) Less(i, j int) bool {
	if s[i].Parent.ProjectSlug == s[j].Parent.ProjectSlug {
		return s[i].Slug < s[j].Slug
	} else {
		return s[i].Parent.ProjectSlug < s[j].Parent.ProjectSlug
	}
}

// ApplicationsListGetOfTypes gets the list of all existing application if no projectID is given.
// If the parameter projectID is different than "", it retrieves the apps in this project only.
// The type of applications to retrieve can be specified in a slice of string
//
// :param apps: Pointer to the rpapiv1.RPApplications structure to fill
// :type apps: *rpapiv1.RPApplications
// :param projectID: ID of the project from which the applications are wanted
// :type projectID: string
// :param types: Types of project to retrieve
// :type types: []string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationsListGetOfTypes(apps *rpapiv1.RPApplications, projectID string, types []string) error {

	// Construct the string containing all the types to retrieve, by default the value is "standard"
	typesStr := "standard"

	if len(types) > 0 {
		// Get rid of the default value
		typesStr = ""
		// Construct the type string
		for index, currentType := range types {
			typesStr += currentType
			if index < (len(types) - 1) {
				typesStr += ","
			}
		}
	}

	// Initialize the call-back to save current page of apps
	savingCB := func(tmpObjects []byte) {
		tmpApps := rpapiv1.RPApplications{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpApps)
		// Add the applications in the final object
		apps.Apps = append(apps.Apps, tmpApps.Apps...)
	}

	// Decide which route to use
	routeToCall := ""
	if projectID == "" {
		routeToCall = allAppsRoute
	} else {
		routeToCall = projectRoutePrefix + projectID + allAppsRoute
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Call the function allowing to get all the object
	err := pagingMngt.getAllObjects(rpAPIInterface, routeToCall, typesStr, nil, savingCB)
	if err != nil {
		return errorInterpreter(err, appObjectType)
	}

	sort.Sort(_AppByLabel(apps.Apps))

	return nil
}

// ApplicationsListByTypePage Get the list of existing apps with the
// page number indicated. If a projectID is specified, it retrieves the apps in this project only.
// The type of applications to retrieve can be specified in a slice of string
//
// :param apps: Pointer to the rpapiv1.RPApplications structure to fill
// :type apps: *rpapiv1.RPApplications
// :param projectID: ID of the project from which the applications are wanted
// :type projectID: string
// :param types: Types of application to retrieve
// :type types: []string
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :param bookmarked: Set to true if only the bookmarked apps are wanted
// :type bookmarked: boolean
// :returns: error -- an error if a problem occurred
func (rpAPIInterface *RPBCKAPI) ApplicationsListByTypePage(apps *rpapiv1.RPApplications, projectID string, types []string, page, perPage int, bookmarked bool) error {
	// Construct the string containing all the types to retrieve, by default the value is "standard"
	typesStr := "standard"

	if len(types) > 0 {
		// Get rid of the default value
		typesStr = ""
		// Construct the type string
		for index, currentType := range types {
			typesStr += currentType
			if index < (len(types) - 1) {
				typesStr += ","
			}
		}
	}

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
		"type":     typesStr,
	}

	if bookmarked {
		queryParamMap["filters"] = "bookmarked==true"
	}

	// Decide which route to use
	routeToCall := ""
	if projectID == "" {
		routeToCall = allAppsRoute
	} else {
		routeToCall = projectRoutePrefix + projectID + allAppsRoute
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(routeToCall, &apps, queryParamMap); err != nil {
		return errorInterpreter(err, appObjectType)
	}

	return nil
}

// ApplicationsSamplesListGet gets the list of all existing applications samples
//
// :param apps: Pointer to the rpapiv1.RPSampleApplications structure to fill
// :type apps: *rpapiv1.RPSampleApplications
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationsSamplesListGet(apps *[]rpapiv1.RPSampleApplication) error {

	// Send the appropriate GET request
	if err := rpAPIInterface.HTTPCli.Get(appSamplesRoute, apps); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ApplicationGet gets the information available for a given application
//
// :param id: id of the project wanted
// :type id: string
// :param app: Pointer to the rpapiv1.RPApplicationFull structure to fill
// :type app: *rpapiv1.RPApplicationFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationGet(id string, app *rpapiv1.RPApplicationFull) error {
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(appIDRoute, id), app); err != nil {
		return errorInterpreter(err, appObjectType)
	}
	return nil
}

// ApplicationGetListFiles gets the list of the files for the given application
//
// :param id: id of the project wanted
// :type id: string
// :param list: Pointer to the rpapiv1.RPFileList structure to fill
// :type list: *rpapiv1.RPFileList
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationGetListFiles(id string, app *rpapiv1.RPFileList) error {
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(appListFilesRoute, id), app); err != nil {
		return errorInterpreter(err, appObjectType)
	}
	return nil
}

// ApplicationRemoveFile asks the Redpesk backend to delete an existing file
//
// :param appID: ID of the application to delete
// :type appID: string
// :param filename: name of the file to remove
// :type filename: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationRemoveFile(appID string, filename string) error {
	queryParamMap := map[string]string{
		"name": filename,
	}
	if err := rpAPIInterface.HTTPCli.Delete(fmt.Sprintf(appRemoveFileRoute, appID), nil, queryParamMap); err != nil {
		return errorInterpreter(err, appObjectType)
	}
	return nil
}

// ApplicationDownloadFile get a file from the redpesk back end
//
// :param appID: ID of the application to delete
// :type appID: string
// :param filename: name of the file to remove
// :type filename: string
// :param filename: name of the file to remove
// :type filename: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationDownloadFile(appID string, filename string, file *rpapiv1.RPFile) error {
	url := fmt.Sprintf(appDownloadFileRoute, appID)
	queryParamMap := map[string]string{
		"name": filename,
	}
	err := rpAPIInterface.HTTPCli.Get(url, file, queryParamMap)
	if err != nil {
		return errorInterpreter(err, appObjectType)
	}
	return nil
}

// ApplicationCreate asks the Redpesk backend to create a new application
//
// :param fatherPrj: Father project of the application to create
// :type fatherPrj: string
// :param app: pointer to the structure containing the data to create the application
// :type app: *rpapiv1.RPApplicationFull
// :param ans: buffer to stor the structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPApplicationFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationCreate(fatherPrj string, app *rpapiv1.RPApplicationFull, ans *rpapiv1.RPApplicationFull) error {
	var tmpAns *rpapiv1.RPApplicationFull

	// Check that the pointer to the structure to send is not nil
	if app == nil {
		return errors.New(appCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPApplicationFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(appCreationRoute, fatherPrj), app, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ApplicationUpdate asks the Redpesk backend to update an existing application
//
// :param appID: ID of the application to update
// :type appID: string
// :param app: pointer to the structure containing the data to create the application
// :type app: *rpapiv1.RPApplicationFull
// :param ans: buffer to stor the structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPApplicationFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationUpdate(appID string, app *rpapiv1.RPApplicationFull, ans *rpapiv1.RPApplicationFull) error {
	var tmpAns *rpapiv1.RPApplicationFull

	// Check that the pointer to the structure to send is not nil
	if app == nil {
		return errors.New(appCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPApplicationFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(appUpdateRoute, appID), app, tmpAns); err != nil {
		return errorInterpreter(err, appObjectType)
	}

	return nil
}

// ApplicationDelete asks the Redpesk backend to delete an existing application
//
// :param appID: ID of the application to delete
// :type appID: string
// :param ans: buffer to store the structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPApplicationFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationDelete(appID string, ans *rpapiv1.RPApplicationFull) error {
	var tmpAns *rpapiv1.RPApplicationFull

	if ans == nil {
		tmpAns = &rpapiv1.RPApplicationFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Delete(fmt.Sprintf(appDeletionRoute, appID), tmpAns); err != nil {
		return errorInterpreter(err, appObjectType)
	}

	return nil
}

// ApplicationBookmark asks the Redpesk backend to bookmark or reset the bookmark of an existing application
//
// :param appID: ID of the application to bookmark or reset
// :type appID: string
// :param ans: rpapiv1.RPApplicationBookmark structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPApplicationBookmark
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationBookmark(appID string, ans *rpapiv1.RPApplicationBookmark) error {
	var tmpAns *rpapiv1.RPApplicationBookmark

	if ans == nil {
		tmpAns = &rpapiv1.RPApplicationBookmark{}
	} else {
		tmpAns = ans
	}

	// Send the post request
	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(appBookmarkRoute, appID), nil, tmpAns); err != nil {
		return errorInterpreter(err, appObjectType)
	}

	return nil
}

// ApplicationBuild asks the Redpesk backend to build an existing application
//
// :param appID: ID of the application to build
// :type appID: string
// :param bldStartObj: Structure containing the fields needed to start the build
// :type bldStartObj: *rpapiv1.RPBuildStart
// :param ans: buffer to store the structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPBuilds
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationBuild(appID string, bldStartObj *rpapiv1.RPBuildStart, bldType rpapiv1.RPBuildType, ans *rpapiv1.RPBuilds) error {
	var tmpAns *rpapiv1.RPBuilds

	// Prepare the query parameters
	queryParamMap := map[string]string{}

	// Now add the right build type wanted
	queryParamMap["build_type"] = string(bldType)

	if ans == nil {
		tmpAns = &rpapiv1.RPBuilds{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(appBuildRoute, appID), bldStartObj, tmpAns, queryParamMap); err != nil {
		return errorInterpreter(err, appObjectType)
	}

	return nil
}

// ApplicationAudit asks the Redpesk backend to audit an existing application
//
// :param appID: ID of the application to audit
// :type appID: string
// :param ans: buffer to store the structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPAudit
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationAudit(appID string, input *rpapiv1.AuditStruct, ans *rpapiv1.RPAudit) error {
	var tmpAns *rpapiv1.RPAudit

	if ans == nil {
		tmpAns = &rpapiv1.RPAudit{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(appAuditRoute, appID), input, tmpAns); err != nil {
		return errorInterpreter(err, appObjectType)
	}

	return nil
}

// ApplicationsBuildsListGet gets the list of all existing builds in the given application.
//
// :param builds: Pointer to the rpapiv1.RPBuilds structure to fill
// :type builds: *rpapiv1.RPBuilds
// :param appID: application ID
// :type appID: string
// :param buildType: build type wanted for the list
// :type buildType: rpapiv1.RPBuildType
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationsBuildsListGet(builds *rpapiv1.RPBuilds, appID string, buildType rpapiv1.RPBuildType) error {

	// Initialize the call-back to save current page of builds
	savingCB := func(tmpObjects []byte) {
		tmpBuilds := rpapiv1.RPBuilds{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpBuilds)
		// Add the applications in the final object
		builds.Builds = append(builds.Builds, tmpBuilds.Builds...)
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Initialize the additional query parameters we want for this request
	additionalQueryParams := map[string]string{"build_type": string(buildType)}

	// Call the function allowing to get all the object
	err := pagingMngt.getAllObjects(rpAPIInterface, fmt.Sprintf(appBuildRoute, appID), "", additionalQueryParams, savingCB)
	if err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ApplicationsBuildsListGetByPage gets the list of all existing builds in the given application.
// but only one page of it
//
// :param builds: Pointer to the rpapiv1.RPBuilds structure to fill
// :type builds: *rpapiv1.RPBuilds
// :param appID: application ID
// :type appID: string
// :param buildType: build type wanted for the list
// :type buildType: rpapiv1.RPBuildType
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :param sortOpt: Optional parameter allowing to sort the list
// :type sortOpt: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationsBuildsListGetByPage(builds *rpapiv1.RPBuilds, appID string, buildType rpapiv1.RPBuildType, page, perPage int, sortOpt string) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":       strconv.Itoa(page),
		"per_page":   strconv.Itoa(perPage),
		"build_type": string(buildType),
	}

	// Add the sorting option if specified
	if sortOpt != "" {
		queryParamMap["sort"] = sortOpt
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(appBuildRoute, appID), builds, queryParamMap); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ApplicationsAuditsListGetByPage gets the list of audits in the given application.
// but only one page of it
//
// :param appID: application ID
// :type appID: string
// :param audits: Pointer to the rpapiv1.RPAudits structure to fill
// :type audits: *rpapiv1.RPAudits
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :param sortOpt: Optional parameter allowing to sort the list
// :type sortOpt: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationsAuditsListGetByPage(appID string, audits *rpapiv1.RPAudits, page, perPage int, sortOpt string) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
	}

	// Add the sorting option if specified
	if sortOpt != "" {
		queryParamMap["sort"] = sortOpt
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(appAuditRoute, appID), audits, queryParamMap); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ApplicationsTestsListGet gets the list of all existing tests in the given application.
// If a build ID is specified, all the tests for this build are retrieved.
//
// :param prjID: project ID
// :type prjID: string
// :param appID: application ID
// :type appID: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationsTestsListGet(prjID, appID string) ([]rpapiv1.RPTest, error) {

	// Call the route in order to get the tests stat for the given project
	prjStats := rpapiv1.RPTestStatsProject{}
	if err := rpAPIInterface.ProjectGetTestsStats(prjID, 0, &prjStats); err != nil {
		return []rpapiv1.RPTest{}, err
	}

	// Now find the right application
	for _, appStats := range prjStats.AppStats {
		if appStats.AppID.String() == appID {
			return appStats.LastCITests, nil
		}
	}

	// If we reach here, nothing has been found
	return []rpapiv1.RPTest{}, nil
}

// ApplicationsTestsListGetByPage gets the list of existing tests in the given application for a given page
//
// :param prjID: project ID
// :type prjID: string
// :param appID: application ID
// :type appID: string
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :param forceReq: Force the request to be sent and don't use the cache
// :type forceReq: bool
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationsTestsListGetByPage(prjID, appID string, page, perPage int, forceReq bool) (rpapiv1.RPTests, error) {

	// Manage a cache since there is not paging managed on the backend side.
	var err error
	if (cachedAppRPTests == nil) || forceReq {
		cachedAppRPTests, err = rpAPIInterface.ApplicationsTestsListGet(prjID, appID)
		if err != nil {
			return rpapiv1.RPTests{}, err
		}
	}

	// Initialize the structure that we will return
	retStruct := rpapiv1.RPTests{
		Page:         page,
		PerPage:      perPage,
		TotalEntries: len(cachedAppRPTests),
		Tests:        []rpapiv1.RPTest{},
	}

	// Calculate the minimum and maximum index wanted by the caller
	minIndex := (page - 1) * perPage
	maxIndex := (page * perPage) - 1

	// Loop over the cached tests
	for index, test := range cachedAppRPTests {
		if (index >= minIndex) && (index <= maxIndex) {
			retStruct.Tests = append(retStruct.Tests, test)
		}
	}

	// If we reach here, nothing has been found
	return retStruct, nil
}

// ApplicationGetFile request a file from the backend. Particularly, it can be a spec file.
//
// :param appID: ID of the application the file is about
// :type appID: string
// :param fileName: Name of the file to retrieve
// :type fileName: string
// :param rpFile: Pointer to the file structure to fill with the answer
// :type rpFile: *rpapiv1.RPFile
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationGetFile(appID string, fileName string, rpFile *rpapiv1.RPFile) error {

	// Create the URL parameters to pass
	queryParamMap := map[string]string{
		"name": fileName,
	}

	// Request the file from the backend
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(appUploadFileRoute, appID), rpFile, queryParamMap); err != nil {
		return errorInterpreter(err, "")
	}

	// Decode the content and update it in rpFile
	codedContent := rpFile.Content
	decodedContentBytes, err := base64.StdEncoding.DecodeString(codedContent)
	if err != nil {
		return errorInterpreter(err, "")
	}

	// Save the decoded content in the structure
	rpFile.Content = string(decodedContentBytes)

	return nil
}

// ApplicationUploadFile post a file on the backend. Particularly, it can be a spec file.
//
// :param appID: ID of the application where the file needs to be uploaded
// :type appID: string
// :param rpFile: Pointer to the structure containing the content and the name of the file to upload
// :type rpFile: rpapiv1.RPFile
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationUploadFile(appID string, rpFile *rpapiv1.RPFile) error {

	// Check that the pointer to the structure to send is not nil
	if rpFile == nil {
		return errors.New(appUploadNilError)
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(appUploadFileRoute, appID), rpFile, nil); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ApplicationGetBranchesList sends a request to the backend in order to retrieve the list of
// branches for the given application
//
// :param appID: UUID of the application for which we want the branches
// :type appID: string
// :param rpBranches: Pointer to the structure to fill with the list of branches
// :type rpBranches: rpapiv1.RPBranchList
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationGetBranchesList(appID string, rpBranches *rpapiv1.RPBranchList) error {
	// Get the list of branches for the given application
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(appBranchesRoute, appID), rpBranches); err != nil {
		return errorInterpreter(err, appObjectType)
	}
	return nil
}

// ApplicationCreateOrSwitchBranch sends a request to the backend in order to create a branch
// or to switch to an existing branch for the given application
//
// :param appID: UUID of the application where a branch needs to be created
// :type appID: string
// :param rpBranch: Pointer to the structure containing the right values to create/switch to a branch
// :type rpBranch: rpapiv1.RPBranchList
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationCreateOrSwitchBranch(appID string, rpBranch *rpapiv1.RPBranch) error {
	// Check that the pointer to the structure to send is not nil
	if rpBranch == nil {
		return errors.New(rpBranchNilError)
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(appBranchesRoute, appID), rpBranch, nil); err != nil {
		return errorInterpreter(err, appObjectType)
	}
	return nil
}

// ApplicationDeleteBranch sends a request to the backend in order to delete a branch
// for the given application
//
// :param appID: UUID of the application where a branch needs to be deleted
// :type appID: string
// :param rpBranch: Pointer to the structure containing the name of the branch to delete
// :type rpBranch: rpapiv1.RPBranchName
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ApplicationDeleteBranch(appID string, rpBranch *rpapiv1.RPBranchName) error {
	// Check that the pointer to the structure to send is not nil
	if rpBranch == nil {
		return errors.New(rpBranchNameNilError)
	}

	if err := rpAPIInterface.HTTPCli.DeleteParameter(fmt.Sprintf(appBranchesRoute, appID), rpBranch, nil); err != nil {
		return errorInterpreter(err, appObjectType)
	}
	return nil
}

// ApplicationPostAction request an action from the backend for a specified application
//
// :param appID: app ID
// :type appID: string
// :param action: action to perform
// :type action: string
// :param postBody: Structure to pass as the body of the request
// :type postBody: interface{}
// :returns: error -- error if something went wrong
func (rpAPIInterface *RPBCKAPI) ApplicationPostAction(appID, action string, postBody interface{}) error {

	// Create the URL parameters to pass
	queryParamMap := map[string]string{
		"action": action,
	}

	// Send the appropriate POST request
	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(appActionRoute, appID), postBody, nil, queryParamMap); err != nil {
		return errorInterpreter(err, appObjectType)

	}
	return nil
}

// ApplicationCopy copies the application to another project
//
// :param appID: app ID
// :type appID: string
// :param rpAppCopy: Structure containing destination-project of the copied application
// :type rpAppCopy: rpapiv1.RPApplicationCopy
// :returns: error -- error if something went wrong
func (rpAPIInterface *RPBCKAPI) ApplicationCopy(appID string, rpAppCopy *rpapiv1.RPApplicationCopy) error {

	// Check that the pointer to the structure to send is not nil
	if rpAppCopy == nil {
		return errors.New(rpApplicationCopyNilError)
	}

	// Send the request
	return rpAPIInterface.ApplicationPostAction(appID, "copy", rpAppCopy)
}

// ApplicationRefreshFiles refresh the files of the gitpkg
//
// :param appID: app ID
// :type appID: string
// :param rpAppCopy: Structure containing destination-project of the copied application
// :type rpAppCopy: rpapiv1.RPApplicationCopy
// :returns: error -- error if something went wrong
func (rpAPIInterface *RPBCKAPI) ApplicationRefreshFiles(appID string) error {
	// Send the request
	return rpAPIInterface.ApplicationPostAction(appID, "refresh_files_package", nil)
}
